// LiveDisplays.h is part of the LiveDisplays plug-in to PYTHIA 8
// Copyright (C) 2019 P. Skands
// LiveDisplays is licenced under the GNU GPL version 2 or later.
// Please respect the MCnet Guidelines for Event Generator Authors and Users,
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the LiveDisplays utilities

#ifndef LIVEDISPLAYS_H
#define LIVEDISPLAYS_H

// Main ROOT headers used by the top-level LiveDisplays App.
#include "TROOT.h"
#include "TApplication.h"
#include "TCanvas.h"
#include "TH1.h"
#include "TStyle.h"
#include "TSystem.h"

// Root headers used by the Display class
#include "TCanvas.h"
#include "TColor.h"
#include "TH1.h"
#include "TH2.h"
#include "TImage.h"
#include "TLatex.h"
#include "TLegend.h"
#include "TLegendEntry.h"
#include "TLine.h"
#include "TStyle.h"
#include "TSystem.h"

// Auxiliary headers used by the Display class
#include "LiveDisplays/VRfile.h"
#include "LiveDisplays/VRgraph.h"

// Pythia 8 headers
#include "Pythia8/Basics.h"
#include "Pythia8/Info.h"
#include "Pythia8/Settings.h"
#include "Pythia8/Pythia.h"
#include "Pythia8/PythiaStdlib.h"

// Put everything in the Pythia 8 namespace

namespace Pythia8 {

class LiveDisplaysApp;

//==========================================================================

// The LiveDisplaysApp class
//
// Handles communication with the computer's windowing system and
// overall tasks such as updating (all) displays.
// Exactly one instance must be created per run

class LiveDisplaysApp : public TApplication {

 public:

  // Constructor (requires a pointer to a Pythia 8 settings database)
  LiveDisplaysApp(Pythia* pythiaPtrIn)
    : TApplication("LiveDisplaysApp",0,0) {

    // Save pointer to Pythia (so we can access Settings and Info)
    pythiaPtr = pythiaPtrIn;

    // Initialise Graphics
    InitializeGraphics();
    gROOT->SetStyle("Plain");
    // Set default hatching width thick enough to show up well on slides
    gStyle->SetHatchesLineWidth(2);
    // Speed up drawing, with openGL
    gStyle->SetCanvasPreferGL(kTRUE);

    // Switch on storage of stat uncertainties for all TH1 histograms
    TH1::SetDefaultSumw2();

    // Initialise update and save frequencies
    nNextUpdateSav = pythiaPtr->settings.mode("LiveDisplays:nUpdate");
    deltaUpdate    = nNextUpdateSav;
    nNextSaveSav   = pythiaPtr->settings.mode("LiveDisplays:nSave");
    deltaSave      = nNextSaveSav;
    fSave          = pythiaPtr->settings.parm("LiveDisplays:fSave");
  }

  // Update (and save) display(s) if settings function(s) tell us to
  bool update(bool forceUpdate=false, bool forceSave=false);

  // Query when we will do next save or update
  int nNextSave() { return nNextSaveSav; }
  int nNextUpdate() { return nNextUpdateSav; }

 private:

  // Global settings
  Pythia* pythiaPtr;

  // Number of calls to update
  double nNextSaveSav, nNextUpdateSav, deltaUpdate, deltaSave, fSave;

};


//==========================================================================

// The LiveDisplay class

// Main display class inheriting from ROOT's TCanvas class. Any number of
// LiveDisplay instances can be created, each containing a single plot.

class LiveDisplay : public TCanvas {

 public:

  // Dummy constructor for empty canvas
  LiveDisplay(string displayNameIn, int sizeX=0, int sizeY=0,
    int offsetX=0, int offsetY=0)
    : TCanvas(displayNameIn.c_str(), displayNameIn.c_str(),
      -offsetX, offsetY, sizeX, sizeY),
        nPad(1), isInitPad(false), hasAxesSav(false), has2d(false),
        hasDataSav(false), hasTitle(false),
        hasLegend(false), forceAlign(false),
        hasLegendChi2(false), showChi2sav(false),
        hasLegendUncVar(false), hasSubTitle1(false), hasSubTitle2(false),
        hasLeftLabel(false), hasRightLabel(false),
        hasLeftLabelPtr(false), hasRightLabelPtr(false),
        rebinLikeDataSav(false), zeroSkippingSav(false) {
    displayNameSav   = displayNameIn;
  }

  // Constructor from TH1D* and Pythia*
  LiveDisplay(TH1D* hisPtr, Pythia* pythiaPtr, int sizeX=0, int sizeY=0,
    int offsetX=0, int offsetY=0)
    : TCanvas(string(string(hisPtr->GetName())+"LiveDisplay").c_str(),
      hisPtr->GetName(), -(offsetX > 0 ? offsetX :
        pythiaPtr->settings.mode("LiveDisplays:offsetX")),
      (offsetY > 0 ? offsetY :
        pythiaPtr->settings.mode("LiveDisplays:offsetY")),
      (sizeX > 0 ? sizeX :
        pythiaPtr->settings.mode("LiveDisplays:sizeX")),
      (sizeY > 0 ? sizeY :
        pythiaPtr->settings.mode("LiveDisplays:sizeY"))),
        nPad(1), isInitPad(false), hasAxesSav(false), has2d(false),
        hasDataSav(false), hasTitle(false),
        hasLegend(false), forceAlign(false),
        hasLegendChi2(false), showChi2sav(false),
        hasLegendUncVar(false), hasSubTitle1(false), hasSubTitle2(false),
        hasLeftLabel(false), hasRightLabel(false),
        hasLeftLabelPtr(false), hasRightLabelPtr(false),
        rebinLikeDataSav(false), zeroSkippingSav(false) {
    displayNameSav   = hisPtr->GetName();
    init(pythiaPtr);
    addTH1DPtr(hisPtr);
    int offsetXnew = (offsetX > 0 ? offsetX :
      pythiaPtr->settings.mode("LiveDisplays:offsetX"))
      + pythiaPtr->settings.mode("LiveDisplays:cascadeX");
    pythiaPtr->settings.mode("LiveDisplays:offsetX",offsetXnew);
    int offsetYnew = (offsetY > 0 ? offsetY :
      pythiaPtr->settings.mode("LiveDisplays:offsetY"))
      + pythiaPtr->settings.mode("LiveDisplays:cascadeY");
    pythiaPtr->settings.mode("LiveDisplays:offsetY",offsetYnew);
  }

  // Constructor from a vector of TH1D* and Pythia*
  LiveDisplay(vector<TH1D*> hisPtr, Pythia* pythiaPtr,
    int sizeX=0, int sizeY=0, int offsetX=0, int offsetY=0)
    : TCanvas(string(string(hisPtr[0]->GetName())+"LiveDisplay").c_str(),
      hisPtr[0]->GetName(), -(offsetX > 0 ? offsetX :
        pythiaPtr->settings.mode("LiveDisplays:offsetX")),
      (offsetY > 0 ? offsetY :
        pythiaPtr->settings.mode("LiveDisplays:offsetY")),
      (sizeX > 0 ? sizeX :
        pythiaPtr->settings.mode("LiveDisplays:sizeX")),
      (sizeY > 0 ? sizeY :
        pythiaPtr->settings.mode("LiveDisplays:sizeY"))),
        nPad(1), isInitPad(false), hasAxesSav(false), has2d(false),
        hasDataSav(false), hasTitle(false),
        hasLegend(false), forceAlign(false),
        hasLegendChi2(false), showChi2sav(false),
        hasLegendUncVar(false), hasSubTitle1(false), hasSubTitle2(false),
        hasLeftLabel(false), hasRightLabel(false),
        hasLeftLabelPtr(false), hasRightLabelPtr(false),
        rebinLikeDataSav(false), zeroSkippingSav(false) {
    displayNameSav   = hisPtr[0]->GetName();
    init(pythiaPtr);
    addTH1DPtr(hisPtr);
    int offsetXnew = (offsetX > 0 ? offsetX :
      pythiaPtr->settings.mode("LiveDisplays:offsetX"))
      + pythiaPtr->settings.mode("LiveDisplays:cascadeX");
    pythiaPtr->settings.mode("LiveDisplays:offsetX",offsetXnew);
    int offsetYnew = (offsetY > 0 ? offsetY :
      pythiaPtr->settings.mode("LiveDisplays:offsetY"))
      + pythiaPtr->settings.mode("LiveDisplays:cascadeY");
    pythiaPtr->settings.mode("LiveDisplays:offsetY",offsetYnew);
  }

  // Constructor from TH2D* and Pythia*
  // (Still not as fully developed as for TH1D)
  LiveDisplay(TH2D* hisPtr, Pythia* pythiaPtr, int sizeX=0, int sizeY=0,
    int offsetX=0, int offsetY=0)
    : TCanvas(string(string(hisPtr->GetName())+"LiveDisplay").c_str(),
      hisPtr->GetName(),
      (offsetX > 0 ? offsetX :
        pythiaPtr->settings.mode("LiveDisplays:offsetX")),
      (offsetY > 0 ? offsetY :
        pythiaPtr->settings.mode("LiveDisplays:offsetY")),
      (sizeX > 0 ? sizeX :
        pythiaPtr->settings.mode("LiveDisplays:sizeX")),
      (sizeY > 0 ? sizeY :
        pythiaPtr->settings.mode("LiveDisplays:sizeY"))),
        nPad(1), isInitPad(false), hasAxesSav(false), has2d(true),
        hasDataSav(false), hasTitle(false),
        hasLegend(false), forceAlign(false),
        hasLegendChi2(false), showChi2sav(false),
        hasLegendUncVar(false), hasSubTitle1(false), hasSubTitle2(false),
        hasLeftLabel(false), hasRightLabel(false),
        hasLeftLabelPtr(false), hasRightLabelPtr(false),
        rebinLikeDataSav(false), zeroSkippingSav(false) {
    displayNameSav   = hisPtr->GetName();
    init(pythiaPtr);
    setTH2DPtr(hisPtr);
    int offsetXnew = (offsetX > 0 ? offsetX :
      pythiaPtr->settings.mode("LiveDisplays:offsetX"))
      + pythiaPtr->settings.mode("LiveDisplays:cascadeX");
    pythiaPtr->settings.mode("LiveDisplays:offsetX",offsetXnew);
    int offsetYnew = (offsetY > 0 ? offsetY :
      pythiaPtr->settings.mode("LiveDisplays:offsetY"))
      + pythiaPtr->settings.mode("LiveDisplays:cascadeY");
    pythiaPtr->settings.mode("LiveDisplays:offsetY",offsetYnew);
  }

  // Init pointers and set settings
  void init(Pythia* pythiaPtr);

  // Functions to Get
  bool logY() { return logYsav; }
  bool linY() { return !logYsav; }
  bool logX() { return logXsav; }
  bool linX() { return !logXsav; }
  bool normalize() { return normalizeSav; }
  int  optStat() { return optStatSav; }
  double normFac() { return normFacSav; }
  string name() { return displayNameSav; }

  // Functions to check whether axes and data have been defined
  bool hasAxes() {return hasAxesSav;}
  bool hasData() {return hasDataSav;}

  // Add pad (for ratio to data or decomposition of vector of histos)
  void addPad();

  // Goto pad
  void gotoPad(int iPadIn = 1);

  // Functions to Set
  void verbose(bool verboseIn) {verboseSav = verboseIn;};
  void optStat(int optStatIn) {optStatSav = optStatIn;}
  void normalize(bool normalizeIn = true, double normFacIn = 1.0) {
    normalizeSav  = normalizeIn;
    normFacSav    = normFacIn;
  }
  void logY(bool logYin = true) {
    logYsav = logYin;
    cd(1);
    if (logYin) gPad->SetLogy();
    else gPad->SetLogy(0);
    gPad->cd(1);
    if (logYin) gPad->SetLogy();
    else gPad->SetLogy(0);
  }
  void linY(bool linYin) { logY(!linYin); }
  void logX(bool logXin) {
    logXsav = logXin;
    gPad->cd(1);
    if (logXin) {
      gPad->SetLogx();
    }
    else gPad->SetLogx(0);
  }
  void linX(bool linXin) { logX(!linXin); }
  void forceLeftLegend(bool leftLegendIn=true) { leftLegend = leftLegendIn; }
  void showRatio(bool showRatioIn=true) {
    // Remove ratio pad if being shown
    if (showRatioIn == false && nPad >= 2 && showRatioSav) nPad--;
    showRatioSav = showRatioIn;
  }
  void ratioRange(double ratioMinIn, double ratioMaxIn) {
    ratioMinSav = ratioMinIn;
    ratioMaxSav = ratioMaxIn;
    if (!hasAxes()) return;
    axesNormPtr->SetMinimum(ratioMinSav);
    axesNormPtr->SetMaximum(ratioMaxSav);
    // Make copy of x axis for overdrawing
    int nBin = axesNormPtr->GetNbinsX();
    delete xCtrPtr;
    delete xUppPtr;
    xCtrPtr = new TLine(axesNormPtr->GetBinLowEdge(1),1.0,
      axesNormPtr->GetBinLowEdge(nBin+1),1.0);
    xUppPtr = new TLine(axesNormPtr->GetBinLowEdge(1),ratioMaxSav,
      axesNormPtr->GetBinLowEdge(nBin+1),ratioMaxSav);
  }
  void showRelUnc(bool showRelUncIn=true) {
    showRelUncSav = showRelUncIn;
  }

  // Check if axes and/or data has been defined
  void setAxes(TH1D* initHis);

  // Set specific color scheme for specific graph
  // Return false if graph does not exist or color scheme not defined
  bool setColorScheme(int iScheme, int iGraph=1) {
    int nGraphs    = graphs.size();
    int nGraphsHis = graphsHis.size();
    int nGraphsTot = nGraphs+nGraphsHis;
    // Set color scheme for data
    if (iGraph == 0) {
      return dataPtr->setColorScheme(iScheme);
    }
    // Set color schem for histograms
    else if (iGraph <= nGraphs) {
      return graphs[iGraph-1].setColorScheme(iScheme);
    }
    else if (iGraph <= nGraphsTot) {
      return graphsHis[iGraph-1-nGraphs].setColorScheme(iScheme);
    }
    else {
      return false;
    }
    return true;
  }

  //------------------------------------------------------------------
  // LiveDisplay: TH1D I/O

  // Add a histogram, not necessarily the first
  void addTH1DPtr(TH1D* hisPtrIn) {
    if (!hasAxes()) {
      setAxes(hisPtrIn);
      graphs.resize(0);
    }
    graphs.push_back(VRgraph(hisPtrIn));
    // Default: assign color set n to graph number n
    graphs[graphs.size()-1].setColorScheme(graphs.size()+graphsHis.size());
    // Check if we should (re)set legend
    if (!hasLegend) setLegend(legendSav);
    // Check if we should add a ratio pad
    if (graphs.size() >= 2 && showRatioSav && nPad <= 1) addPad();
  }

  // Add a 2D histogram (replaces any other histograms)
  void setTH2DPtr(TH2D* hisPtrIn) {
    // Set TH2D pointer
    th2dPtr = hisPtrIn;
    has2d   = true;
    // Switch off stats options
    th2dPtr->SetStats(optStatSav);
    // Reset all 1D content
    graphs.resize(0);
    nPad = 1;
    //
    cd();
    // If several pads, first do main plot
    gotoPad(1);
    // Make sure we have room to plot Z (colour) scale
    gPad->SetRightMargin(0.145);
    gPad->SetBottomMargin(0.15);
    gPad->SetTopMargin(0.1);
    gPad->SetLeftMargin(0.125);
    // Set color scale
    const Int_t NRGBs = 6;
    const Int_t NCont = 20;
    Int_t MyPalette[NCont];
    Double_t r[NRGBs]    = {0., 0.2, 0.4, 0.6, 1.0, 1.0};
    Double_t g[NRGBs]    = {0., 0.0, 0.4, 0.6, 1.0, 0.7};
    Double_t b[NRGBs]    = {0., 0.7, 0.4, 0.6, 0.2, 0.7};
    Double_t stop[NRGBs] = {0., 0.25, 0.4, .58, 0.7, 1.0};
    Int_t FI = TColor::CreateGradientColorTable(NRGBs, stop, r, g, b, NCont);
    for (int i=0;i<NCont;i++) MyPalette[i] = FI+i;
    //MyPalette[0]=0;
    gStyle->SetPalette(NCont,MyPalette);
    gStyle->SetNumberContours(NCont);
  }

  // Add a histogram (with vector uncertainties), not necessarily the first
  void addTH1DPtr(vector<TH1D*> hisPtrIn) {
    if (!hasAxes()) {
      setAxes(hisPtrIn[0]);
      graphs.resize(0);
    }
    graphs.push_back(VRgraph(hisPtrIn));
    graphs[graphs.size()-1].setColorScheme(graphs.size()+graphsHis.size());

    // Check if we should (re)set legend
    if (!hasLegend) setLegend(legendSav);

    // Add pad for ratios
    if (graphsHis.size() + graphs.size() >= 2 && nPad <= 1
      && showRatioSav) addPad();

    // Add pad for decomposition of vector band
    if (hisPtrIn.size() >= 2 && nPad <= 2 && showRelUncSav) addPad();

  }

  // Get histogram pointer corresponding to graph i, variation j
  TH1D* getTH1DPtr(int i=0, int j=0) {
    if ( i <= int(graphs.size())-1 ) return graphs[i].getTH1DPtr(j);
    else return NULL;
  }

  //------------------------------------------------------------------
  // LiveDisplay: FILE I/O

  // Function to convert a list of integers to vector<int>
  void getFormat(vector<int>& format, int i1, int i2, int i3=-1, int i4=-1,
    int i5=-1, int i6=-1, int i7=-1, int i8=-1, int i9=-1, int i10=-1,
    int i11=-1, int i12=-1, int i13=-1, int i14=-1, int i15=-1, int i16=-1,
    int i17=-1, int i18=-1, int i19=-1) {
    // Store file format
    format.push_back(i1);
    format.push_back(i2);
    if (i3 != -1) format.push_back(i3); else return;
    if (i4 != -1) format.push_back(i4); else return;
    if (i5 != -1) format.push_back(i5); else return;
    if (i6 != -1) format.push_back(i6); else return;
    if (i7 != -1) format.push_back(i7); else return;
    if (i8 != -1) format.push_back(i8); else return;
    if (i9 != -1) format.push_back(i9); else return;
    if (i10 != -1) format.push_back(i10); else return;
    if (i11 != -1) format.push_back(i11); else return;
    if (i12 != -1) format.push_back(i12); else return;
    if (i13 != -1) format.push_back(i13); else return;
    if (i14 != -1) format.push_back(i14); else return;
    if (i15 != -1) format.push_back(i15); else return;
    if (i16 != -1) format.push_back(i16); else return;
    if (i17 != -1) format.push_back(i17); else return;
    if (i18 != -1) format.push_back(i18); else return;
    if (i19 != -1) format.push_back(i19); else return;
  }

  // Set file to be treated as experimental data
  bool setDataFile(string dataFileName, bool hasSysErrors = true) {
    if (hasSysErrors)
      return setDataFile(dataFileName,10,20,30,40,51,52,61,62);
    else return setDataFile(dataFileName,10,20,30,40,51,52);
  }
  bool setDataFile(string dataFileName, int i1, int i2, int i3=-1, int i4=-1,
    int i5=-1, int i6=-1, int i7=-1, int i8=-1, int i9=-1, int i10=-1,
    int i11=-1, int i12=-1, int i13=-1, int i14=-1, int i15=-1, int i16=-1,
    int i17=-1, int i18=-1, int i19=-1);

  // Add file containing saved histogram(s) (eg from previous runs)
  // (Default is to ignore any additional systematic errors)
  bool addHisFile(string hisFileName, bool hasSysErrors = false) {
    if (hasSysErrors) return addHisFile(hisFileName,10,20,30,40,51,52,61,62);
    else return addHisFile(hisFileName,10,20,30,40,51,52);
  }
  bool addHisFile(string hisFileName, int i1, int i2, int i3=-1, int i4=-1,
    int i5=-1, int i6=-1, int i7=-1, int i8=-1, int i9=-1, int i10=-1,
    int i11=-1, int i12=-1, int i13=-1, int i14=-1, int i15=-1, int i16=-1,
    int i17=-1, int i18=-1, int i19=-1);

  // force update of histogram files, useful when performing several runs
  // in parallel, with the contents of data files changing dynamically.
  bool refreshHisFiles();

  //------------------------------------------------------------------
  // Update LiveDisplay
  void update();
  // Wrapper to update as generic TCanvas, using ROOT's Notify method.
  virtual Bool_t Notify() { update(); return true; }

  // Helper function to add a simple line to display.
  void addTLine(TLine* newLine) {
    linesPtr.push_back(newLine);
  }

  // Save display image to graphics file, with access to same
  // formats as ROOT TPad::Print()
  void saveLiveDisplay(string baseNameIn="none", string fileType="def");

  // save main theory histogram to data file (in standard column format)
  // Note: extended format allows theory uncertainty variations to be included
  bool saveHistogram(string baseNameIn="none", string ext="dat");

  // Print out chi2 values, first of MC curves, then of saved histograms
  void printChi2();

  // Set title
  void title(string titleIn);
  void titleSize(double sizeIn) { titleTexPtr->SetTextSize(sizeIn); }

  // Set Subtitle 1 properties (usually used for Data reference)
  void subTitle1(string subTitleIn);
  void subTitle1y(double yIn) { subTitle1ySav = yIn; }
  void subTitle1size(double sizeIn) { subTitle1Ptr->SetTextSize(sizeIn); }

  // Set Subtitle 2 properties (usually used for MC names and versions)
  void subTitle2(string subTitleIn);
  void subTitle2y(double yIn) { subTitle2ySav = yIn; }
  void subTitle2size(double sizeIn) { subTitle2Ptr->SetTextSize(sizeIn); }

  // Access to TLatex pointers of titles, for user modifications
  TLatex* title() { return titleTexPtr;}
  TLatex* subTitle1() {return subTitle1Ptr;}
  TLatex* subTitle2() {return subTitle2Ptr;}

  // Set Legend
  // Optionally force alignment: align = -1(left), 0(auto), 1(right)
  void setLegend(vector<string> legendIn, int alignIn=0);
  void setLegend(vector<string> legendIn, double xLeg,
    double xWid=0.33, double dy=0.056, double yLegTop=0.86);

  // Optionally show chi2 information in legend
  void showChi2(bool showChi2in = true) { showChi2sav = showChi2in; }
  bool showChi2() { return showChi2sav; }

  //------------------------------------------------------------------------
  // setLegendUncVar():
  // Set Legend for uncertainty bands
  void setUncertaintyLabels(vector<string> labelsIn);

  //------------------------------------------------------------------------
  // setLeftLabel(), setRightLabel()
  // Set Labels for beam/energy type (top left and right white-on-black)
  void leftLabel(string labelIn) {
    if (labelIn != "" && labelIn != " ")
      hasLeftLabel = true;
    else
      hasLeftLabel = false;
    leftLabelSav = labelIn;
  }
  void rightLabel(string labelIn) {
    if (labelIn != "" && labelIn != " ")
      hasRightLabel = true;
    else
      hasRightLabel = false;
    rightLabelSav = labelIn;
  }

  // Set rebinning properties
  void rebinLikeData(bool rebinLikeDataIn=true) {
    rebinLikeDataSav = rebinLikeDataIn;
    // Check if rebinning histograms to same binning as data
    if (hasDataSav && dataPtr!= 0 && rebinLikeDataSav) {
      for (int iGraph=0; iGraph<int(graphs.size()); ++iGraph)
        graphs[iGraph].setRebinAxis(dataPtr,zeroSkippingSav);
      for (int iGraph=0; iGraph<int(graphsHis.size()); ++iGraph)
        graphsHis[iGraph].setRebinAxis(dataPtr,zeroSkippingSav);
    }
  }
  void zeroSkipping(bool zeroSkippingIn=true) {
    zeroSkippingSav  = zeroSkippingIn;
    for (int iGraph=0; iGraph<int(graphs.size()); ++iGraph)
      graphs[iGraph].zeroSkipping(zeroSkippingIn);
    for (int iGraph=0; iGraph<int(graphsHis.size()); ++iGraph)
      graphsHis[iGraph].zeroSkipping(zeroSkippingIn);
  }

 private:

  // LiveDisplay: private data members and functions

  // Canvas name
  string displayNameSav;

  // Internal variables
  bool logYsav, logXsav, normalizeSav, verboseSav;
  bool showRatioSav, showRelUncSav, ratioLogYsav;
  int optStatSav;
  double normFacSav, ratioMinSav, ratioMaxSav;

  // Input containers
  VRfile* dataFilePtr;
  vector<VRfile*> hisFilePtrs;

  // PADS
  int nPad;
  bool isInitPad;

  // PLOT

  // Use histogram to define axes
  TH1D* axesPtr;
  TH1D* axesNormPtr;
  TH1D* axesVarPtr;
  TLine* xCtrPtr;
  TLine* xUppPtr;
  bool hasAxesSav;

  // Optional possibility to use display for 2D histograms
  TH2D* th2dPtr;
  bool has2d;

  // MC Graphs
  vector<VRgraph> graphs;
  vector<VRgraphNorm> graphsNorm;

  // His graphs (for comparison to saved data files)
  vector<VRgraphHis> graphsHis;
  vector<VRgraphNorm> graphsHisNorm;
  vector<string> hisFileNames;

  // RelUnc pad
  vector<VRgraphVar> graphsVar;

  // Data graph (and normlaized one- and two-sigma variants)
  VRgraphData* dataPtr;
  VRgraphData* dataTwoSigmaPtr;
  VRgraphNorm* dataNormPtr;
  VRgraphNorm* dataTwoSigmaNormPtr;
  bool hasDataSav;

  // Plot title
  TLatex* titleTexPtr;
  string titleSav;
  bool hasTitle;

  // Objects
  vector<TLine*> linesPtr;

  // LEGEND

  // Main Legend
  TLegend* legendPtr;
  bool hasLegend;
  bool leftLegend;
  vector<string> legendSav;
  int alignSav;
  bool forceAlign;
  double xLegSav, xWidSav, dySav, yLegTopSav;

  // Chi2 Legend
  TLegend* legendChi2Ptr;
  bool hasLegendChi2, showChi2sav;

  // Uncertainty band variation legends
  vector<string> uLabelsSav;
  TLegend* legendUncVarTopPtr;
  TLegend* legendUncVarBotPtr;
  bool hasLegendUncVar;

  // Title
  double titleSizeSav;

  // Subtitle 1: normally used for data reference
  TLatex* subTitle1Ptr;
  bool hasSubTitle1;
  double subTitle1ySav, subTitle1sizeSav;

  // Subtitle 2: normally used for MC names and versions
  TLatex* subTitle2Ptr;
  bool hasSubTitle2;
  string subTitle2Sav;
  double subTitle2ySav, subTitle2sizeSav;

  // Left and right labels
  string leftLabelSav, rightLabelSav;
  TLatex* leftLabelPtr;
  TLatex* rightLabelPtr;
  bool hasLeftLabel, hasRightLabel;
  bool hasLeftLabelPtr, hasRightLabelPtr;

  // Rebinning properties
  bool rebinLikeDataSav, zeroSkippingSav;

  // Save formats
  bool savePDFsav, saveEPSsav, savePNGsav, saveJPEGsav, saveROOTsav, saveTEXsav;
  string saveFilePrefixSav, saveFileSuffixSav;

};

//==========================================================================

// Simple utility to convert a Pythia Hist to a ROOT TH1 object.
// Note that statistical uncertainties are not preserved.

void Hist2TH1(TH1* th1Ptr, Pythia8::Hist* histPtr);

} // End namespace Pythia8

#endif

