// VRfile.h is a part of the LiveDisplays plug-in to the PYTHIA 8 generator.
// Copyright (C) 2019 P. Skands.
// LiveDisplays is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users,
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for VRfile

#ifndef LIVEDISPLAYSFILE_H
#define LIVEDISPLAYSFILE_H

// Include Pythia's standard libraries
#include "Pythia8/PythiaStdlib.h"

// Put everything in the Pythia8 namespace

namespace Pythia8 {

// ------------------------------------------------------------------------
// class VRfile:
// Base class to contain tabulated histograms
// (either saved MC histos or experimental data points)
// Base class implements "standard" format

class VRfile {

 public:

  // Constructor
  VRfile() {}

  // Constructor
  VRfile(vector<int> formatIn) {
    setFormat(formatIn);
  }

  // Destructor
  virtual ~VRfile() {}

  // Reset data containers
  void init(bool verboseIn = false) {
    xLsav.resize(0);
    xCsav.resize(0);
    xRsav.resize(0);
    ySav.resize(0);
    uTotHiSav.resize(0);
    uTotLoSav.resize(0);
    uStatHiSav.resize(0);
    uStatLoSav.resize(0);
    hasNamesSav = false;
    pointNameSav.resize(0);
    fileSRC = "none";
    fileREF = "none";
    fileECM = "none";
    verboseSav = verboseIn;
    sumWeights = 0.;
    nSysSav = 0;
  }

  void setFormat(vector<int> formatIn) {
    format.resize(0);
    for (unsigned int iFmt=0; iFmt<formatIn.size(); ++iFmt) {
      format.push_back(formatIn[iFmt]);
    }
  }

  // Read data points from file
  bool readFile(string fileNameIn, bool verboseIn=false);

  // Read line from data file.
  // Made virtual so inherited classes can use different data formats.
  virtual bool readLine(ifstream& inStream);

  // Finalize read-in: fill in empty arrays and compute normalization.
  // Also made virtual so can be adapted to other data formats, if needed.
  virtual bool finalize();

  // Size
  int size() {return min(xCsav.size(),ySav.size());}
  int nSys() {return nSysSav;}

  // Functions to get stored data
  double getXL(int i) {
    return (i < int(xLsav.size()) && i >= 0) ? xLsav[i] : 0.0;}
  double getXC(int i) {
    return (i < int(xCsav.size()) && i >= 0) ? xCsav[i] : 0.0;}
  double getXR(int i) {
    return (i < int(xRsav.size()) && i >= 0) ? xRsav[i] : 0.0;}
  double getY(int i) {
    return (i < int(ySav.size()) && i >= 0) ? ySav[i] : 0.0;}
  double getUtotHi(int i) {
    return (i < int(uTotHiSav.size()) && i >= 0) ? uTotHiSav[i] : 0.0;}
  double getUtotLo(int i) {
    return (i < int(uTotLoSav.size()) && i >= 0) ? uTotLoSav[i] : 0.0;}
  double getUstatHi(int i) {
    return (i < int(uStatHiSav.size()) && i >= 0) ? uStatHiSav[i] : 0.0;}
  double getUstatLo(int i) {
    return (i < int(uStatLoSav.size()) && i >= 0) ? uStatLoSav[i] : 0.0;}
  double getUsysHi(int i) {
    return (i < int(uSysHiSav.size()) && i >= 0) ? uSysHiSav[i] : 0.0;}
  double getUsysLo(int i) {
    return (i < int(uSysLoSav.size()) && i >= 0) ? uSysLoSav[i] : 0.0;}
  double getUsys2Hi(int i) {
    return (i < int(uSys2HiSav.size()) && i >= 0) ? uSys2HiSav[i] : 0.0;}
  double getUsys2Lo(int i) {
    return (i < int(uSys2LoSav.size()) && i >= 0) ? uSys2LoSav[i] : 0.0;}


  double getSumWeights() { return sumWeights; }
  string getSRC() { return fileSRC; }
  string getREF() { return fileREF; }
  string getECM() { return fileECM; }

  // Do data points have individual names?
  bool hasPointNames() { return hasNamesSav; }
  string getPointName(int i) {
    return (i < int(pointNameSav.size()) && i >= 0) ? pointNameSav[i] : "";}

  // Vector containing column formats and number of systematic-error columns
  vector<int> format;
  int nSysSav;

  // Generic data storage members
  vector<double> xLsav, xCsav, xRsav, ySav,
    uTotHiSav, uTotLoSav, uStatHiSav, uStatLoSav,
    uSysLoSav, uSysHiSav, uSys2LoSav, uSys2HiSav;

  // Storage for integral
  double sumWeights;

  // Storage of SRC (name), REF (paper/analysis reference), and ECM
  string fileSRC, fileREF, fileECM;

  // Storage of names for points
  bool hasNamesSav;
  vector<string> pointNameSav;

  // Verbose output
  bool verboseSav;

};

} // End namespace Pythia8

#endif
