// VRgraph.h is a part of the LiveDisplays plug-in to the PYTHIA 8 generator.
// Copyright (C) 2019 P. Skands.
// LiveDisplays is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users,
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for VRgraph and derived classes.
// They are based on, and inherit from, the ROOT TGraphAsymErrors class.
// Histograms are based on ROOT's TH1D class.
//   VRgraph     : main graph class, based on a pointer to a TH1D
//   VRgraphData : graphs representing experimental data, from a file
//   VRgraphHis  : graphs representing saved histograms, from a file
//   VRgraphNorm : one graph normalized by another one,
//                 e.g., for ratio of theory to data
//   VRgraphVar  : for vector of graph variations,
//                 e.g., relative uncertainties

#ifndef LIVEDISPLAYSGRAPH_H
#define LIVEDISPLAYSGRAPH_H

// Include ROOT headers required by VRgraph
#include "TColor.h"
#include "TGraphAsymmErrors.h"
#include "TH1.h"
#include "TROOT.h"

// Include additional LiveDisplays headers needed by VRgraph
#include "LiveDisplays/VRfile.h"

// Put all the rest in Vincia namespace

namespace Pythia8 {

//=====================================================================
// class VRgraph. Base class for LiveDisplays graphs

class VRgraph : public TGraphAsymmErrors {

 public:

  // Constructors
  VRgraph() {
    hasOuterSav     = false;
    rebin           = false;
    zeroSkippingSav = false;
    TINY = 1e-6;
  }
  VRgraph(TH1D* initHis) {
    hisPtrSav.push_back(initHis);
    rebin           = false;
    zeroSkippingSav = false;
    hasOuterSav     = false;
    TINY = 1e-6;
  }
  VRgraph(vector<TH1D*> initHis) {
    hisPtrSav       = initHis;
    rebin           = false;
    zeroSkippingSav = false;
    hasOuterSav = (hisPtrSav.size() >= 2) ? true : false;
    TINY = 1e-6;
  }

  // Force rebinning according to some other histogram
  void setRebinAxis(VRgraph* xGraphPtr, bool zeroSkippingIn=false) {
    rebin        = true;
    rebinAxisPtr = xGraphPtr;
    zeroSkippingSav = zeroSkippingIn;
  }

  // zeroSkipping
  void zeroSkipping(bool zeroSkippingIn=true) {
    zeroSkippingSav = zeroSkippingIn;
  }

  // Outer band
  bool hasOuter() {return hasOuterSav;}
  int nOuter() { return hisPtrSav.size() -1; }
  TGraphAsymmErrors* getOuter() {return &outer;}
  virtual void DrawOuter(string drawOptions="") {
    if (hasOuter()) outer.Draw(drawOptions.c_str());
  }

  // Update graph
  virtual bool update(bool normalizeIn=true, double normFacIn=1.0,
    double scatterX = 0.0);

  // Set color scheme (copy from another graph)
  virtual void setColorScheme(VRgraph* graph);

  // Set colors according to predefined (enumerated) color schemes
  // (definitions corresponding to each scheme defined in .cc file)
  // (0 is special, used for data, > 0 for theory curves)
  // Returns false if the specified colour scheme does not exist
  virtual bool setColorScheme(int iScheme=0);

  // Get Pointer to user histogram (or variation, if exists)
  TH1D* getTH1DPtr (int i=0) {
    if ( i <= int(hisPtrSav.size())-1 ) return hisPtrSav[i];
    else return NULL;
  }

 protected:

  // Pointer to user histogram (vector if contains uncertainty variations)
  vector<TH1D*> hisPtrSav;
  // Container for outer uncertainty band.
  TGraphAsymmErrors outer;

  //
  double sumWeights;
  bool hasOuterSav;

  //
  bool normalizeSav;
  double normFacSav;

  //
  bool rebin;
  bool zeroSkippingSav;
  VRgraph* rebinAxisPtr;

  // Value determining whether bins are compatible
  // apparently, initializing this value where it is declared is not
  // standard C++03 (but standard C++11)
  double TINY;

};

//=====================================================================
// class VRgraphData
// Class for containing graphs representing experimental data points

class VRgraphData : public VRgraph {

 public:

  // Constructors
  VRgraphData() {
    dataFilePtr = 0;
    hasOuterSav = true;
    rebin       = false;
    firstUpdate = true;
  }
  VRgraphData(VRfile* dataFilePtrIn) {
    hasOuterSav = true;
    setDataFilePtr(dataFilePtrIn);
    rebin       = false;
    firstUpdate = true;
  }

  // Set data file Ptr
  void setDataFilePtr(VRfile* dataFilePtrIn) {
    dataFilePtr = dataFilePtrIn;
    setColorScheme(0);
    rebin       = false;
    firstUpdate = true;
    // For files containing data from multiple experiments, allow to use
    // different symbols for each experiment. Set marker of main histogram
    // to be simple dot.
    if (dataFilePtr != 0 && dataFilePtr->hasPointNames()) SetMarkerStyle(1);
  }

  // Update graph
  virtual bool update(bool normalizeIn=true, double normFacIn=1.0,
    double sigmaTot = 1.0);

  // Auxiliary containers used for files containing multiple data sets
  map<string, VRgraph> auxGraphs;

 protected:

  // Pointer to data histogram
  VRfile* dataFilePtr;

  // Variable to set false after first update.
  bool firstUpdate;

};

//=====================================================================
// class VRgraphHis
// Class for containing graphs representing saved histograms

class VRgraphHis : public VRgraph {

 public:

  // Constructors
  VRgraphHis() {
    hasOuterSav     = false;
    rebin           = false;
    zeroSkippingSav = false;
  }
  VRgraphHis(VRfile* hisFilePtrIn) {
    setHisFilePtr(hisFilePtrIn);
    rebin           = false;
    zeroSkippingSav = false;
  }

  // Set data file Ptr
  void setHisFilePtr(VRfile* hisFilePtrIn) {
    hisFilePtr = hisFilePtrIn;
  }

  // Update graph
  virtual bool update(bool normalizeIn=true, double normFacIn=1.0,
    double = 0.0);

 protected:

  // Pointer to user histogram
  VRfile* hisFilePtr;

  // Format specifier
  vector<int> format;

};

//=====================================================================
// class VRgraphNorm
// Class for ratios of histograms (e.g., normalized to data)

class VRgraphNorm : public VRgraph {

 public:

  // Constructors
  VRgraphNorm() {
    hasOuterSav = false;
    chi2sav       = 0.0;
    chi2errSav    = 1.0;
    rebin       = false;
  }
  VRgraphNorm(VRgraph* numeratorIn, VRgraph* denominatorIn) {
    hasOuterSav = false;
    setNumerator(numeratorIn);
    setDenominator(denominatorIn);
    chi2sav       = 0.0;
    chi2errSav    = 1.0;
    rebin       = false;
  }

  // Functions to set
  void setNumerator(VRgraph* graphIn) {
    numeratorPtr = graphIn;
    if (numeratorPtr) {
      hasNumerator = true;
      hasOuterSav  = numeratorPtr->hasOuter();
    }
    chi2sav       = 0.0;
    chi2errSav    = 1.0;
  }
  void setDenominator(VRgraph* graphIn) {
    denominatorPtr = graphIn;
    if (denominatorPtr) hasDenominator = true;
    chi2sav       = 0.0;
    chi2errSav    = 1.0;
  }

  // Get chi2 value (incl 5% uncertainty) and MC stat uncertainty
  double getChi2() { return chi2sav; }
  double getChi2err() { return chi2errSav; }

  // Update graph
  virtual bool update(bool = true,
    double ratioMin = 0.5, double ratioMax = 1.5);

 protected:

  VRgraph* numeratorPtr;
  VRgraph* denominatorPtr;
  bool hasNumerator;
  bool hasDenominator;
  double chi2sav, chi2errSav;


};

//=====================================================================
// class VRgraphVar
// Class for variations of histograms (e.g., relative uncertainties)

class VRgraphVar : public VRgraph {

 public:

  // Constructors
  VRgraphVar() {
    hasOuterSav = false;
    rebin       = false;
  }
  VRgraphVar(VRgraph* referenceIn,int iVarIn =1) {
    hasOuterSav = referenceIn->hasOuter();
    iVarSav     = min(referenceIn->nOuter(),max(0,iVarIn));
    setReference(referenceIn);
    rebin       = false;
  }

  // Functions to set
  void setReference(VRgraph* graphIn) {
    referencePtr = graphIn;
    if (referencePtr) {
      hasReference = true;
      if (referencePtr->hasOuter()) hasOuterSav = true;
    }
  }

  // Update graph
  virtual bool update(bool = true, double = 0.0, double = 0.0);

  // Set line and fill style for each variation,
  // according to predefined scheme (see .cc file)
  void setStyle(int iVar=1);

 protected:

  VRgraph* referencePtr;
  bool hasReference;

 private:
  int iVarSav;

};

} // End namespace Pythia8

#endif
