// Vincia.h is a part of the PYTHIA event generator.
// Copyright (C) 2019 Torbjorn Sjostrand.
// PYTHIA is licenced under the GNU GPL v2 or later, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// This file contains global header information for Vincia.

#ifndef VINCIA_H
#define VINCIA_H

// Maths headers
#include <limits>
#include <cmath>

// Include Pythia 8 headers
#include "Pythia8/Pythia.h"
#include "Pythia8/Event.h"           // Event
#include "Pythia8/ParticleData.h"    // ParticleData
#include "Pythia8/PartonSystems.h"   // PartonSystems
#include "Pythia8/StandardModel.h"   // AlphaStrong

// Include Vincia headers
#include "Pythia8/VinciaAntennaFunctions.h"
#include "Pythia8/VinciaCommon.h"
#include "Pythia8/VinciaFSR.h"
#include "Pythia8/VinciaISR.h"
#include "Pythia8/VinciaQED.h"
#include "Pythia8/VinciaMG5cxxInterface.h"

// Define namespace inside which Vincia lives
namespace Pythia8 {

//==========================================================================

// The VinciaPlugin class

class VinciaPlugin {

 public:

  VinciaPlugin() {
    cerr << "\n (VinciaPlugin():) cannot construct VinciaPlugin :(\n\n";
  }

  // Constructor from Pythia
  VinciaPlugin(Couplings* couplingsPtrIn, Info* infoPtrIn,
    ParticleData* particleDataPtrIn, PartonSystems* partonSystemsPtrIn,
    Rndm* rndmPtrIn, Settings* settingsPtrIn, SusyLesHouches* slhaPtrIn,
    UserHooks* userHooksPtrIn) {

    isConstructed = false;

    // First set pointers to Pythia classes
    pythiaPtr         = 0;
    couplingsPtr      = couplingsPtrIn;
    infoPtr           = infoPtrIn;
    particleDataPtr   = particleDataPtrIn;
    partonSystemsPtr  = partonSystemsPtrIn;
    rndmPtr           = rndmPtrIn;
    settingsPtr       = settingsPtrIn;
    slhaPtr           = slhaPtrIn;
    userHooksPtr      = userHooksPtrIn;

    // Init FSR shower pointers and default settings
    // ( beyond those set by the non-virtual TimeShower::initPtr() )
    fsrShower.initVinciaPtrs(&colour,&isrShower,&qedShower,&mecs,&resolution,
      &vinCom,&vinWeights);

    // Init ISR shower pointers and default settings
    // ( beyond those set by the non-virtual SpaceShower::initPtr() )
    isrShower.initVinciaPtrs(&colour,&fsrShower,&qedShower,&mecs,&resolution,
      &vinCom,&vinWeights);

  }

  // Constructor as plugin (given Pythia object,
  // using C++11 delegated constructor to call the Pythia-internal one first
  VinciaPlugin(Pythia* pythia,
    string xmlDir="../share/Vincia/xmldoc",
    string xmlIndexFile="VinciaIndex.xml") :
    VinciaPlugin(&pythia->couplings,&pythia->info,&pythia->particleData,
      &pythia->partonSystems,&pythia->rndm, &pythia->settings,
      &pythia->slhaInterface.slha, 0) {

    // Set Pythia 8 shower pointers to Vincia
    pythiaPtr = pythia;
    pythia->setShowerPtr(&fsrShower, &fsrShower, &isrShower);

    // Find path to PYTHIA data files.
    const char* PYTHIA8DATA = "PYTHIA8DATA";
    char* envPath = getenv(PYTHIA8DATA);
    if (envPath == 0 || *envPath == '\0') {
      cout<<" VINCIA Error: you have not set the PYTHIA8DATA environment variable.\n";
      cout<<" VINCIA Hint: if your local PYTHIA install is in the directory $HOME/yourpythia, try setting PYTHIA8DATA to $HOME/yourpythia/share/Pythia8/xmldoc"<<endl;
      cout<<"              alternatively, for a default global PYTHIA install, use /usr/share/Pythia8/xmldoc"<<endl;
      return;
    }

    // Find path to VINCIA data files, i.e. xmldoc directory location.
    // Environment variable takes precedence, else use constructor input.
    string xmlPath        = "";
    bool existsVINCIADATA = false;
    const char* VINCIADATA = "VINCIADATA";
    envPath = getenv(VINCIADATA);
    if (envPath != 0 && *envPath != '\0') {
      int i = 0;
      while (*(envPath+i) != '\0') xmlPath += *(envPath+(i++));
      existsVINCIADATA = true;
    }
    else xmlPath = xmlDir;
    if (xmlPath[ xmlPath.length() - 1 ] != '/') xmlPath += "/";
    // Set the VINCIADATA environment variable, for use in main programs
    setenv(VINCIADATA,xmlPath.c_str(),0);
    // Save the presumed location of VINCIA runtime data files
    VINCIA_SHARE = xmlPath+"../";
    // Add the VINCIA list of parameters to the PYTHIA 8 settings database
    settingsPtr->init(xmlPath+xmlIndexFile, true);

    // Verbose level
    setVerbose(settingsPtr->mode("Vincia:verbose"));

    // Check for DATA directory
    string msg="";
    if (verbose >= normal) {
      if (existsVINCIADATA){
	msg="VINCIADATA environment variable ok.";
      }
      else{
	msg="VINCIADATA environment variable undefined. Using constructor instead.";
      }
      printOut(__METHOD_NAME__,msg);
    }
    if (verbose >= quiet) {
      msg="using configuration directory : "+xmlPath;
      printOut(__METHOD_NAME__,msg);
    }

    isConstructed = true;

  }

  // Initialize Vincia & Pythia
  bool init(string commandFile="none", int subrun=-999) {

    if(!isConstructed){
      printErr("VinciaPlugin::init()",
	       "Construction of plugin failed.");
      return false;
    }

    // Get verbose level
    setVerbose(settingsPtr->mode("Vincia:verbose"));

    // Verbose output
    string msg="";
    if (verbose >= veryloud){
      msg="begin --------------";
      printOut(__METHOD_NAME__,msg);
    }

    // Vincia on/off
    vinciaOn    = settingsPtr->flag("Vincia");

    // Initialise VINCIA settings
    if (vinciaOn) {

      // Get list of Pythia readString commands given to Settings so far
      vector<string> readStringHistory = settingsPtr->getReadHistory();

      // Reset tune parameters using default (VINCIA-specific) tune
      string defaultTune = VINCIA_SHARE + "/tunes/default.cmnd";
      // Check if this file exists, otherwise look in VINCIA_SHARE
      if (fileExists(defaultTune)) {
	if (verbose >= quiteloud){
	  msg="reading default TUNE preset from "+defaultTune;
	  printOut(__METHOD_NAME__,msg);
	}
        pythiaPtr->readFile(defaultTune);
      }

      // Then reapply readString commands to enable user modifications
      // (but ignore any Tune:ee or Tune:pp commands)
      if (verbose >= quiteloud){
	msg="reapplying user settings";
	printOut(__METHOD_NAME__,msg);
      }
      for (int i=0; i<(int)readStringHistory.size(); ++i) {
        string lineNow = readStringHistory[i];
	string lineLower = toLower(lineNow);
        if (lineLower.find("tune:ee") == string::npos && lineLower.find("tune:pp")
	    == string::npos){
	  pythiaPtr->readString(lineNow);
	}
      }

      // Read any command file given to init
      if (commandFile != "none") {
	if (verbose >= normal){
	  msg="processing "+commandFile;
	  printOut(__METHOD_NAME__,msg);
	}
        if (subrun >= 0) pythiaPtr->readFile(commandFile,subrun);
        else pythiaPtr->readFile(commandFile);
      }

      // Check if verbose setting changed
      verbose = settingsPtr->mode("Vincia:verbose");

      if (verbose >= quiteloud){
        msg="setting pointers...";
	printOut(__METHOD_NAME__,msg);
      }

      // FSR and ISR antenna sets
      antennaSetFSR.initPtr(particleDataPtr, settingsPtr, rndmPtr, &dglap);
      antennaSetISR.initPtr(particleDataPtr, settingsPtr, rndmPtr, &dglap);

      // QED Shower module
      qedShower.initPtr(infoPtr, particleDataPtr, partonSystemsPtr,
        rndmPtr, settingsPtr, &vinCom);

      // Hand antenna set pointers to shower and matching objects
      fsrShower.initAntPtr(&antennaSetFSR);
      isrShower.initAntPtr(&antennaSetISR);
      mecs.initAntPtr(&antennaSetFSR,&antennaSetISR);
    }

    // If VINCIA off, let us know about it
    else {

      // Set Pythia 8 shower pointers back to Pythia internal ones
      pythiaPtr->setShowerPtr(0, 0, 0);

      if (verbose >= normal){
	msg="Vincia is switched off";
	printOut(__METHOD_NAME__,msg);
      }
    }

    // Pass pointers on to objects that require them
    resolution.initPtr(settingsPtr);
    rambo.initPtr(rndmPtr);
    vinCom.initPtr(particleDataPtr,settingsPtr,rndmPtr);
    mg5cxxInterface.initPtr(particleDataPtr,settingsPtr,couplingsPtr,
      rndmPtr,slhaPtr,&vinCom);
    mecs.initPtr(settingsPtr, infoPtr, rndmPtr,
      partonSystemsPtr, &mg5cxxInterface, &vinCom);
    colour.initPtr(particleDataPtr,settingsPtr,partonSystemsPtr,rndmPtr);
    vinWeights.initPtr(settingsPtr, infoPtr, &vinCom);

    // Initialise classes not initialised by shower::init() methods.
    vinCom.init();
    resolution.init();
    colour.init();
    vinWeights.init();

    // Initialise Pythia
    // (Note this also calls the ISR and FSR init() methods since those inherit
    // from the base Pythia shower classes)
    if(!pythiaPtr->init()){
      printErr("VinciaPlugin::init()",
	       "Vincia failed to initiate.");
      return false;
    }

    // MECs depend on Pythia Couplings so must be initialised after Pythia
    mecs.init();

    // Print VINCIA header and list of parameters
    if (verbose >= 1 && vinciaOn) {
      fsrShower.header();
    }

    // Verbose output
    if(verbose >= veryloud){
      msg="end --------------";
      printOut(__METHOD_NAME__,msg);
    }

    return true;

  }

  // Methods to return shower pointers
  TimeShower* getTimesShower() { return &fsrShower; }
  TimeShower* getTimesDecShower() { return &fsrShower; }
  SpaceShower* getSpaceShower() { return &isrShower; }

  //*********

  // Automatically set verbose level in all inheriting members
  void setVerbose(int verboseIn) {
    verbose = verboseIn;
    vinCom.setVerbose(verboseIn);
    resolution.setVerbose(verboseIn);
    fsrShower.setVerbose(verboseIn);
    qedShower.setVerbose(verboseIn);
    isrShower.setVerbose(verboseIn);
    colour.setVerbose(verboseIn);
    mg5cxxInterface.setVerbose(verboseIn);
    mecs.setVerbose(verboseIn);
  }


  // Shortcut for post-Vincia event weight
  double weight(int iWeight=0) {
    // If Vincia is on, return Vincia weight
    if (settingsPtr->flag("Vincia"))
      return vinWeights.weight(iWeight);
    // Else just wrap Pythia's weights
    return infoPtr->weight(iWeight);
  }
  string weightLabel(int iWeight=0) {
    // If Vincia is on, use Vincia's weight labels
    if (settingsPtr->flag("Vincia"))
      return vinWeights.weightLabel(iWeight);
    // Else just wrap Pythia's weight counter
    return infoPtr->weightLabel(iWeight);
  }
  int  nWeights() {
    // If Vincia is on, use Vincia's weight counter
    if (settingsPtr->flag("Vincia"))
      return vinWeights.nWeights();
    // Else just wrap Pythia's weight counter
    return infoPtr->nWeights();
  }
  bool reweightingOccurred() { return vinWeights.reweightingOccurred(); }

  // Utilities for printing Info and Internal histograms
  void printInfo() {
    fsrShower.printInfo(true);
    isrShower.printInfo(true);
  }
  void printHistos() { fsrShower.printHistos(); }
  void writeHistos( string fileName = "vincia", string lastName = "dat" ) {
    fsrShower.writeHistos(fileName,lastName);
  }
  const Hist& getDiagnosticHistogram( string name ) {
    return fsrShower.getDiagnosticHistogram(name);
  }

  // Main VINCIA on/off switch
  bool vinciaOn;

  // Instantiate Vincia Objects
  VinciaCommon        vinCom;
  Resolution          resolution;
  VinciaFSR           fsrShower;
  VinciaISR           isrShower;
  QEDShower           qedShower;
  Colour              colour;
  ResScaleHook        resScaleHook;
  VinciaWeights       vinWeights;
  MECs                mecs;

  // Auxiliary objects
  VinciaMG5cxxInterface mg5cxxInterface;
  Rambo                 rambo;

  // Vectors of antenna functions
  DGLAP         dglap;
  AntennaSetFSR antennaSetFSR;
  AntennaSetISR antennaSetISR;

  // Location of runtime VINCIA files
  string VINCIA_SHARE;

  // Pointers to Pythia classes
  Pythia*        pythiaPtr;
  Couplings*     couplingsPtr;
  Info*          infoPtr;
  ParticleData*  particleDataPtr;
  PartonSystems* partonSystemsPtr;
  Rndm*          rndmPtr;
  Settings*      settingsPtr;
  SusyLesHouches* slhaPtr;
  UserHooks*     userHooksPtr;

 private :

  // Verbose level
  int verbose;
  bool isConstructed;

}; // end Vincia plugin

} // end Vincia namespace
#endif
