// VinciaAntennaFunctions.h is a part of the PYTHIA event generator.
// Copyright (C) 2019 Torbjorn Sjostrand.
// PYTHIA is licenced under the GNU GPL v2 or later, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// This file contains header information for the AntennaFunction base class,
// its derived classes for FF, IF, and II antenna functions, the
// AntennaSetFSR and AntennaSetISR classes, and the MEC class for matrix-
// element corrections.

#ifndef VINCIAANTENNAFUNCTIONS_H
#define VINCIAANTENNAFUNCTIONS_H

#include "Pythia8/Basics.h"
#include "Pythia8/Event.h"
#include "Pythia8/PythiaStdlib.h"

#include "Pythia8/VinciaCommon.h"
#include "Pythia8/VinciaMG5cxxInterface.h"

namespace Pythia8 {

//==========================================================================

// A class containing DGLAP splitting functions for limit checking

class DGLAP {

 public:

  // Constructor
  DGLAP() {}

  // Helicity-dependent Altarelli-Parisi kernels (mu = m/Q)
  // Note: Pg2gg is written with standard factor 2 normalization convention
  double Pg2gg(double z, int hA=9, int hB=9, int hC=9);
  double Pg2qq(double z, int hA=9, int hB=9, int hC=9, double mu=0.);
  double Pq2qg(double z, int hA=9, int hB=9, int hC=9, double mu=0.);
  double Pq2gq(double z, int hA=9, int hB=9, int hC=9, double mu=0.);

  // Wrappers to get unpolarized massive Altarelli-Pariis kernels (mu = m/Q)
  double Pg2qq(double z, double mu) { return Pg2qq(z,9,9,9,mu); }
  double Pq2qg(double z, double mu) { return Pq2qg(z,9,9,9,mu); }
  double Pq2gq(double z, double mu) { return Pq2gq(z,9,9,9,mu); }

  // Altarelli-Parisi kernels with linear in/out polarizations for gluons
  //   pol = +1 for in-plane, -1 for out-of-plane
  double Pg2ggLin(double z, int polA=9, int polB=9, int polC=9);
  double Pg2qqLin(double z, int polA=9, int hB=9, int hC=9, double mu=0.);
  double Pq2qgLin(double z, int hA=9, int hB=9, int polC=9,double mu=0.);
  double Pq2gqLin(double z, int hA=9, int polB=9, int hC=9,double mu=0.);

};

//==========================================================================

// The AntennaFunction base class:
// Base class implementation for all AntennaFunction objects

class AntennaFunction {

 public:

  // Constructor
  AntennaFunction() {
    // Init flags
    isInitPtr = false;
    isInit    = false;
    // Default values (reset by init)
    verbose              = 1;
    // Initialize LH and RH maps (DO NOT CHANGE)
    RH[9]  = true;
    RH[1]  = true;
    RH[-1] = false;
    LH[9]  = true;
    LH[1]  = false;
    LH[-1] = true;
    chargeFacSav   = 0.0;
    kineMapSav     = 0;
    modeSLC        = -1;
    // Sector shower parameters
    sectorShower   = false;
    sectorDampSav  = 0.0;       // yGGsym = 1 - yxg - ygg + sectorDamp*ygg
    // Global shower parameters
    alphaSav       = 0.0;
    //
    preFacFiniteTermSav = 0.0;
    antMinSav           = 0.0;
    // Initialise dummy vectors for massless / unpolarised antennae
    hDum.clear();
    mDum.clear();
    for (int i=0; i<4; ++i) {
      hDum.push_back(9);
      mDum.push_back(0.);
    }
    hA = 9;
    hB = 9;
    hi = 9;
    hj = 9;
    hk = 9;
    mi = 0.;
    mj = 0.;
    mk = 0.;
  };

  // Destructor
  virtual ~AntennaFunction() {};

  //----------------------------------------------------------------------
  // List of VIRTUAL functions to be redefined by inheriting classes

  // Names of this antenna, for VINCIA, and for humans
  virtual string vinciaName()   const {return "Vincia:AntennaFunction";};

  // Parton types (AB -> 0i 1j 2k): needed by soft- and collinear-limit checks
  virtual int idA()    const {return 0;}
  virtual int idB()    const {return 0;}
  virtual int id1()    const {return 0;}

  // The (dimensionful) antenna function
  virtual double antFun(vector<double> invariants, vector<double> mNew,
    vector<int> helBef, vector<int> helNew) = 0;

  // Optional implementation of the DGLAP kernels for collinear-limit checks
  // Defined as PI/sij + PK/sjk, i.e., equivalent to antennae.
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> mNew, vector<int> helBef, vector<int> helNew) = 0;

  //----------------------------------------------------------------------
  // List of VIRTUAL functions that MAY be redefined (if needed)

  // Default initialization
  virtual bool init();

  // Construct baseName from idA, idB, and id1
  virtual string baseName()    const {
    string name = id2str(id1()) + "/" + id2str(idA()) + id2str(idB());
    return name;
  }

  // Wrapper that can modify baseName to more human readable form if required.
  virtual string humanName()   const {
    return baseName();
  }

  // Method to intialise mass values
  virtual void initMasses(vector<double>* masses) {
    if (masses->size() >= 3) {
      mi = masses->at(0);
      mj = masses->at(1);
      mk = masses->at(2);
    } else {
      mi = 0.0;
      mj = 0.0;
      mk = 0.0;
    }
  }

  // Method to initialise internal helicity variables
  // Return value = number of helicity configurations to average over.
  virtual int initHel(vector<int>* helBef, vector<int>* helNew) {

    // Initialise as unpolarised
    hA = 9;
    hB = 9;
    hi = 9;
    hj = 9;
    hk = 9;

    // Check if one or more partons have explicit helicities
    if (helNew->size() >= 3) {
      hi = helNew->at(0);
      hj = helNew->at(1);
      hk = helNew->at(2);
    }
    if (helBef->size() >= 2) {
      hA = helBef->at(0);
      hB = helBef->at(1);
    }

    // Check if helicity assignments make sense
    bool physHel = true;
    if ( hA != 1 && hA != -1 && hA != 9) physHel = false;
    if ( hB != 1 && hB != -1 && hB != 9) physHel = false;
    if ( hi != 1 && hi != -1 && hi != 9) physHel = false;
    if ( hj != 1 && hj != -1 && hj != 9) physHel = false;
    if ( hk != 1 && hk != -1 && hk != 9) physHel = false;
    if (!physHel) {
      if (verbose >= normal){
        stringstream ss;
        ss<<"Unphysical helicity configuration: "
          << hA <<" "<< hB << " -> "<< hi <<" "<< hj <<" "<< hk;
        printErr(__METHOD_NAME__,ss.str());
      }
      return 0;
    }

    // How many configurations are we averaging over?
    int nAvg = 1;
    if (hA == 9) nAvg *= 2;
    if (hB == 9) nAvg *= 2;
    return nAvg;
  }

  // Wrapper for helicity-summed/averaged antenna function
  double antFun(vector<double> invariants, vector<double> masses) {
    return antFun(invariants, masses, hDum, hDum);
  }

  // Wrapper for massless, helicity-summed/averaged antenna function
  double antFun(vector<double> invariants) {
    return antFun(invariants, mDum, hDum, hDum);
  }

  // Wrapper without helicity assignments
  double AltarelliParisi(vector<double> invariants,
    vector<double> masses) {
    return AltarelliParisi(invariants, masses, hDum, hDum);
  }

  // Wrapper for massless helicity-summed/averaged DGLAP kernels
  double AltarelliParisi(vector<double> invariants) {
    return AltarelliParisi(invariants, mDum, hDum, hDum);
  }

  // Function to check singularities, positivity, etc.
  virtual bool check();

  //----------------------------------------------------------------------
  // List of NON-VIRTUAL functions (i.e., that may not be redefined)

  // Initialize pointers:
  //    ParticleData : required for masses, colour charges, etc
  //    Rndm : required to generate random test points to check antenna
  //    Settings : required for all coefficients, switches, etc.
  void initPtr(ParticleData* particleDataPtrIn, Settings* settingsPtrIn,
    Rndm* rndmPtrIn, DGLAP* dglapPtrIn) {
    particleDataPtr = particleDataPtrIn;
    settingsPtr     = settingsPtrIn;
    rndmPtr         = rndmPtrIn;
    dglapPtr        = dglapPtrIn;
    isInitPtr       = true;
    isInit          = false;
  };

  // -------
  // Functions to GET

  // Get parameters
  double chargeFac() { return chargeFacSav; }
  int    kineMap() { return kineMapSav; }
  double alpha() { return alphaSav; }
  double sectorDamp() { return sectorDampSav; }

  // Functions to get Altarelli-Parisi energy fractions from invariants
  double zA(vector<double> invariants) {
    double yij = invariants[1]/invariants[0];
    double yjk = invariants[2]/invariants[0];
    return (1.-yjk)/(1.+yij);
  }
  double zB(vector<double> invariants) {
    double yij = invariants[1]/invariants[0];
    double yjk = invariants[2]/invariants[0];
    return (1.-yij)/(1.+yjk);
  }

  //----------------------------------------------------------------------
  // id2str()
  // Auxiliary function to translate an ID code to a string,
  // for constructing antenna names

  string id2str(int id) const {
    // Vector bosons
    if (id == 21) return "g";
    else if (id == 22) return "gamma";
    else if (id == 23) return "Z";
    else if (abs(id) == 24) return "W";
    // Fermions
    else if (id >= 1 && id <= 4) return "q";
    else if (-id >= 1 && -id <= 4) return "qbar";
    else if (id == 5) return "b";
    else if (id == -5) return "bbar";
    else if (id == 6) return "t";
    else if (id == -6) return "tbar";
    else if (id >= 11 && id <= 20 && id%2 == 1) return "l-";
    else if (id >= 11 && id <= 20 && id%2 == 0) return "nu";
    else if (-id >= 11 && -id <= 20 && id%2 == 1) return "l+";
    else if (-id >= 11 && -id <= 20 && id%2 == 0) return "nubar";
    // Octet fermion: use gluino
    else if (id == 1000021) return "~g";
    // Charged scalar: use H+
    else if (id == 37) return "H+";
    else if (id == -37) return "H-";
    // Coloured scalars: use squarks
    else if (id >= 1000000 && id <= 1000010) return "~q";
    else if (-id >= 1000000 && -id <= 1000010) return "~q*";
    // Unknown particle
    else return "X";
  }

  //----------------------------------------------------------------------
  // List of PROTECTED data members
  // (Not private, so can be accessed by inheriting classes)

 protected:

  // Is initialized
  bool isInitPtr, isInit;

  // Charge factor, kinematics map, and subleading-colour treatment
  double chargeFacSav;
  int kineMapSav, modeSLC;
  bool sectorShower;

  // The alpha collinear-partitioning parameter
  double alphaSav;

  // The sector-shower collinear dampening parameter
  double sectorDampSav;

  // Shorthand for commonly used variable(s)
  double term;

  //
  double preFacFiniteTermSav, antMinSav;
  bool   isMinVar;

  // Variables for internal storage of masses and helicities
  double mi, mj, mk;
  int hA, hB, hi, hj, hk;

  // Map to tell whether a given helicity value maps to L- and/or R-handed
  // Defined by constructor and not to be changed dynamically.
  map<int,bool> LH;
  map<int,bool> RH;

  // Verbose level
  int verbose;

  // Pointers to Pythia-8 classes
  ParticleData* particleDataPtr;
  Settings*     settingsPtr;
  Rndm*         rndmPtr;

  // Pointer to VINCIA DGLAP class
  DGLAP*        dglapPtr;

  // Dummy vectors
  vector<double> mDum;
  vector<int> hDum;

};

//==========================================================================

// Derived classes for Final-Final AntennaFunctions

//==========================================================================

// class QQEmitFF

class QQEmitFF : public AntennaFunction {

 public:

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const {return "Vincia:QQEmitFF";};

  // Functions needed by soft- and collinear-limit checks (AB -> 0i 1j 2k)
  virtual int idA()    const {return 1;}
  virtual int idB()    const {return -1;}
  virtual int id1()    const {return 21;}

  // The (dimensionless) antenna function
  virtual double antFun(vector<double> invariants, vector<double> mNew,
    vector<int> helBef, vector<int> helNew);

  // Function to give Altarelli-Parisi limits of this antenna
  // Defined as PI/yij + PK/yjk, i.e., equivalent to antennae.
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> /* mNew */, vector<int> helBef, vector<int> helNew) {
    int h0Now = helNew[0];
    int h1Now = helNew[1];
    int h2Now = helNew[2];
    int hANow = helBef[0];
    int hBNow = helBef[1];
    if (hANow != h0Now || hBNow != h2Now) return -1.;
    else return dglapPtr->Pq2qg(zA(invariants),hANow,h0Now,h1Now)/invariants[1]
           + dglapPtr->Pq2qg(zB(invariants),hBNow,h2Now,h1Now)/invariants[2];
  }

};

//==========================================================================

// class QGEmitFF

class QGEmitFF : public AntennaFunction {

 public:

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const {return "Vincia:QGEmitFF";};

  // Parton types (AB -> 0i 1j 2k): needed by soft- and collinear-limit checks
  virtual int idA()    const {return 1;}
  virtual int idB()    const {return 21;}
  virtual int id1()    const {return 21;}

  // The (dimensionless) antenna function
  virtual double antFun(vector<double> invariants,
    vector<double> mNew, vector<int> helBef, vector<int> helNew);

  // Function to give Altarelli-Parisi limits of this antenna
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> /* mNew */, vector<int> helBef, vector<int> helNew) {
    int h0Now = helNew[0];
    int h1Now = helNew[1];
    int h2Now = helNew[2];
    int hANow = helBef[0];
    int hBNow = helBef[1];
    if (h0Now != hANow) return -1;
    // Compute (sum of) DGLAP kernel(s)
    double sum(0.);
    if (h2Now == hBNow){
      sum += dglapPtr->Pq2qg(zA(invariants),hANow,h0Now,h1Now)/invariants[1];
    }
    sum += dglapPtr->Pg2gg(zB(invariants),hBNow,h2Now,h1Now)/invariants[2];
    return sum;
  }

};

//==========================================================================

// class GQEmitFF

class GQEmitFF : public QGEmitFF {

 public:

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const {return "Vincia:GQEmitFF";};

  // Parton types (AB -> 0i 1j 2k): needed by soft- and collinear-limit checks
  virtual int idA()    const {return 21;}
  virtual int idB()    const {return -1;}
  virtual int id1()    const {return 21;}

  // The antenna function (derived from QGEmit by swapping)
  virtual double antFun(vector<double> invariants,
    vector<double> mNew, vector<int> helBef, vector<int> helNew) {
    vector<double> invariantsSwapped = invariants;
    invariantsSwapped[1] = invariants[2];
    invariantsSwapped[2] = invariants[1];
    vector<double> mSwapped = mNew;
    mSwapped[0] = mNew[2];
    mSwapped[2] = mNew[0];
    vector<int> helBefSwapped = helBef;
    helBefSwapped[0] = helBef[1];
    helBefSwapped[1] = helBef[0];
    vector<int> helNewSwapped = helNew;
    helNewSwapped[0] = helNew[2];
    helNewSwapped[2] = helNew[0];
    return QGEmitFF::antFun(invariantsSwapped, mSwapped,
      helBefSwapped, helNewSwapped);
  }

  // Function to give Altarelli-Parisi limits of this antenna
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> /* mNew */, vector<int> helBef, vector<int> helNew) {
    int h0Now = helNew[0];
    int h1Now = helNew[1];
    int h2Now = helNew[2];
    int hANow = helBef[0];
    int hBNow = helBef[1];
    if (h2Now != hBNow) return -1;
    // Compute (sum of) DGLAP kernel(s)
    double sum(0.);
    if (h0Now == hANow){
      sum += dglapPtr->Pq2qg(zB(invariants),hBNow,h2Now,h1Now)/invariants[2];
    }
    sum += dglapPtr->Pg2gg(zA(invariants),hANow,h0Now,h1Now)/invariants[1];
    return sum;
  }

};

//==========================================================================

// class GGEmitFF

class GGEmitFF : public AntennaFunction {

 public:

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const {return "Vincia:GGEmitFF";};

  // Parton types (AB -> 0i 1j 2k): needed by soft- and collinear-limit checks
  virtual int idA()    const {return 21;}
  virtual int idB()    const {return 21;}
  virtual int id1()    const {return 21;}

  // The (dimensionless) antenna function
  virtual double antFun(vector<double> invariants,
    vector<double> mNew, vector<int> helBef, vector<int> helNew);

  // Function to give Altarelli-Parisi limits of this antenna
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> /* mNew */, vector<int> helBef, vector<int> helNew) {
    int h0Now = helNew[0];
    int h1Now = helNew[1];
    int h2Now = helNew[2];
    int hANow = helBef[0];
    int hBNow = helBef[1];
    double sum(0.);
    if ( hBNow == h2Now ){
      sum += dglapPtr->Pg2gg(zA(invariants),hANow,h0Now,h1Now)/invariants[1];
    }
    if ( hANow == h0Now ){
      sum += dglapPtr->Pg2gg(zB(invariants),hBNow,h2Now,h1Now)/invariants[2];
    }
    return sum;
  }

};

//==========================================================================

// class GXSplitFF

class GXSplitFF : public AntennaFunction {

 public:

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const {return "Vincia:GXSplitFF";};

  // Parton types (AB -> 0i 1j 2k): needed by soft- and collinear-limit checks
  virtual int idA()    const {return 21;}
  virtual int idB()    const {return  0;}
  virtual int id1()    const {return -1;}

  // The (dimensionless) antenna function
  virtual double antFun(vector<double> invariants,
    vector<double> mNew, vector<int> helBef, vector<int> helNew);

  // Function to give Altarelli-Parisi limits of this antenna
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> /* mNew */, vector<int> helBef, vector<int> helNew) {
    int h0Now = helNew[0];
    int h1Now = helNew[1];
    int h2Now = helNew[2];
    int hANow = helBef[0];
    int hBNow = helBef[1];
    if (hBNow != h2Now) return 0.0;
    else
      return dglapPtr->Pg2qq(zA(invariants),hANow,h0Now,h1Now)/invariants[1];
  }

};

//==========================================================================

// Derived classes for Sector Final-Final antenna functions
// (Obtained by explicit symmetrisation of global ones)

//==========================================================================

// class QQEmitFFsec (identical to global one)

class QQEmitFFsec : public QQEmitFF {

 public:

  // The (dimensionless) antenna function (this one is same as global)
  virtual double antFun(vector<double> invariants,
    vector<double> mNew, vector<int> helBef, vector<int> helNew) {
    return QQEmitFF::antFun(invariants, mNew, helBef, helNew);
  }

};

//==========================================================================

// class QGEmitFFsec

class QGEmitFFsec : public QGEmitFF {

 public:

  // The (dimensionless) antenna function
  virtual double antFun(vector<double> invariants,
    vector<double> mNew, vector<int> helBef, vector<int> helNew) {
    double ant = QGEmitFF::antFun(invariants, mNew, helBef, helNew);
    // Check if helicity vectors empty
    if (helBef.size() < 2) {
      helBef.push_back(9);
      helBef.push_back(9);
    }
    if (helNew.size() < 3) {
      helNew.push_back(9);
      helNew.push_back(9);
      helNew.push_back(9);
    }
    // Check if j has same helicity as parent gluon.
    int hG = helBef[1];
    int hjNow = helNew[1];
    if ( hG == hjNow ) {
      // Define j<->k symmetrisation term
      vector<double> invariantsSym = invariants;
      double s02 = invariants[0] - invariants[1] - invariants[2];
      vector<int> helSym = helNew;
      invariantsSym[1] = s02 + sectorDampSav * invariants[2];
      helSym[1] = helNew[2];
      helSym[2] = helNew[1];
      ant += QGEmitFF::antFun(invariantsSym, mNew, helBef, helSym);
    }
    return ant;
  }

};

//==========================================================================

// class GQEmitFFsec

class GQEmitFFsec : public QGEmitFFsec {

 public:

  // Parton types (AB -> 0i 1j 2k): needed by soft- and collinear-limit checks
  virtual int idA()    const {return 21;}
  virtual int idB()    const {return -1;}
  virtual int id1()    const {return 21;}

  // The antenna function (derived from QGEmit by swapping)
  virtual double antFun(vector<double> invariants,
    vector<double> mNew, vector<int> helBef, vector<int> helNew) {
    vector<double> invariantsSwapped = invariants;
    invariantsSwapped[1] = invariants[2];
    invariantsSwapped[2] = invariants[1];
    vector<double> mSwapped = mNew;
    mSwapped[0] = mNew[2];
    mSwapped[2] = mNew[0];
    vector<int> helBefSwapped = helBef;
    helBefSwapped[0] = helBef[1];
    helBefSwapped[1] = helBef[0];
    vector<int> helNewSwapped = helNew;
    helNewSwapped[0] = helNew[2];
    helNewSwapped[2] = helNew[0];
    return QGEmitFFsec::antFun(invariantsSwapped, mSwapped,
      helBefSwapped, helNewSwapped);
  }

  // Function to give Altarelli-Parisi limits of this antenna
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> /* mNew */, vector<int> helBef, vector<int> helNew) {
    int h0Now = helNew[0];
    int h1Now = helNew[1];
    int h2Now = helNew[2];
    int hANow = helBef[0];
    int hBNow = helBef[1];
    if (h2Now != hBNow) return -1;
    // Compute (sum of) DGLAP kernel(s)
    double sum(0.);
    if (h0Now == hANow){
      sum += dglapPtr->Pq2qg(zB(invariants),hBNow,h2Now,h1Now)/invariants[2];
    }
    sum += dglapPtr->Pg2gg(zA(invariants),hANow,h0Now,h1Now)/invariants[1];
    return sum;
  }

};

//==========================================================================

// class GGEmitFFsec

class GGEmitFFsec : public GGEmitFF {

 public:

  // The (dimensionless) antenna function
  virtual double antFun(vector<double> invariants, vector<double> mNew,
    vector<int> helBef, vector<int> helNew) {
    double ant = GGEmitFF::antFun(invariants, mNew, helBef, helNew);
    // Check if helicity vectors empty
    if (helBef.size() < 2) {
      helBef.push_back(9);
      helBef.push_back(9);
    }
    if (helNew.size() < 3) {
      helNew.push_back(9);
      helNew.push_back(9);
      helNew.push_back(9);
    }
    // Check if j has same helicity as parent gluon 0.
    int hjNow = helNew[1];
    if ( helBef[0] == hjNow ) {
      // Define i<->j symmetrisation term
      vector<double> invariantsSym = invariants;
      double s02 = invariants[0] - invariants[1] - invariants[2];
      vector<int> helSym = helNew;
      helSym[0] = helNew[1];
      helSym[1] = helNew[0];
      invariantsSym[2] = s02 + sectorDampSav * invariants[1];
      ant += GGEmitFF::antFun(invariantsSym, mNew, helBef, helSym);
    }
    // Check if j has same helicity as parent gluon 1.
    if ( helBef[1] == hjNow ) {
      // Define j<->k symmetrisation term
      vector<double> invariantsSym = invariants;
      double s02 = invariants[0] - invariants[1] - invariants[2];
      vector<int> helSym = helNew;
      helSym[1] = helNew[2];
      helSym[2] = helNew[1];
      invariantsSym[1] = s02 + sectorDampSav * invariants[2];
      ant += GGEmitFF::antFun(invariantsSym, mNew, helBef, helSym);
    }
    return ant;
  }

};

//==========================================================================

// class GXSplitFFsec

class GXSplitFFsec : public GXSplitFF {

 public:

  // The (dimensionless) antenna function
  virtual double antFun(vector<double> invariants, vector<double> mNew,
    vector<int> helBef, vector<int> helNew) {
    // The sector antenna is just 2 * global.
    return 2*GXSplitFF::antFun(invariants,mNew,helBef,helNew);
  }

};

//==========================================================================

// Derived classes for Initial-Initial antenna functions
// (the same for global and sector cases since even the global ones
// include sector terms representing "emission into the initial state".

//==========================================================================

// class AntennaFunctionII

// Base class for initial-initial and initial-final antenna functions.
// Base class implements II.

class AntennaFunctionIX : public AntennaFunction {

 public:

  // Method to initialise (can be different than that of the base class)
  virtual bool init();

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:AntennaFunctionIX"; };

  // Parton types AB -> 0a 1j 2b with A,B,a,b initial and j final
  virtual int idA()    const { return 0; }
  virtual int idB()    const { return 0; }
  virtual int id0()    const { return 0; }
  virtual int id1()    const { return 0; }
  virtual int id2()    const { return 0; }

  // Functions to get Altarelli-Parisi energy fractions
  virtual double zA(vector<double> invariants) {
    double sAB = invariants[0];
    double sjb = invariants[2];
    return sAB/(sAB+sjb);
  }
  virtual double zB(vector<double> invariants) {
    double sAB = invariants[0];
    double saj = invariants[1];
    return sAB/(sAB+saj);
  }

  // Function to tell if this is an II antenna
  virtual bool isIIant() { return true; }

  // Function to check singularities, positivity, etc.
  virtual bool check();

};

//==========================================================================

// class QQEmitII

class QQEmitII : public AntennaFunctionIX {

 public:

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:QQemitII"; };

  // Parton types AB -> 0a 1j 2b with A,B,a,b initial and j final
  virtual int idA()    const { return 1; }
  virtual int idB()    const { return -1; }
  virtual int id0()    const { return 1; }
  virtual int id1()    const { return 21; }
  virtual int id2()    const { return -1; }

  // The antenna function
  virtual double antFun(vector<double> invariants, vector<double> masses,
    vector<int> helBef, vector<int> helNew);

  // AP splitting kernel for collinear limit checks
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> /* mNew */, vector<int> helBef, vector<int> helNew);

};

//==========================================================================

// class GQEmitII

class GQEmitII : public AntennaFunctionIX {

 public:

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:GQemitII"; };

  // Parton types AB -> 0a 1j 2b with A,B,a,b initial and j final
  virtual int idA()    const { return 21; }
  virtual int idB()    const { return 1; }
  virtual int id0()    const { return 21; }
  virtual int id1()    const { return 21; }
  virtual int id2()    const { return 1; }

  // The antenna function
  virtual double antFun(vector<double> invariants, vector<double> masses,
    vector<int> helBef, vector<int> helNew);

  // AP splitting kernel for collinear limit checks
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> /* mNew */, vector<int> helBef, vector<int> helNew);

};

//==========================================================================

// class GGEmitII

class GGEmitII : public AntennaFunctionIX {

 public:

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:GGemitII"; };

  // Parton types AB -> 0a 1j 2b with A,B,a,b initial and j final
  virtual int idA()    const { return 21; }
  virtual int idB()    const { return 21; }
  virtual int id0()    const { return 21; }
  virtual int id1()    const { return 21; }
  virtual int id2()    const { return 21; }

  // The antenna function
  virtual double antFun(vector<double> invariants, vector<double> masses,
    vector<int> helBef, vector<int> helNew);

  // AP splitting kernel
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> /* mNew */, vector<int> helBef, vector<int> helNew);

};

//==========================================================================

// class QXSplitII

// Splitting in the forwards sense aka quark backwards evolving to a gluon and
// emitting an antiquark in the final state

class QXSplitII : public AntennaFunctionIX {

 public:

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:QXsplitII"; };

  // Parton types AB -> 0a 1j 2b with A,B, a,b initial and j final
  virtual int idA()    const { return 1; }
  virtual int idB()    const { return 0; }
  virtual int id0()    const { return 21; }
  virtual int id1()    const { return -1; }
  virtual int id2()    const { return 0; }

  // The antenna function
  virtual double antFun(vector<double> invariants, vector<double> masses,
    vector<int> helBef, vector<int> helNew);

  // AP splitting kernel
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> /* mNew */, vector<int> helBef, vector<int> helNew);

  // Mark that this function has no zB collinear limit
  virtual double zB(vector<double>) {return -1.0;}

};

//==========================================================================

// class GXConvII

// Gluon evolves backwards into a quark and emits a quark in the final state

class GXConvII : public AntennaFunctionIX {

 public:

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:GXconvII"; };

  // Parton types AB -> 0a 1j 2b with A,B,a,b initial and j final
  virtual int idA()    const { return 21; }
  virtual int idB()    const { return 0; }
  virtual int id0()    const { return 2; }
  virtual int id1()    const { return 2; }
  virtual int id2()    const { return 0; }

  // The antenna function
  virtual double antFun(vector<double> invariants, vector<double> masses,
    vector<int> helBef, vector<int> helNew);

  // AP splitting kernel
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> /* mNew */, vector<int> helBef, vector<int> helNew);

  // Mark that this function has no zB collinear limit
  virtual double zB(vector<double>) {return -1.0;}

};

//==========================================================================

// Derived classes for Global Initial-Final antenna functions

//==========================================================================

// class AntennaFunctionIF

// Base class for IF antenna functions. (Base class implements QQEmitIF)

class AntennaFunctionIF : public AntennaFunctionIX {

 public:

  // Method to initialise (can be different than that of the base class)
  virtual bool init();

  virtual bool check();

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:AntennaFunctionIF"; };

  // Parton types AB -> 0a 1j 2b with A,a initial and B,b,j final
  virtual int idA()    const { return 1; }
  virtual int idB()    const { return -1; }
  virtual int id0()    const { return 1; }
  virtual int id1()    const { return 21; }
  virtual int id2()    const { return -1; }

  // Functions to get Altarelli-Parisi energy fractions
  virtual double zA(vector<double> invariants) {
    double sAK = invariants[0];
    double sjk = invariants[2];
    return sAK/(sAK+sjk);
  }
  virtual double zB(vector<double> invariants) {
    double sAK = invariants[0];
    double saj = invariants[1];
    return (sAK-saj)/sAK;
  }

  // Methods to tell II, IF, and RF apart.
  virtual bool isIIant() { return false; }
  virtual bool isRFant() { return false; }

  // Check for resonances
  virtual bool checkRes();

  // Called in checkRes
  virtual void getTestMasses(vector<double> &masses){
    masses.clear();
    masses.resize(4,0.0);
  }

  // Called in checkRes
  virtual bool getTestInvariants(vector<double> &invariants,
    vector<double> masses,
    double yaj, double yjk){
    invariants.clear();
    if(masses.size()!=4){
      return false;
    }
    double mA=masses[0];
    double mK=masses[2];
    double mAK=masses[3];
    double sAK = mA*mA +mK*mK - mAK*mAK;
    double sjk = sAK*yjk/(1.-yjk);
    if((sjk+sAK)==0.0){
      return false;
    }
    double saj = yaj*(sAK+sjk);
    double sak = sAK + sjk -saj;
    double det = saj*sjk*sak - saj*saj*mK*mK - sjk*sjk*mA*mA;
    if(det <0.){
      return false;
    }

    invariants.push_back(sAK);
    invariants.push_back(saj);
    invariants.push_back(sjk);
    invariants.push_back(sak);
    return true;
  }


 protected:

  //massive eikonal factor
  //NB - is positive definite - proportional to gram determinant
  double massiveEikonal(double saj, double sjk, double sak,
    double m_a, double m_k){
    double antsoft = 2.0*sak/(saj*sjk);
    antsoft -= 2.0*m_a*m_a/(saj*saj);
    antsoft -= 2.0*m_k*m_k/(sjk*sjk);
    return antsoft;
  }

  //overloaded version
  double massiveEikonal(vector<double> invariants, vector<double> masses){
    //1) extract invariants
    double saj = invariants[1];
    double sjk = invariants[2];
    double sak = invariants[3];

    //2) extract masses
    double mares = masses[0];
    double mkres = masses[2];

    //3) Massive eikonal
    double antsoft = massiveEikonal(saj,sjk,sak,mares,mkres);
    return antsoft;
  }

  double gramDet(vector<double> invariants, vector<double> masses){
    //1) extract invariants
    double saj = invariants[1];
    double sjk = invariants[2];
    double sak = invariants[3];

    //2) extract masses
    double mares = masses[0];
    double mjres = masses[1];
    double mkres = masses[2];

    double det = 0.25*(saj*sjk*sak - saj*saj*mkres*mkres -sak*sak*mjres*mjres
      - sjk*sjk*mares*mares
      + 4.0*mares*mares*mjres*mjres*mkres*mkres);
    return det;
  }

  // wrapper for comparing to AP functions - sums over flipped invariants
  // where appropriate
  double antFunCollLimit(vector<double> invariants,vector<double> masses);

};

//==========================================================================

// class QQEmitIF

class QQEmitIF : public AntennaFunctionIF {

 public:

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:QQemitIF"; };

  // Parton types AB -> 0a 1j 2b with A,a initial and B,b,j final
  virtual int idA()    const { return 1; }
  virtual int idB()    const { return -1; }
  virtual int id0()    const { return 1; }
  virtual int id1()    const { return 21; }
  virtual int id2()    const { return -1; }

  // The antenna function
  virtual double antFun(vector<double> invariants, vector<double> masses,
    vector<int> helBef, vector<int> helNew);

  // AP splitting kernel
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> /* mNew */, vector<int> helBef, vector<int> helNew);

  // Functions to get Altarelli-Parisi energy fractions
  virtual double zA(vector<double> invariants) {
    double sAK = invariants[0];
    double sjk = invariants[2];
    return sAK/(sAK+sjk);
  }
  virtual double zB(vector<double> invariants) {
    double sAK = invariants[0];
    double saj = invariants[1];
    return (sAK-saj)/sAK;
  }

};

//==========================================================================

// class QGEmitIF

class QGEmitIF : public AntennaFunctionIF {

 public:

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:QGemitIF"; };

  // Parton types AB -> 0a 1j 2b with A,a initial and B,b,j final
  virtual int idA()    const { return 1; }
  virtual int idB()    const { return 21; }
  virtual int id0()    const { return 1; }
  virtual int id1()    const { return 21; }
  virtual int id2()    const { return 21; }

  // The antenna function
  virtual double antFun(vector<double> invariants, vector<double> masses,
    vector<int> helBef, vector<int> helNew);

  // AP splitting kernel
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> /* mNew */, vector<int> helBef, vector<int> helNew);

};

//==========================================================================

// class GQEmitIF

class GQEmitIF : public AntennaFunctionIF {

 public:

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:GQemitIF"; };

  // Parton types AB -> 0a 1j 2b with A,a initial and B,b,j final
  virtual int idA()    const { return 21; }
  virtual int idB()    const { return 1; }
  virtual int id0()    const { return 21; }
  virtual int id1()    const { return 21; }
  virtual int id2()    const { return 1; }

  // The antenna function
  virtual double antFun(vector<double> invariants, vector<double> masses,
    vector<int> helBef, vector<int> helNew);

  // AP splitting kernel
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> /* mNew */, vector<int> helBef, vector<int> helNew);

};

//==========================================================================

// class GGEmitIF

class GGEmitIF : public AntennaFunctionIF {

 public:

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:GGemitIF"; };

  // Parton types AB -> 0a 1j 2b with A,a initial and B,b,j final
  virtual int idA()    const { return 21; }
  virtual int idB()    const { return 21; }
  virtual int id0()    const { return 21; }
  virtual int id1()    const { return 21; }
  virtual int id2()    const { return 21; }

  // The antenna function
  virtual double antFun(vector<double> invariants, vector<double> masses,
    vector<int> helBef, vector<int> helNew);

  // AP splitting kernel
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> /* mNew */, vector<int> helBef, vector<int> helNew);

};

//==========================================================================

// class QXSplitIF

// Splitting in the forwards sense aka quark backwards evolving to a gluon
// and emitting an antiquark in the final state

class QXSplitIF : public AntennaFunctionIF {

 public:

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:QXsplitIF"; };

  // Parton types AB -> 0a 1j 2b with A,a initial and B,b,j final
  virtual int idA()    const { return 1; }
  virtual int idB()    const { return 0; }
  virtual int id0()    const { return 21; }
  virtual int id1()    const { return -1; }
  virtual int id2()    const { return 0; }

  // The antenna function
  virtual double antFun(vector<double> invariants, vector<double> masses,
    vector<int> helBef, vector<int> helNew);

  // AP splitting kernel
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> /* mNew */, vector<int> helBef, vector<int> helNew);

  // Mark that this function does not have a zB collinear limit
  virtual double zB(vector<double>) { return -1.0; }

};

//==========================================================================

// class GXConvIF

// Gluon evolves backwards into a quark and emits a quark in the final state

class GXConvIF : public AntennaFunctionIF {

 public:

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:GXconvIF"; };

  // Parton types AB -> 0a 1j 2b with A,a initial and B,b,j final
  virtual int idA()    const { return 21; }
  virtual int idB()    const { return 0; }
  virtual int id0()    const { return 2; }
  virtual int id1()    const { return 2; }
  virtual int id2()    const { return 0; }

  // The antenna function
  virtual double antFun(vector<double> invariants, vector<double> masses,
    vector<int> helBef, vector<int> helNew);

  // AP splitting kernel
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> /* mNew */, vector<int> helBef, vector<int> helNew);

  // Mark that this function does not have a zB collinear limit.
  virtual double zB(vector<double>) { return -1.0; }

};

//==========================================================================

// class XGSplitIF

// Gluon splitting in the final state

class XGSplitIF : public AntennaFunctionIF {

 public:

  // Names (remember to redefine both for each inherited class)
  virtual string vinciaName()    const { return "Vincia:XGsplitIF"; };

  // Parton types AB -> 0a 1j 2b with A,a initial and B,b,j final
  virtual int idA()    const { return 0; }
  virtual int idB()    const { return 21; }
  virtual int id0()    const { return 0; }
  virtual int id1()    const { return -1; }
  virtual int id2()    const { return 1; }

  // The antenna function
  virtual double antFun(vector<double> invariants, vector<double> masses,
    vector<int> helBef, vector<int> helNew);

  // AP splitting kernel
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> /* mNew */, vector<int> helBef, vector<int> helNew);

  // Mark that this function does not have a zA collinear limit
  virtual double zA(vector<double>) { return -1.0; }

};

//==========================================================================

// Derived classes for Sector Initial-Final antenna functions
// (Note only the final-state leg needs to be symmetrised, as the
// global IF functions already contain sector terms on their initial-state
// legs to account for the absence of "emission into the initial state".)

//==========================================================================

// class QGEmitIFsec

class QGEmitIFsec : public QGEmitIF {

 public:

 public:

  // The (dimensionless) antenna function
  virtual double antFun(vector<double> invariants,
    vector<double> mNew, vector<int> helBef, vector<int> helNew) {
    double ant = QGEmitIF::antFun(invariants, mNew, helBef, helNew);
    // Check if helicity vectors empty
    if (helBef.size() < 2) {
      helBef.push_back(9);
      helBef.push_back(9);
    }
    if (helNew.size() < 3) {
      helNew.push_back(9);
      helNew.push_back(9);
      helNew.push_back(9);
    }
    // Check if j has same helicity as parent final-state gluon.
    int hG = helBef[1];
    int hjNow = helNew[1];
    if ( hG == hjNow ) {
      // invariants = sAK, saj, sjk
      // Define j<->k symmetrisation term: sak = sAK - saj + sjk
      vector<double> invariantsSym = invariants;
      double s02 = invariants[0] - invariants[1] + invariants[2];
      vector<int> helSym = helNew;
      invariantsSym[1] = s02 + sectorDampSav * invariants[2];
      helSym[1] = helNew[2];
      helSym[2] = helNew[1];
      ant += QGEmitIF::antFun(invariantsSym, mNew, helBef, helSym);
    }
    return ant;
  }

};

//==========================================================================

// Derived classes for Resonance-Final AntennaFunctions

//==========================================================================

class QQEmitRF : public QQEmitIF {

 public:

  // Names (remember to redefine both for each inherited class)
  string vinciaName()    const { return "Vincia:QQEmitRF"; };
  // Parton types AB -> ijk with A,i initial and B,k,j final
  int idA()    const { return 6; }
  int idB()    const { return 5; }
  int id0()    const { return 6; }
  int id1()    const { return 21; }
  int id2()    const { return 5; }

  double zA(vector<double>){ return -1;}
  bool isRFant() {return true;}

  void getTestMasses(vector<double> &masses){
    masses.clear();
    //Top mass
    double mA = particleDataPtr->m0(6);
    masses.push_back(mA);
    //Gluon mass
    double mjres = 0.0;
    masses.push_back(mjres);
    //Bottom mass
    double mK = particleDataPtr->m0(5);
    masses.push_back(mK);
    //W mass
    double mAK = particleDataPtr->m0(24);
    masses.push_back(mAK);
  }

  //AP with dummy helicities
  virtual double AltarelliParisi(vector<double> invariants,
    vector<double> masses, vector<int>,vector<int>){
    double sjk = invariants[2];
    double mkres = masses[2];

    double z = zB(invariants);
    double mu2 = mkres*mkres/sjk;
    double Pz = dglapPtr->Pq2gq(z,9,9,9,mu2);
    return Pz/sjk;

  };
  /*
    double AltarelliParisi(vector<double> invariants, vector<double> masses,
    vector<int> hPre, vector<int> hPost){
    cout<<" Not implemented "<<endl;
    return 0.;
    }
  */
  /* private: */

  /*  //collinear piece of the anntenna function */
  /*  double antColl(double sAK, double saj, double sjk){ */
  /*    double coll_term = (sjk/saj + saj/sjk); */
  /*    coll_term /= sAK; */
  /*    return coll_term; */
  /*  }; */

};

//--------------------------------------------------------------------------

class QGEmitRF : public QGEmitIF {

 public:

  // Names (remember to redefine both for each inherited class)
  string vinciaName()    const { return "Vincia:QGEmitRF"; };

  // Parton types AB -> ijk with A,i initial and B,k,j final
  int idA()    const { return 6; }
  int idB()    const { return 21; }
  int ida()    const { return 6; }
  int idb()    const { return 21; }
  int id1()    const { return 21; }

  double zA(vector<double>){ return -1;}
  bool isRFant() {return true;}

  void getTestMasses(vector<double> &masses){
    masses.clear();
    //Top mass
    double mA = particleDataPtr->m0(6);
    masses.push_back(mA);
    //Gluon mass
    double mjres = 0.0;
    masses.push_back(mjres);
    double mK = 0.0;
    masses.push_back(mK);
    //X mass
    double mAK = 0.6*particleDataPtr->m0(6);
    masses.push_back(mAK);
  };

  //AP with dummy helicities and masses
  virtual double AltarelliParisi(vector<double> invariants,vector<double>,
    vector<int>,vector<int>){
    double sjk = invariants[2];
    double z = zB(invariants);
    double Pz = dglapPtr->Pg2gg(z,9,9,9);
    return Pz/sjk;
  };

  /* private: */

  /*  //collinear piece of the anntenna function */
  /*  double antColl(double sAK, double saj, double sjk){ */
  /*    double coll_term = sjk/saj + saj/sjk*(sAK -saj)/sAK; */
  /*    coll_term /= sAK; */
  /*    return coll_term; */
  /*  }; */
};


//--------------------------------------------------------------------------

class XGSplitRF : public XGSplitIF {

 public:

  // Names (remember to redefine both for each inherited class)
  string vinciaName()    const { return "Vinica:XGSplitRF"; };

  double zA(vector<double>){ return -1;}
  bool isRFant() {return true;}

  // Parton types AB -> ijk with A,i initial and B,k,j final
  int idA()    const { return 6; }
  int idB()    const { return 21; }
  int ida()    const { return 6; }
  int idb()    const { return 2; }
  int id1()    const { return -2; }

  void getTestMasses(vector<double> &masses){
    masses.clear();
    //Top mass
    double mA = particleDataPtr->m0(6);
    masses.push_back(mA);
    //Gluon mass
    double mjres = 0.0;
    masses.push_back(mjres);
    double mK = 0.0;
    masses.push_back(mK);
    //X mass
    double mAK = 0.6*particleDataPtr->m0(6);
    masses.push_back(mAK);
  };

  //AP with dummy helicities
  double AltarelliParisi(vector<double> invariants, vector<double> masses,
    vector<int>,vector<int>){
    //1) extract invariants
    double sAK = invariants[0];
    double saj = invariants[1];
    double sjk = invariants[2];
    //2) extract mass
    double mkres = masses[2];
    double m2q = mkres*mkres;
    double Q2 = sjk + 2*m2q;
    double mu2 = m2q/Q2;
    double z = (sAK+saj-Q2)/sAK;
    double Pz = dglapPtr->Pg2qq(z,9,9,9,mu2);
    return Pz/Q2;
  };


};

//==========================================================================

// class GGEmitIFsec

class GGEmitIFsec : public GGEmitIF {

 public:

  // The (dimensionless) antenna function
  virtual double antFun(vector<double> invariants,
    vector<double> mNew, vector<int> helBef, vector<int> helNew) {
    double ant = GGEmitIF::antFun(invariants, mNew, helBef, helNew);
    // Check if helicity vectors empty
    if (helBef.size() < 2) {
      helBef.push_back(9);
      helBef.push_back(9);
    }
    if (helNew.size() < 3) {
      helNew.push_back(9);
      helNew.push_back(9);
      helNew.push_back(9);
    }
    // Check if j has same helicity as parent final-state gluon.
    int hG = helBef[1];
    int hjNow = helNew[1];
    if ( hG == hjNow ) {
      // invariants = sAK, saj, sjk
      // Define j<->k symmetrisation term: sak = sAK - saj + sjk
      vector<double> invariantsSym = invariants;
      double s02 = invariants[0] - invariants[1] + invariants[2];
      vector<int> helSym = helNew;
      invariantsSym[1] = s02 + sectorDampSav * invariants[2];
      helSym[1] = helNew[2];
      helSym[2] = helNew[1];
      ant += GGEmitIF::antFun(invariantsSym, mNew, helBef, helSym);
    }
    return ant;
  }

};

//==========================================================================

// class XGSplitIFsec

// Gluon splitting in the final state

class XGSplitIFsec : public XGSplitIF {

 public:

  // The (dimensionless) antenna function
  virtual double antFun(vector<double> invariants, vector<double> mNew,
    vector<int> helBef, vector<int> helNew) {
    // The sector antenna is just 2 * global.
    return 2*XGSplitIF::antFun(invariants,mNew,helBef,helNew);
  }

};

//==========================================================================

// The AntennaSetFSR class:
// Contains a list of FSR AntennaFunction Functions

class AntennaSetFSR {

 public:

  // Default Constructor does nothing
  AntennaSetFSR() {
    isInitPtr=false;
    isInit=false;
  };

  // Destructor: delete the antennae
  virtual ~AntennaSetFSR() {
    for (map<int,AntennaFunction*>::iterator it = antFunPtrs.begin();
         it != antFunPtrs.end(); ++it) delete it->second;
    antFunPtrs.clear();
  }

  // Initialize pointers
  void initPtr(ParticleData* particleDataPtrIn, Settings* settingsPtrIn,
    Rndm* rndmPtrIn, DGLAP* dglapPtrIn) {
    particleDataPtr = particleDataPtrIn;
    settingsPtr     = settingsPtrIn;
    rndmPtr         = rndmPtrIn;
    dglapPtr        = dglapPtrIn;
    isInitPtr       = true;
  }

  // Initialize antenna set (optionally with min or max variation)
  void init();

  // Function to chek if an antenna with the given index exists
  bool exists(int iAntIn) { return antFunPtrs.count(iAntIn); }

  // Gets an antenna iAntIn from the AntennaSet
  AntennaFunction* getAnt(int iAntIn) {
    if (exists(iAntIn)) return antFunPtrs[iAntIn];
    else return NULL;
  }

  // Method to return all iAntPhys values that are defined in antFunPtr
  vector<int> getIant() {
    vector<int> iAnt;
    map<int,AntennaFunction*>::iterator it;
    for (it = antFunPtrs.begin(); it != antFunPtrs.end(); ++it)
      iAnt.push_back(it->first);
    return iAnt;
  }

  // Get Vincia name (e.g., "Vincia:QQ:emit")
  string vinciaName(int iAntIn) {
    if (exists(iAntIn)) return antFunPtrs[iAntIn]->vinciaName();
    else return "noVinciaName";
  }

  // Get human name (e.g., "g/qq")
  string humanName(int iAntIn) {
    if (exists(iAntIn)) return antFunPtrs[iAntIn]->humanName();
    else return "noHumanName";
  }

 private:

  // Use a map of AntennaFunction pointers, create them with new on
  // initialization
  map<int,AntennaFunction*> antFunPtrs;

  // Pointers to Pythia-8 classes, needed to initialise antennae
  bool isInitPtr;
  bool isInit;
  ParticleData* particleDataPtr;
  Settings*     settingsPtr;
  Rndm*         rndmPtr;

  // Pointer to VINCIA DGLAP class
  DGLAP*        dglapPtr;

  // Verbose level
  int verbose;

}; // End class AntennaSetFSR

//==========================================================================

// The AntennaSetISR class

class AntennaSetISR {

 public:

  // Default Constructor
  AntennaSetISR() {
    isInitPtr=false;
    isInit=false;
  };

  // Destructor: delete the antennae
  ~AntennaSetISR(){
    for (vector<AntennaFunctionIX*>::iterator it = antFunPtrs.begin();
         it != antFunPtrs.end(); ++it) delete *it;
  }
  // Initialize pointers
  void initPtr(ParticleData* particleDataPtrIn, Settings* settingsPtrIn,
    Rndm* rndmPtrIn, DGLAP* dglapPtrIn) {
    particleDataPtr = particleDataPtrIn;
    settingsPtr     = settingsPtrIn;
    rndmPtr         = rndmPtrIn;
    dglapPtr        = dglapPtrIn;
    isInitPtr       = true;
  }

  // Initialize antenna set
  void init();

  // Function to chek if an antenna with the given index exists
  bool exists(int iAntIn) {
    if (!isInit) return false;
    return ( iAntIn<0 || iAntIn >= int(antFunPtrs.size()) ) ? false : true;
  }

  // Gets an antenna iAntIn from the AntennaSetISR
  AntennaFunctionIX* getAnt(vector<AntennaFunctionIX*>::size_type iAntIn) {
    if (exists(iAntIn)) return antFunPtrs[iAntIn];
    else return NULL;
  }

  // Get Vincia name (e.g., "Vincia:QQemitII")
  string vinciaName(vector<AntennaFunctionIX*>::size_type iAntIn) {
    if (iAntIn <= antFunPtrs.size()) return antFunPtrs[iAntIn]->vinciaName();
    else return "noVinciaName";
  }

 private:

  // use a vector of AntennaFunctionIX pointers, create them with new on
  // initialization
  vector<AntennaFunctionIX*> antFunPtrs;

  // Pointers to Pythia-8 classes, needed to initialise antennae
  bool isInitPtr;
  bool isInit;
  ParticleData* particleDataPtr;
  Settings*     settingsPtr;
  Rndm*         rndmPtr;

  // Pointer to VINCIA DGLAP class
  DGLAP*        dglapPtr;

  // Verbose level
  int verbose;

}; // End class AntennaSetISR

//==========================================================================

// class MECs
// for computing matrix-element corrections to antenna functions

class MECs {

 public:

  // Constructor.
  MECs() : isInitPtr(false), isInit(false) {};

  // Destructor.
  virtual ~MECs() {};

  // Initialize Pointers
  void initPtr(Settings* settingsPtrIn, Info* infoPtrIn, Rndm* rndmPtrIn,
    PartonSystems* partonSystemsPtrIn,
    VinciaMG5cxxInterface* mg5cxxInterfacePtrIn, VinciaCommon* vinComPtrIn) {
    settingsPtr      = settingsPtrIn;
    infoPtr          = infoPtrIn;
    rndmPtr          = rndmPtrIn;
    partonSystemsPtr = partonSystemsPtrIn;
    mg5cxxInterfacePtr = mg5cxxInterfacePtrIn;
    vinComPtr        = vinComPtrIn;
    isInitPtr        = true;
  }

  // Initialize Pointers to Antenna sets
  void initAntPtr(AntennaSetFSR* antFSRusr, AntennaSetISR* antISRusr) {
    antSetFSR = antFSRusr;
    antSetISR = antISRusr;
  }

  // Initialize
  void init();

  // Function to return ME class (Born type) for a parton configuration.
  // Can be called from either of the ISR::prepare() or FSR::prepare()
  // functions, or from the ISR::branch() or FSR::branch() functions.
  // Return : >= 0 We have an ME for this configuration, associated with the
  //              (arbitrary) integer code label denoted by the return value.
  //          < 0 We don't have an ME / no ME should be used for this system.
  bool prepare(const int iSys, Event& event);

  // Function to assign helicities to particles (using Matrix Elements)
  bool polarise(const int iSys, Event& event);

  // Method to ensure massive partons are treated as unpolarised
  bool depolariseMassive(const int iSys, Event& event);

  // Make list of particles as vector<Particle>
  vector<Particle> makeParticleList(const int iSys, const Event& event,
    const vector<Particle> pNew = vector<Particle> (),
    const vector<int> iOld = vector<int> ());

  // Check if state already has helicities
  // Return true if any particle in state already has a specified helicity
  bool isPolarised(int iSys, Event& event, bool checkIncoming) {
    for (int i=0; i<partonSystemsPtr->sizeAll(iSys); ++i) {
      int ip=partonSystemsPtr->getAll(iSys,i);
      if (ip == 0) continue;
      if (event[ip].isFinal() || checkIncoming) {
        if (event[ip].pol() != 9) return true;
      }
    }
    return false;
  }

  // Wrapper function to return a specific antenna function
  AntennaFunction* getAntFSR(const int iAnt) {
    return antSetFSR->getAnt(iAnt); }
  AntennaFunctionIX* getAntISR(const int iAnt) {
    return antSetISR->getAnt(iAnt); }

  // Function to determine if MECs are requested at this order for this system
  bool doMEC(const int iSys, const int nBranch);

  // Get squared Matrix Element
  double getME2(const vector<Particle> parts, const int nIn);
  double getME2(const int iSys, const Event& event);

  // Get ME Correction factor
  double getMEC(const int, const Event&) {
    // TMP: dummy implementation
    return 1.;
  }

  // Return number of partons added since Born (as defined by prepare).
  int sizeOutBorn(const int iSys) { return sizeOutBornSav[iSys]; }

  // Function to set level of verbose output
  void setVerbose(const int verboseIn) { verbose = verboseIn; }

  // Header
  void header();

  // Verbose level
  int verbose;

 private:

  // Is initialized?
  bool isInitPtr, isInit;

  // Pointers to PYTHIA 8 objects
  Settings*      settingsPtr;
  Info*          infoPtr;
  Rndm*          rndmPtr;
  PartonSystems* partonSystemsPtr;

  // Pointers to VINCIA objects
  VinciaMG5cxxInterface* mg5cxxInterfacePtr;
  VinciaCommon*  vinComPtr;

  // Antenna Sets
  AntennaSetFSR* antSetFSR;
  AntennaSetISR* antSetISR;

  // Matching settings
  bool matchingFullColour;
  int  nMECs2to1, nMECs2to2, nMECs2toN, nMECsResDec, nMECsMPI;
  int  nFlavZeroMass;

  // Map from iSys to Born multiplicity and ME class; set in prepare()
  map<int,int> sizeOutBornSav;

};

} // End Vincia namespace

#endif
