// VinciaFSR.h is a part of the PYTHIA event generator.
// Copyright (C) 2019 Torbjorn Sjostrand.
// PYTHIA is licenced under the GNU GPL v2 or later, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// This file contains header information for the VinciaFSR class for
// QCD final-state antenna showers, and auxiliary classes.

#ifndef VINCIA_RESFSR_H
#define VINCIA_RESFSR_H

#include "Pythia8/TimeShower.h"
#include "Pythia8/VinciaAntennaFunctions.h"
#include "Pythia8/VinciaCommon.h"
#include "Pythia8/VinciaISR.h"
#include "Pythia8/VinciaQED.h"
#include "Pythia8/VinciaWeights.h"
#include "Pythia8/VinciaDiagnostics.h"

namespace Pythia8 {

// Forward declarations
class VinciaISR;
class ResScaleHook;

//==========================================================================

// UserHooks that are is to set the scale in resonance decays

class ResScaleHook : public UserHooks {
 public:
  ResScaleHook() { };
  virtual ~ResScaleHook() { };
  // Start resonance showers at m
  bool canSetResonanceScale() { return true; }
  virtual double scaleResonance(int iRes, const Event& event) {
    return event[iRes].m(); }
};

//==========================================================================

// Helper struct for passing trial-alphaS information to Brancher class
// Note: vectors of evolutionWindow could be encapsulated in a class,
// containing only a single instance of Rndm*.

struct EvolutionWindow {
  int runMode;
  double alphaSmax;
  double b0;
  double kMu2;
  double lambda2;
  double qMin;
  map<int,double> mass;
};

//==========================================================================

//Helper struct to store information about junctions that involved resonances
//that have now decayed

struct resJunctionInfo {
  //number of junction in event record
  int iJunction;
  //which of the three ends of iJunction are we talking about
  int iEndCol;
  //What is the colour tag of iEndCol in iJunctions
  int iEndColTag;
  //pythia index of quark that should be on the end.
  //specifically it should be the quark from the decay of the resonance
  int iEndQuark;
  //iEndColTag and iEnd Quark will get updated as the quark radiates.
  //vector of colours lines between res and end quark: needed
  //in order to track if a gluon in chain splits to find correct junction
  //end.
  vector<int> colours;

};

//==========================================================================

// The Brancher class:
// Base class containing a generic set of "parent partons"
// as well as virtual methods for generating trial branchings.

class Brancher {

 public:

  // Default Constructor
  Brancher() {};

  // Create branchelemental for antenna(e) with arbitrary parents in iIn
  Brancher(int iSysIn, Event& event, vector<int> iIn) {
    reset(iSysIn, event, iIn);
  };

  // Wrapper for simple 2- (or 3-) parton antennae
  Brancher(int iSysIn, Event& event, int i0, int i1, int i2=0) {
    vector<int> iIn;
    iIn.push_back(i0);
    iIn.push_back(i1);
    if (i2 >= 1) iIn.push_back(i2);
    reset(iSysIn,event,iIn);
  }

  // Destructor
  virtual ~Brancher() {};

  // Reset (common functionality implemented in base class)
  virtual void reset(int iSysIn, Event& event, vector<int> iIn);

  // Wrapper for simple 2- (or 3-) parton antennae
  void reset(int iSysIn, Event& event, int i0, int i1, int i2=0) {
    vector<int> iIn;
    iIn.push_back(i0);
    iIn.push_back(i1);
    if (i2 >= 1) iIn.push_back(i2);
    reset(iSysIn,event,iIn);
  }

  // Methods to get (explicit for up to three parents, otherwise just use iVec)
  int i0() const { return (iSav.size() >= 1) ? iSav[0] : -1; }
  int i1() const { return (iSav.size() >= 2) ? iSav[1] : -1; }
  int i2() const { return (iSav.size() >= 3) ? iSav[2] : -1; }
  int iVec(unsigned int i) const { return (iSav.size() > i) ? iSav[i] : -1; }
  vector<int> iVec() { return iSav; }
  int id0() const { return (idSav.size() >= 1) ? idSav[0] : -1; }
  int id1() const { return (idSav.size() >= 2) ? idSav[1] : -1; }
  int id2() const { return (idSav.size() >= 3) ? idSav[2] : -1; }
  vector<int> idVec() const { return idSav; }
  int colType0() const { return (colTypeSav.size() >= 1) ? colTypeSav[0] : -1;}
  int colType1() const { return (colTypeSav.size() >= 2) ? colTypeSav[1] : -1;}
  int colType2() const { return (colTypeSav.size() >= 3) ? colTypeSav[2] : -1;}
  vector<int> colTypeVec() const { return colTypeSav; }
  int col0() const { return (colSav.size() >= 1) ? colSav[0] : 0; }
  int col1() const { return (colSav.size() >= 2) ? colSav[1] : 0; }
  int col2() const { return (colSav.size() >= 3) ? colSav[2] : 0; }
  vector<int> colVec() const { return colSav; }
  int acol0() const { return (acolSav.size() >= 1) ? acolSav[0] : 0; }
  int acol1() const { return (acolSav.size() >= 2) ? acolSav[1] : 0; }
  int acol2() const { return (acolSav.size() >= 3) ? acolSav[2] : 0; }
  vector<int> acolVec() const { return acolSav; }
  int h0() const { return (hSav.size() >= 1) ? hSav[0] : -1; }
  int h1() const { return (hSav.size() >= 2) ? hSav[1] : -1; }
  int h2() const { return (hSav.size() >= 3) ? hSav[2] : -1; }
  vector<int> hVec() const { return hSav; }
  double m0() const { return (mSav.size() >= 1) ? mSav[0] : -1; }
  double m1() const { return (mSav.size() >= 2) ? mSav[1] : -1; }
  double m2() const { return (mSav.size() >= 3) ? mSav[2] : -1; }
  vector<double> mVec() const { return mSav; }
  vector<double> getmPostVec() { return mPostSav; };
  int colTag(){return colTagSav;};

  // Method to get maximum value of evolution scale for this brancher
  // (Must be redefined in base classes.
  virtual double getQ2Max(int) = 0;

  // Methods to return saved / derived quantities
  int    system() const { return systemSav; }
  double mAnt() const { return mAntSav; }
  double m2Ant() const { return m2AntSav; }
  double sAnt() const { return sAntSav; }
  double kallenFac() const { return kallenFacSav; }
  double enhanceFac() const { return enhanceSav; }
  double q2Trial() const {
    return q2NewSav;
  }
  int iAntPhys() const { return iAntSav; }

  //-------------------------------------------------------------------------
  // Virtual methods that MUST be redefined in derived classes
  // Init (for functionality specific to derived classes)
  virtual void init() { iAntSav = -1; }
  
  // Generate a new Q2 scale
  virtual double genQ2(int evTypeIn, double Q2MaxNow, Rndm* rndmPtr,
    const EvolutionWindow* evWindowPtrIn, double colFac,
    vector<double> headroomIn, vector<double> enhanceFacIn,
    int verboseIn) = 0;

  // Generate complementary invariant(s) for saved trial.
  // Return false if no physical kinematics possible.
  virtual bool genInvariants(vector<double>& invariants,Rndm*,int) {
    // Base class returns false;
    invariants.clear();
    return false;
  }

  // Compute antPhys / antTrial, given an input value for antPhys
  virtual double pAccept(const double,int=0) {
    // Base class returns zero.
    return 0.;
  }

  // Compute pT scale of trial branching
  virtual double getpTscale(){
    if(invariantsSav.size() ==3){
      double sIK = invariantsSav[0];
      double y12 = invariantsSav[1] / sIK;
      double y23 = invariantsSav[2] / sIK;
      return sIK * y12 * y23;
    } else return 0.;
  }

  // 
  virtual double getXj(){
    if(invariantsSav.size() ==3){
      double sIK = invariantsSav[0];
      double y12 = invariantsSav[1] / sIK;
      double y23 = invariantsSav[2] / sIK;
      return y12 + y23;
    } else return 1.0;
  }

  // What kind of particle(s) are being created.
  // (E.g., return 21 for gluon emission, quark flavour for splitting, etc)
  virtual int idNew() const { return 0; }
  virtual double mNew() const { return 0.0; }

  virtual bool getNewParticles(Event& event, vector<Vec4> momIn,
    vector<int> hIn, vector<Particle> &pNew,Rndm* rndmPtr,
    Colour* colourPtr) = 0;

  // List function
  virtual void list(string header="none") const;

  //-------------------------------------------------------------------------
  // Virtual methods that CAN be redefined in derived classes

  // Set post-branching IDs and masses. Base class is for gluon emission.
  virtual void setidPost(){
    idPostSav.clear();
    idPostSav.push_back(id0());
    idPostSav.push_back(21);
    idPostSav.push_back(id1());
  };
  virtual vector<double> setmPostVec() {
    mPostSav.clear();
    mPostSav.push_back(mSav[0]); //mi
    mPostSav.push_back(0.0); //mj
    mPostSav.push_back(mSav[1]); //mk
    return mPostSav;
  }

  virtual void setStatPost(){
    statPostSav.resize(iSav.size()+1,51);
  };

  virtual void setMaps(int){
    mothers2daughters.clear();
    daughters2mothers.clear();
  }

  // Return index of new particle (slightly arbitrary choice for splittings)
  virtual int iNew(){
    if(i0()>0 ){
      if(mothers2daughters.find(i0())!= mothers2daughters.end())
        return mothers2daughters[i0()].second;
      else return 0;      
    }
    else return 0;
  }

  // Method returns pos of resonance if there is one participating in decay
  // -1 otherwise
  virtual int posR(){ return -1; }

  // Method returns pos of colour-connected daughter to resonance
  // if there is one participating in decay -1 otherwise
  virtual int posF(){ return -1; }

  //-------------------------------------------------------------------------
  // Non-virtual methods 

  int getBranchType(){ return branchType;}
  bool isSwapped(){ return swapped;}
  vector<double> getInvariants(){ return invariantsSav;}

  // This method allows to reset enhanceFac if we do an accept/reject
  void resetEnhanceFac(const double enhanceIn) { enhanceSav = enhanceIn; }

  // Method to see if a saved trial has been generated,
  // and to erase memory of it.
  bool hasTrial() const {return hasTrialSav;}
  // Method to mark new trial needed *without* erasing current one.
  void needsNewTrial() {hasTrialSav = false;}
  // Method to mark new trial needed *and* erase current one.
  void eraseTrial() {
    hasTrialSav = false;
    q2NewSav = 0.;
  }

  // Publicly accessible members for storing mother/daughter connections
  map< int, pair<int,int> > mothers2daughters;
  map< int, pair<int,int> > daughters2mothers;

 protected:

  // Data members for storing information about parent partons
  int systemSav;
  vector<int> iSav, idSav, colTypeSav, hSav, colSav, acolSav;
  vector<int> idPostSav;
  vector<int> statPostSav;
  vector<double> mSav, mPostSav;
  int colTagSav;
  int evTypeSav;

  //all alphaS information
  const EvolutionWindow* evWindowSav;

  // Saved antenna mass parameters
  double mAntSav, m2AntSav, kallenFacSav, sAntSav;

  // Data members for storing information about generated trial branching.
  bool   hasTrialSav;
  double headroomSav, enhanceSav, q2BegSav, q2NewSav;
  vector<double> invariantsSav;

  // Find out which branching type we are doing
  int branchType;
  //   1 : QCD emission
  //   2 : QCD splitting
  //   3 : QED emission
  //   4 : QED splitting
  //   5 : QCD resonance emission

  // Indices of FF antenna functions
  int iAntSav;
  // 0 = iQQemitFF
  // 1 = iQGemitFF
  // 2 = iGQemitFF
  // 3 = iGGemitFF
  // 4 = iGXsplitFF
  // 5 = iQQemitRF
  // 6 = iQGemitRF
  // 7 = iXGsplitRF

  //if true, flip identities of A and B
  bool swapped;

};

//==========================================================================

// BrancherEmitFF: branchelemental for 2->3 gluon emissions

class BrancherEmitFF : public Brancher {

 public:

  // Create branchelemental for antenna(e) with parents in iIn
  BrancherEmitFF(int iSysIn, Event& event, vector<int> iIn) {
    reset(iSysIn, event, iIn);
  }

  // Wrapper to provide simple 2-parton systems as parents
  BrancherEmitFF(int iSysIn, Event& event, int i0, int i1) {
    reset(iSysIn, event, i0, i1);
  }

  // Destructor
  virtual ~BrancherEmitFF() {};

  // Method to initialise members specific to BrancherEmitFF
  virtual void init();

  // Generate a new Q2 scale
  double genQ2(int evTypeIn, double Q2MaxNow, Rndm* rndmPtr,
    const EvolutionWindow* evWindowPtrIn, double colFac,
    vector<double> headroomIn, vector<double> enhanceFacIn,int verboseIn);
  /* double genQ2(int evTypeIn, double q2BegIn, double colFacIn, */
  /*   EvolutionWindow* evWindowPtrIn, double headroomIn=1.,
       double enhanceFacIn=1.); */

  // Generate invariants. Return false if no physical kinematics possible.
  virtual bool genInvariants(vector<double>& invariants, Rndm* rndmPtr,
    int verboseIn);

  // Compute antPhys / antTrial, given antPhys
  // Note: antPhys should be normalised to include charge and coupling factors
  virtual double pAccept(const double antPhys,int=0);

  double getQ2Max(int evType){
    if(evType == 1){        return sAntSav/4.; }
    else if (evType == 2) { return sAntSav/9.; }
    else if (evType == 3) { return sAntSav/2.; }
    else { return 0.; }
  }

  // Method to make mothers2daughters and daughters2mothers pairs.
  virtual void setMaps(int sizeOld);

  // Flavour and mass of emitted particle
  virtual int idNew() const { return 21; }
  virtual double mNew() const { return 0.0; }

  // Generic getter method. Assumes setter methods called earlier
  virtual bool getNewParticles(Event& event, vector<Vec4> momIn,
    vector<int> hIn, vector<Particle> &pNew,Rndm* rndmPtr,Colour* colourPtr);

 private:

  // Data members to store information about generated trials.
  double colFacSav;

};

//==========================================================================

// BrancherSplitFF: branchelemental for 2->3 gluon splittings

class BrancherSplitFF : public Brancher {

 public:

  // Constructors
  // Create branchelemental for antenna(e) with parents in iIn
  BrancherSplitFF(int iSysIn, Event& event, vector<int> iIn) {
    reset(iSysIn, event, iIn);
  }

  // Wrapper to provide simple 2-parton systems as parents
  BrancherSplitFF(int iSysIn, Event& event, int i0, int i1, bool col2acol) {
    reset(iSysIn, event, i0, i1);
    // Set if it is the anticolour or colour side of the gluon that
    // participates in the antenna (used to decide pTj or pTi measure)
    if (col2acol) { isXGsav = false; }
    else { isXGsav = true; }
  }

  // Destructor
  virtual ~BrancherSplitFF() {};

  // Method to initialise data members specific to BrancherSplitFF
  virtual void init();

  // Method to check if this antenna corresponds to splitting off the
  // colour (true) or anticolour side (false) of the parent gluon.
  virtual bool isXG() const { return isXGsav; }

  // Flavour and mass of emitted particle
  virtual int idNew() const { return idFlavSav; }
  virtual double mNew() const { return mFlavSav; }

  // Generate a new Q2 scale
  double genQ2(int evTypeIn, double Q2MaxNow, Rndm* rndmPtr,
    const EvolutionWindow* evWindowPtrIn, double colFac,
    vector<double> headroomIn, vector<double> enhanceFacIn,int verboseIn);

  // Method to generate complementary invariant(s) for saved trial.
  // Returns false if no physical kinematics possible.
  virtual bool genInvariants(vector<double>& invariants, Rndm* rnmdPtr,
    int verboseIn);

  // Compute antPhys / antTrial, given antPhys
  // Note: antPhys should be normalised to include charge and coupling factors
  virtual double pAccept(const double antPhys,int);

  double getQ2Max(int evType){
    if(evType == 1) {       return sAntSav/4.; }
    else if (evType == 2) { return sAntSav; }
    else if (evType == 3) { return sAntSav; }
    else { return 0.; }
  }

  virtual vector<double> setmPostVec() {
    mPostSav.clear();
    mPostSav.push_back(mFlavSav); //mi
    mPostSav.push_back(mFlavSav); //mj
    mPostSav.push_back(mSav[1]); //mk
    return mPostSav;
  }

  virtual void setidPost(){
    idPostSav.clear();
    idPostSav.push_back(idFlavSav);
    idPostSav.push_back(-idFlavSav);
    idPostSav.push_back(id1());
  };

  virtual void setStatPost(){
    statPostSav.resize(iSav.size()+1,51);
    statPostSav[2]=52;
  };

  // Method to create the mothers2daughters and daughters2mothers maps
  virtual void setMaps(int sizeOld);

  virtual bool getNewParticles(Event& event, vector<Vec4> momIn,
    vector<int> hIn, vector<Particle> &pNew, Rndm*, Colour*);

 private:

  // Data members for storing information on generated trials.
  int     idFlavSav;
  double  mFlavSav;

  // Data member to store whether this is really an XG antenna or a
  // GX one (ie if it is the anticolour or colour side of the gluon
  // which is participating in the LC antenna; in the former case,
  // we use pT(qbar) = pTj as the measure; in the latter pT(q) = pTi.)
  bool    isXGsav;

};

//==========================================================================

//Class for storing information on antennae between a coloured resonance and
//final state parton, and generating a new emission
class BrancherEmitRF : public Brancher {

 public:

  //constructors
  BrancherEmitRF(){};

  BrancherEmitRF(int iSysIn, Event& event, vector<int> allIn,
    unsigned int posResIn, unsigned int posFIn, double Q2cut) {
    reset(iSysIn, event, allIn);
    init(event,allIn,posResIn,posFIn,Q2cut);
  }

  void resetResBrancher(int iSysIn, Event& event, vector<int> allIn,
    unsigned int posResIn, unsigned int posFIn, double Q2cut) {
    reset(iSysIn, event, allIn);
    init(event,allIn,posResIn,posFIn,Q2cut);
  }

  //Overloaded version of init - does nothing.
  void init() {};

  virtual void init(Event& event, vector<int> allIn, unsigned int posResIn,
    unsigned int posFIn, double Q2cut);

  //destructor
  ~BrancherEmitRF(){};

  virtual vector<double> setmPostVec() {
    mPostSav.clear();
    mPostSav.push_back(mRes); //ma
    mPostSav.push_back(0.0); //mj
    mPostSav.push_back(mFinal); //mk
    mPostSav.push_back(mRecoilers); //mAK
    return mPostSav;
  }

  virtual void setidPost(){
    idPostSav.clear();
    //copy idsIn
    idPostSav=idSav;
    //insert gluon in second position
    idPostSav.insert(idPostSav.begin()+1,21);
  };

  virtual void setStatPost(){
    statPostSav.resize(iSav.size()+1,52);
    statPostSav[posFinal]=51;
    statPostSav[posFinal+1]=51;
  };

  virtual int iNew() {
    if(posFinal>0 && iSav[posFinal]>0 
      && mothers2daughters.find(iSav[posFinal])!= mothers2daughters.end())
      return mothers2daughters[iSav[posFinal]].second;
    return 0;
  }

  virtual void setMaps(int sizeOld);

  //generic method.
  //assumes setter methods called earlier
  virtual bool getNewParticles(Event& event, vector<Vec4> momIn,
    vector<int> hIn, vector<Particle> &pNew, Rndm* rndmPtr, Colour*);

  //function returns pos of resonance
  int posR(){
    return int(posRes);
  }

  //function returns pos of colour-connected daughter to resonance
  int posF(){
    return int(posFinal);
  }

  double getQ2Max(int evType){
    if(evType==1){ return Q2MaxSav; }
    else { return 0.; }
  }

  virtual double genQ2(int evTypeIn, double Q2MaxNow, Rndm* rndmPtr,
    const EvolutionWindow* evWindowPtrIn, double colFac,
    vector<double> headroomIn, vector<double> enhanceFacIn,
    int verboseIn);

  virtual bool genInvariants(vector<double>& invariants,Rndm* rndmPtr,
    int verboseIn);

  virtual double pAccept(const double,int);

 protected:

  //private methods
  double KallenFunction(double x, double y, double z){
    double result = x*x + y*y + z*z;
    result-= 2.*(x*y + x*z + y*z);
    return result;
  }

  virtual double zetaIntSingleLim(double zetaLim){
    //zeta = (sjk+sAK)/sAK
    double x = zetaLim-1;
    double intZ = x + log(x);
    return intZ;
  }

  double zetaIntegral(double zLow, double zHigh){
    double zetaInt = zetaIntSingleLim(zHigh)-zetaIntSingleLim(zLow);
    return zetaInt;
  }

  double getsAK(double mA, double mK, double mAK){
    double sAKcalc = mA*mA +mK*mK - mAK*mAK;
    return sAKcalc;
  }

  double zetaMinCalc(double mA, double mK, double mAK,double Q2cut){
    double sajMax = mA*mA -(mAK + mK)*(mAK + mK);
    double zetaMinOut = 1.0/(1.0 - Q2cut/sajMax);
    return zetaMinOut;
  }

  double zetaMaxCalc(double mA, double mK, double mAK){
    double sjkMax = (mA-mAK)*(mA-mAK) - mK*mK;
    double sAKcalc = getsAK(mA,mK,mAK);
    double zetaMaxOut = 1.0 + sjkMax/sAKcalc;
    return zetaMaxOut;
  }

  virtual double getZetaNext(Rndm* rndmPtr){
    //Returns the solution for zeta to
    // R = I(zeta,zetamin)/I(zetamax,zetamin)
    double R = rndmPtr->flat();
    //I(zetamax,zetamin)
    double intZrange =zetaIntegral(zetaMin,zetaMax);
    double intZMin = zetaIntSingleLim(zetaMin);
    //exp(I(zeta))
    double expIntZeta = exp(intZrange*R + intZMin);
    double lambWFact =LambertW(expIntZeta);
    //Now invert to get zeta
    return 1. + lambWFact;
  }

  virtual double calcQ2Max(double mA, double mAK, double mK){
    double aM2 = (mA-mAK)*(mA-mAK) - mK*mK;
    double bM2 = mAK*(mA-mAK) + mK*mK;
    double cM = mA-mAK;
    double Q2 = aM2*aM2*mA/(2.0*cM*bM2);
    return Q2;
  }

  bool vetoPhSpPoint(double saj, double sjk, double sak,int verboseIn);
  double getEjMax(double cosTheta, double mA, double mAK, double mK);

  // Private data members

  // Save reference to position in vectors of resonance
  // and colour connected parton
  unsigned int posRes;
  unsigned int posFinal;

  //mass of resonance
  double mRes;
  //mass of final state parton in antennae
  double mFinal;
  //collective mass of rest of downstream decay products of resonance
  //these will just take recoil
  double mRecoilers;

  double sAK;

  //limits of zeta Integral
  double zetaMin;
  double zetaMax;

  //max Q2 for this brancher
  //(still an overestimate - but better than just mass of res)
  double Q2MaxSav;

  //Integral of zeta over whole phase space
  double zetaIntSave;

  double colFacSav;

  //bool to store whether the colour flow is from
  // R to F (e.g. t -> bW+)
  // or F to R (e.g. tbar -> bbarW-)
  bool colFlowRtoF;

  map<unsigned int,unsigned int> posNewtoOld;

};

//Class for storing information on antennae between a coloured resonance and
//final state parton, and generating a new emission
class BrancherSplitRF : public BrancherEmitRF {

 public:

  //constructor
  BrancherSplitRF(int iSysIn, Event& event, vector<int> allIn,
    unsigned int posResIn, unsigned int posFIn, double Q2cut){
    reset(iSysIn, event, allIn);
    init(event,allIn,posResIn,posFIn,Q2cut);
  }

  //Overloaded version of init - does nothing.
  void init() {};

  void init(Event& event, vector<int> allIn, unsigned int posResIn,
    unsigned int posFIn, double Q2cut);

  //destructor
  ~BrancherSplitRF(){};

  vector<double> setmPostVec() {
    mPostSav.clear();
    mPostSav.push_back(mRes);
    mPostSav.push_back(mFlavSav);
    mPostSav.push_back(mFlavSav);
    mPostSav.push_back(mRecoilers);
    return mPostSav;
  }

  virtual void setidPost(){
    idPostSav.clear();
    //copy idsIn
    idPostSav=idSav;
    //modify the splitting gluon to antiquark
    if(colFlowRtoF){
      idPostSav[posFinal] = -idFlavSav;
      //insert quark in second position
      idPostSav.insert(idPostSav.begin()+1,idFlavSav);
    }
    else{
      idPostSav[posFinal] = idFlavSav;
      //insert quark in second position
      idPostSav.insert(idPostSav.begin()+1,-idFlavSav);
    }
  };

  virtual void setStatPost(){
    statPostSav.resize(iSav.size()+1,52);
    statPostSav[1]=51;
    statPostSav[posFinal+1]=51;
  };

  //generic method.
  //assumes setter methods called earlier
  virtual bool getNewParticles(Event& event, vector<Vec4> momIn,
    vector<int> hIn, vector<Particle> &pNew, Rndm*, Colour*);
  
  double genQ2(int evTypeIn, double Q2MaxNow, Rndm* rndmPtr,
    const EvolutionWindow* evWindowPtrIn, double colFac,
    vector<double> headroomIn, vector<double> enhanceFacIn,
    int verboseIn);
  bool genInvariants(vector<double>& invariants,Rndm* rndmPtr,int verboseIn);
  double pAccept(const double,int);

 private:

  virtual double getZetaNext(Rndm* rndmPtr){
    //zeta integral for splitters is just flat
    double R = rndmPtr->flat();
    return zetaMin+R*(zetaMax-zetaMin);
  }

   double zetaMinCalc(double mA, double mK, double mAK,double Q2cut){
     //Calculate for massless j (underestimate for zetamin)
     double sajMax = mA*mA -(mAK + mK)*(mAK + mK);
     return Q2cut/sajMax + 1.- sajMax/sAK;
   }

  /* double calcQ2Max(double ma, double mAK, double mq){ */
  /*   double norm = 1.0 - 1.0/zetaMax; */
  /*   double sajmax = 2.0*(ma*ma -2.0*mq*mAK - mAK*mAK); */
  /*   double Q2 = norm*(sajmax + mq*mq); */
  /*   return Q2; */
  /* } */

  int     idFlavSav;
  double  mFlavSav;

};

//==========================================================================

// The VinciaFSR class for resonant decays
// Inherits from TimeShower in Pythia 8 so can be used as alternative to
// TimesShower.

class VinciaFSR : public Pythia8::TimeShower {

  // Allow VinciaPlugin and SpaceAntShower to access private information
  friend class VinciaPlugin;
  friend class VinciaISR;

 public:

  // ----------------------------------------------------------------------

  // CONSTRUCTOR AND DESTRUCTOR

  // Constructor.
  VinciaFSR(){
    verbose=0;
    headerIsPrinted=false;
    isInit=false;
    isPrepared=false;
    diagnosticsPtr=0;
  };

  // Destructor.
  ~VinciaFSR(){};

  // -----------------------------------------------------------------------

  // FUNCTIONS INHERITED FROM PYTHIA 8 FOR INTERFACING

  // -----------------------------------------------------------------------

  // Methods that control main flow of shower

  // Initialize alphaStrong and related pTmin parameters.
  void init( BeamParticle* beamAPtrIn = 0,
    BeamParticle* beamBPtrIn = 0);

  // Possible limitation of first emission (last two arguments purely
  // dummy in Vincia implementation).
  bool limitPTmax( Event& event, double Q2Fac = 0., double Q2Ren = 0.);

  // Top-level routine to do a full time-like shower in resonance decay.
  int shower( int iBeg, int iEnd, Event& event, double pTmax,
    int nBranchMax = 0);

  // Method to add QED showers in hadron decays
  int showerQED(int iBeg, int iEnd, Event& event, double pTmax);

  // Method to add QED showers to partons below colour resolution scale
  int showerQEDafterRemnants(Event& event);
  
  // Used for global recoil scheme in Pythia8 (No Vincia implementation yet)
  // void prepareGlobal(Event&);

  // Prepare system for evolution.
  void prepare( int iSys, Event& event, bool limitPTmaxIn);

  // Update antenna list after each ISR emission.
  void update( int iSys, Event& event, bool hasWeakRad = false);

  // Update antenna list after a multiple interactions rescattering.
  // (No Vincia implementation yet)
  // void rescatterUpdate( int iSys, Event& event);

  // Select next pT in downwards evolution.
  double pTnext( Event& event, double pTbegAll, double pTendAll,
    bool isFirstTrial = false, bool doTrialIn = false);

  // Branch event, including accept/reject veto
  bool branch( Event& event, bool isInterleaved=false);

  // Utility to print antenna list; for debug mainly.
  void list() const;

  // Initialize data members for calculation of uncertainty bands.
  // (No Vincia implementation yet.)
  // virtual bool initUncertainties();

  // Tell whether FSR has done a weak emission.
  // (No Vincia implementation yet.)
  // virtual bool getHasWeaklyRadiated() {return hasWeaklyRadiated;}

  // Tell which system was the last processed one.
  int system() const {return iSysWin;}

  // Potential enhancement factor of pTmax scale for hardest emission.
  // Used if limitPTmax = true.
  double enhancePTmax() { return pTmaxFudge; }

  // Provide the pT scale of the last branching in the above shower.
  double pTLastInShower() {return pTLastAcceptedSav;}

  // The following methods for merging not yet implemented in Vincia
  // Event clustered()
  // map<string, double> getStateVariables ()
  //  int , string )
  // bool isTimelike()
  // vector<string> getSplittingName()
  // double getSplittingProb()
  // bool allowedSplitting()
  // vector<int> getRecoilers()
  
  // END OF FUNCTIONS INHERITED FROM PYTHIA 8 FOR INTERFACING

  // -----------------------------------------------------------------------

  // PUBLIC NON-INHERITED MEMBERS

  // Initialise pointers to Vincia objects
  void initVinciaPtrs(Colour* colourPtrIn, VinciaISR* isrPtrIn,
    QEDShower* qedPtrIn,MECs* mecsPtrIn,Resolution* resolutionPtrIn,
    VinciaCommon* vinComPtrIn,VinciaWeights* vinWeightsPtrIn) {
    colourPtr     = colourPtrIn;
    isrPtr        = isrPtrIn;
    qedShowerPtr  = qedPtrIn;
    mecsPtr       = mecsPtrIn;
    resolutionPtr = resolutionPtrIn;
    vinComPtr     = vinComPtrIn;
    weightsPtr    = vinWeightsPtrIn;
  }

  // Initialize Pointers to Antenna sets
  void initAntPtr(AntennaSetFSR* antSetIn) { antSetPtr = antSetIn; }

  // Wrapper function to return a specific antenna inside antenna set
  AntennaFunction* getAnt(int iAnt) { return antSetPtr->getAnt(iAnt); }
  // Wrapper to return all iAntSav that are contained in antSetPtr.
  vector<int> getIant() { return antSetPtr->getIant(); }

  // Print header information (version, settings, parameters, etc)
  void header();

  //Initialize internal and diagnostic histograms
  //void initVinciaHistos();

  //Other info that can be called by Vincia
  // Print final statistics information
  void printInfo( bool pluginCall = false );

  // Print internal and diagnostic histrograms
  void printHistos() {
    // Print histograms in shower class
    map<string,Hist>::iterator iH;
    for (iH=vinciaHistos.begin();iH!=vinciaHistos.end();++iH) {
      string Hname=iH->first;
      if ( vinciaHistos[Hname].getEntries() >= 1) {
        cout<<Hname<<vinciaHistos[Hname]<<endl;
      }
    }
  }

  // Print internal and diagnostic histrograms
  void writeHistos(string fileName="vincia", string suffix="dat") {
    // Print histograms in shower class
    map<string,Hist>::const_iterator iH;
    for (iH=vinciaHistos.begin();iH!=vinciaHistos.end();++iH) {
      string Hname=iH->first;
      if ( vinciaHistos[Hname].getEntries() >= 1) {
        string file = fileName + "-Hist-" + Hname + "." + suffix;
        file = sanitizeFileName(file);
        cout << "Writing " << file << endl;
        iH->second.table(file);
      }
    }
  }

  const Hist& getDiagnosticHistogram( string name ) {
    return vinciaHistos[name];
  }

  /* // Print internal and diagnostic histrograms */
  /* void printHistos();     */
  /* // Print internal and diagnostic histrograms */
  /* void writeHistos();    */
  /* // Allow the user to get a pointer to an internal histogram */
  /* // Note: points to an entry of a map, so must be updated each */
  /* // time it is used. */
  /* Hist* getHistPtr(string histIn);    */
  /* const Hist& getDiagnosticHistogram( string name ); */

  /* // Function to ask if a system is hard */
  /* bool isSysHard(int iSys); */

  // Information.
  // -------------------------------------------------------------------------
  // Get number of systems
  int getNsys() {return nBranchFSR.size();}
  // Get number of branchings in a system (return -1 if no such system)
  int getNbranch(int iSys = -1) {
    // iSys < 0 : return sum over all
    int n = 0;
    if (iSys < 0) {
      for (int i=0; i<(int)nBranchFSR.size(); ++i) n += nBranchFSR[i];
    } else if (iSys < (int)nBranchFSR.size()) n = nBranchFSR[iSys];
    else n = -1;
    return n;
  }
  // Get scale of branchings; use (0,1) for first branching in 1st system.
  // (Could be extended so (i,0) would return starting scale for system i.)
  double getQbranch(int iSys, int iBranch) {
    if (iSys < 4 && iSys >= 0 && iBranch <= 10 && iBranch >= 1)
      return qBranch[iSys][iBranch];
    else return 0.;
  }
  double getPTphys(int iSys, int iBranch) {
    if (iSys < 4 && iSys >= 0 && iBranch <= 10 && iBranch >= 1)
      return pTphys[iSys][iBranch];
    else return 0.;
  }

  void doforceQuit(int nBranchQuitIn){
    allowforceQuit=true;
    nBranchQuit=nBranchQuitIn;
  };

  void setDiagnostics(VinciaDiagnostics* diagnosticsPtrIn){
    diagnosticsPtr=diagnosticsPtrIn;
  }

  // Check event
  bool check(Event &event);

  // Set verbose level
  void setVerbose(int verboseIn) { verbose = verboseIn; }

 private:

  // -----------------------------------------------------------------------

  // Private members

  // Evolution windows
  void initEvolutionWindows(void);
  double getQ2Window(int iWindow,double q2cutoff);
  double getLambda(int nFin,AlphaStrong * aSptr);
  double getkMu2(bool isEmit);
  double getMu2(bool isEmit);

  // Reset (or clear) sizes of all containers
  void clearContainers();

  //Method to set up antennae: called in prepare
  bool getAntennae(int iSys, Event& event);

  // Set starting scale of shower (power vs wimpy) for system iSys
  void setStartScale(int iSys, Event& event);

  // Auxiliary methods to generate trial scales for various shower components
  bool q2NextEmit(const double q2Begin, double q2End);
  bool q2NextSplit(const double q2Begin, double q2End);
  bool q2NextResEmit(const double q2Begin, const double q2End);
  bool q2NextResSplit(const double q2Begin, const double q2End);
  bool q2NextEmitQED(double q2Begin, const double q2End);
  bool q2NextSplitQED(double q2Begin, const double q2End);

  template < class Brancher >
  bool q2NextBranch( vector<Brancher> &brancherVec,
    const map<double, EvolutionWindow> &evWindows,
    const int evType,
    const double q2Begin, const double q2End, bool isEmit);

  bool branchQED(Event& event);

  bool rejectEarly(AntennaFunction* &antFunPtr,bool doMEC);
  double getAntPhys(AntennaFunction* &antFunPtr);
  double pAcceptCalc(double antPhys);
  bool genFullKinematics(int kineMap,Event event,vector<Vec4> &pPost);
  bool acceptTrial(Event& event);
  bool getNewParticles(Event& event, AntennaFunction* antFunPtr,
    vector<Particle> &newParts);
  vector<int> genHelicities(AntennaFunction* antFunPtr);
  bool updateEvent(Event& event,resJunctionInfo & junctionInfoIn);
  void updatePartonSystems();
  void saveEmitter(int iSysIn, Event& event, int i0, int i1);
  void saveResEmitter(int iSys, Event& event, vector<int> allIn,
    unsigned int posResIn, unsigned int posFIn,bool colMode);
  //Create a new res emission brancher
  void saveResSplitter(int iSysIn, Event& event, vector<int> allIn,
    unsigned int posResIn, unsigned int posFIn,bool colMode);
  void saveSplitter(int iSysIn, Event& event, int i0, int i1, bool col2acol);
  template < class Brancher >
  void updateBranchers ( vector<Brancher> &brancherVec,
    map< pair<int, bool>, unsigned int > &lookupBrancher,
    Event& event, int iOld, int iNew );
  template < class Brancher >
  void updateBrancher ( vector<Brancher> &brancherVec,
    map< pair<int, bool>, unsigned int > &lookupBrancher,
    Event& event, int iOld1, int iOld2,
    int iNew1, int iNew2);
  void updateEmitters(Event& event, int iOld, int iNew);
  void updateEmitter(Event& event, int iOld1, int iOld2, int iNew1, int iNew2);
  void updateSplitters(Event& event, int iOld, int iNew);
  void updateSplitter(Event& event, int iOld1, int iOld2, int iNew1, int iNew2,
    bool col2acol);
  void removeSplitter(int iRemove);
  bool updateResBranchers(int iSysRes, Event& event, int iRes);
  void updateResBranchers(int iSysRes, Event& event, vector<int> resSysAll,
    unsigned int posRes, unsigned int posPartner, bool isCol);
  bool updateAntennae(Event& event);
  // Method to update QCD antennae after a QED branching
  bool updateAfterQED(Event& event, int sizeOld);
  void printLookup(map< pair<int, bool>, unsigned int > lookupBrancher,
    string name);
  void printLookup(){
    cout<<endl;
    cout<<"  --------"
        <<"  Brancher lookup maps"
        <<"  -------------------------------------------------------------"
        <<endl;
    printLookup(lookupBrancherRF,"BrancherRF");
    printLookup(lookupSplitterRF,"SplitterRF");
    printLookup(lookupBrancherFF,"BrancherFF");
    printLookup(lookupSplitter,"SplitterFF");
    cout<<"  --------"
        <<"       End lookup     "
        <<"  -------------------------------------------------------------"
        <<endl<<endl;

  };

  vector<double> getHeadroom(int iSys, bool isEmit, double q2Next);
  vector<double> getEnhance(int iSys, bool isEmit, double q2Next);

  // - - - - - - - - - - - - - - - - - - - - - - - - -
  // Private data members

  // Initialized?
  bool isInit; //true when class initialises successfully
  bool isPrepared; //true when prepare executes successfully

  // Beams
  BeamParticle* beamAPtr;
  BeamParticle* beamBPtr;
  double eCMBeamsSav, m2BeamsSav;

  // Main on/off switches
  bool doFSR, doISR, doQED;

  // Parameter setting which kind of 2->4 modifications (if any) are used
  int mode2to4;

  // Shower parameters
  bool helicityShower, sectorShower;
  int evolutionType;
  int evTypeEmit, evTypeSplit, nGluonToQuark;
  double q2CutoffEmit, q2CutoffSplit;
  int nFlavZeroMass;
  // kineMapFFsplit;
  /* double m2thresSplitS, m2thresSplitC, m2thresSplitB, m2thresSplitT; */
  //double q2CutoffResEmit;
  map<int,int> resSystems;
  int kMapResEmit;
  int kMapResSplit;

  // Factorization scale and shower starting settings
  int    pTmaxMatch, pTdampMatch;
  double pTmaxFudge, pT2maxFudge, pT2maxFudgeMPI, pT2dampFudge;

  // AlphaS parameters
  bool useCMW;
  int alphaSorder, renormType;
  double alphaSvalue, alphaSmax, alphaSmuFreeze, alphaSmuMin;
  double aSkMu2Emit, aSkMu2Split;
  // Calculated values
  double mu2freeze, mu2min;

  //map of qmin evolution window
  map<double, EvolutionWindow> evWindowsEmit;
  map<double, EvolutionWindow> evWindowsSplit;

  //Lists of different types of antennae;
  vector<BrancherEmitRF> resEmitters;
  vector<BrancherSplitRF> resSplitters;
  vector<BrancherEmitFF> emitters;
  vector<BrancherSplitFF> splitters;

  //look up resemitter
  //bool switches between R (true) or F (false)
  //NB: multiply res index by sign of colour index involved in branching
  //to avoid a multiply-valued map
  map< pair<int, bool>, unsigned int > lookupBrancherRF;
  map< pair<int, bool>, unsigned int > lookupSplitterRF;
  //lookup emitter
  //bool switches between col and anticol end
  map< pair<int, bool>, unsigned int > lookupBrancherFF;
  //lookup splitter
  //bool switches between splitter and recoiler
  map< pair<int, bool>, unsigned int > lookupSplitter;

  // Current winner
  Brancher* winnerPtr;
  bool winnerQED;
  double q2WinSav, pTLastAcceptedSav;

  // Variables set by branch()
  int iSysWin, iAntWin;
  //index of latest emission (slightly arbritrary for splittings
  //but only used to populate some internal histograms
  int iNewSav;

  // Storage of the post-branching configuration while it is being built.
  vector<Particle> pNew;
  // Total and MEC accept probability
  vector<double> pAccept;

  // Colour Reconnection parameters
  bool doCR, CRjunctions;

  // Uncertainty Bands and Weights
  bool uncertaintyBands;
  double uncertaintyBandsCutoff;

  // Enhancement switches and parameters
  bool enhanceInHard, enhanceInResDec, enhanceInMPI;
  double enhanceAll, enhanceBottom, enhanceCharm, enhanceCutoff;

  // Possibility to allow user veto of emission step.
  bool hasUserHooks, canVetoEmission;

  // - - - - - - - - - - - - - - - - - - - - - - - - -
  // Variables storing characteristics of each system

  // Flags to tell a few basic properties of each parton system
  map<int, bool> isHardSys, isResonanceSys, polarisedSys, doMECsSys;
  map<int, bool> stateChangeSys;
  bool stateChangeLast;

  // Save initial FSR starting scale system by system
  map<int, double> Q2hat;

  // Count the number of branchings in the system
  map<int, int> nBranch, nBranchFSR;

  // Total mass of showering system
  map<int, double> mSystem;

  // Count numbers of quarks and gluons
  map<int, int> nG, nQ, nLep, nGam;

  //save headroom and enhancement factors for each systems
  //for both emission and splitting branchers
  map< pair<int, pair<bool,bool> > , vector<double> > headroomSav;
  map< pair<int, pair<bool,bool> > , vector<double> > enhanceSav;

  //information iabout resonances that participate in junctions
  map<int, bool> hasResJunction;
  map<int, resJunctionInfo> junctionInfo;

  // - - - - - - - - - - - - - - - - - - - - - - - - -
  // Statistics and Histograms

  // Internal histograms
  VinciaDiagnostics* diagnosticsPtr;
  bool doDiagnostics;

  map<string,Hist> vinciaHistos;

  // Bool to know whether we have binned the first QBranch already
  bool firstQBranchBinned;

  // Statistics
  long nTrialsSum;
  vector<long> nTrials, nTrialsAccepted, nFailedVeto, nFailedHull, nFailedKine;
  vector<long> nFailedMass, nFailedCutoff, nClosePSforHQ, nSectorReject;

  // Verbose setting, P>1 warning settings, and version number
  int verbose;
  bool allowforceQuit; //only true if debugging!
  bool forceQuit; //only true if debugging!
  int nBranchQuit;

  double pAcceptFussSav, pAcceptFussPT0;
  bool headerIsPrinted;
  double qBranch[4][11], pTphys[4][11];

  // Counter for numbers of events
  long nAccepted, nSelected;
  int nVetoUserHooks, nFailHadLevel, nCallPythiaNext;

  // - - - - - - - - - - - - - - - - - - - - - - - - -
  // Pointers to VINCIA objects

  AntennaSetFSR*  antSetPtr;
  MECs*           mecsPtr;
  Colour*         colourPtr;
  Resolution*     resolutionPtr;
  VinciaISR*      isrPtr;
  QEDShower*      qedShowerPtr;
  VinciaCommon*   vinComPtr;
  VinciaWeights*  weightsPtr;
  // Pointer to AlphaS instances
  AlphaStrong* aSemitPtr;
  AlphaStrong* aSsplitPtr;

};

}

#endif
