// VinciaISR.h is a part of the PYTHIA event generator.
// Copyright (C) 2019 Torbjorn Sjostrand.
// PYTHIA is licenced under the GNU GPL v2 or later, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// This file contains header information for the VinciaISR class for
// QCD initial-state antenna showers (II and IF), and auxiliary classes.

#ifndef VINCIA_ISR_H
#define VINCIA_ISR_H

#include "Pythia8/SpaceShower.h"

#include "Pythia8/VinciaAntennaFunctions.h"
#include "Pythia8/VinciaCommon.h"
#include "Pythia8/VinciaQED.h"
#include "Pythia8/VinciaWeights.h"
#include "Pythia8/VinciaFSR.h"

namespace Pythia8 {

// Forward declarations
class VinciaFSR;
class MECs;

// Classes contained in this file
class TrialGeneratorISR;
class BranchelementalISR;
class VinciaISR;

// ===================================================================

// Base class for Initial-State TrialGenerators
// Note: base class is coded for a soft-eikonal trial function

class TrialGeneratorISR {

 public:

  // -----------------------------------------------------------------
  // Constructor and pointer initialization (non-virtual)
  // -----------------------------------------------------------------

  // Constructor
  TrialGeneratorISR() : isInitPtr(false), isInit(false) {};

  // Destructor
  virtual ~TrialGeneratorISR() {};
  
  // Init Pointers
  void initPtr(Rndm* rndmPtrIn, Settings* settingsPtrIn, Info* infoPtrIn);

  // -----------------------------------------------------------------
  // Virtual functions
  // Defining the particular kind of trial generator
  // -----------------------------------------------------------------

  // Name of trial generator
  virtual string name() {
    return "TrialGeneratorISR";
  }

  // Initialize
  virtual void init(double mcIn, double mbIn);

  // Trial antenna function
  virtual double aTrial(double saj, double sjb, double sAB);

  // Evolution scale
  virtual double getQ2(double saj, double sjb, double sAB) {
    return (saj*sjb/(saj+sjb+sAB));
  }
  virtual double getQ2max(double sAB, double, double) {
    return (0.25*pow2(shhSav-sAB)/shhSav);
  }

  // Generate new Q value, with first-order running alphaS
  virtual double genQ2run(double q2old, double sAB, double zMin, double zMax,
    double colFac, double PDFratio, double b0, double kR, double Lambda,
    double eA, double eB, double headroomFac=1.0, double enhanceFac=1.0);

  // Generate new Q value, with constant trial alphaS
  virtual double genQ2(double q2old, double sAB, double zMin, double zMax,
    double colFac, double alphaSvalue, double PDFratio, double eA, double eB,
    double headroomFac=1.0, double enhanceFac=1.0);

  // Generate new Q value, with running of the PDFs towards the mass threshold
  virtual double genQ2thres(double /*q2old*/, double /*sAB*/, double /*zMin*/,
    double /*zMax*/, double /*colFac*/, double /*alphaSvalue*/,
    double /*PDFratio*/, int /*idA*/, int /*idB*/, double /*eA*/,
    double /*eB*/, bool /*useMpdf*/, double /*headroomFac=1.0*/,
    double /*enhanceFac=1.0*/ ) {
    return 0.0;
  }

  // Generate a new zeta value in [zMin,zMax]
  virtual double genZ(double zMin, double zMax);

  // The zeta integral
  virtual double getIz(double zMin, double zMax);

  // The zeta boundaries, for a given value of the evolution scale
  virtual double getZmin(double Qt2, double sAB, double eA, double eBeamUsed);
  virtual double getZmax(double Qt2, double sAB, double eA, double eBeamUsed);

  // Inverse transforms
  // To obtain saj and sjb from Qt2 and zeta
  virtual double getS1j(double Qt2, double zeta, double sAB);
  virtual double getSj2(double Qt2, double zeta, double sAB);

  // Compute trial PDF ratio
  virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr,
    int iSys, int idA, int idB, double eA, double eB,
    double Qt2A, double Qt2B);

  // Return last trial PDF ratio
  virtual double getTrialPDFratio() { return trialPDFratioSav; }

  // -----------------------------------------------------------------
  // Non-Virtual functions
  // -----------------------------------------------------------------

  bool checkInit();

  int trialFlav() { return trialFlavSav; }

 protected:

  // Pointers
  Rndm* rndmPtr;
  Settings* settingsPtr;
  Info* infoPtr;

  double TINYPDFtrial;

  // Use m or pT evolution for collinear singularities
  bool useMevolSav;

  // s hadron hadron
  double shhSav;

  // For conversion trial generators
  int trialFlavSav;
  int nGtoQISRSav;

  // Masses
  double mbSav;
  double mcSav;

  // Saved trial PDF ratio
  double trialPDFratioSav;

 private:

  // Status
  bool isInitPtr, isInit;

  // Verbose level
  int verbose;
};

// ===================================================================

// Soft-eikonal trial function for II (clone of base class)

class TrialIISoft : public TrialGeneratorISR {

 public:

  // Name of trial generator
  virtual string name() {
    return "TrialIISoft";
  }

};

// ===================================================================

// A collinear trial function for II (derived base class)

class TrialIIGCollA : public TrialGeneratorISR {

 public:

  // Name of trial generator
  virtual string name() {
    return "TrialIIGCollA";
  }

  // Trial antenna function
  virtual double aTrial(double saj, double sjb, double sAB);

  // Evolution scale
  virtual double getQ2(double saj, double sjb, double sAB) {
    return (saj*sjb/(saj+sjb+sAB));
  }
  virtual double getQ2max(double sAB, double, double) {
    return (0.25*pow2(shhSav-sAB)/shhSav);
  }

  // Generate a new Q value, with first-order running alphaS
  virtual double genQ2run(double q2old, double sAB, double zMin, double zMax,
    double colFac, double PDFratio, double b0, double kR, double Lambda,
    double eA, double eB, double headroomFac=1.0, double enhanceFac=1.0);

  // Generate a new Q value, with constant trial alphaS
  virtual double genQ2(double q2old, double sAB, double zMin, double zMax,
    double colFac, double alphaSvalue, double PDFratio,
    double eA, double eB, double headroomFac=1.0, double enhanceFac=1.0);

  // Generate a new zeta value in [zMin,zMax]
  virtual double genZ(double zMin, double zMax);

  // The zeta integral
  virtual double getIz(double zMin, double zMax);

  // The zeta boundaries, for a given value of the evolution scale
  virtual double getZmin(double Qt2, double sAB, double eA, double eBeamUsed);
  virtual double getZmax(double Qt2, double sAB, double eA, double eBeamUsed);

  // Inverse transforms
  // To obtain saj and sjb from Qt2 and zeta
  virtual double getS1j(double Qt2, double zeta, double sAB);
  virtual double getSj2(double Qt2, double zeta, double sAB);

  // Trial PDF ratio
  virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr,
    int iSys, int idA, int idB, double eA, double eB,
    double Qt2A, double Qt2B);

};


// ===================================================================

// B collinear trial function for II (derived base class)

class TrialIIGCollB : public TrialIIGCollA {

 public:

  // Name of trial generator
  virtual string name() {
    return "TrialIIGCollB";
  }

  // Trial antenna function
  virtual double aTrial(double saj, double sjb, double sAB) {
    return TrialIIGCollA::aTrial(sjb, saj, sAB);
  }

  // Inverse transforms
  // To obtain saj and sjb from Qt2 and zeta
  virtual double getS1j(double Qt2, double zeta, double sAB) {
    return TrialIIGCollA::getSj2(Qt2, zeta, sAB);
  }
  virtual double getSj2(double Qt2, double zeta, double sAB) {
    return TrialIIGCollA::getS1j(Qt2, zeta, sAB);
  }

  // Trial PDF ratio
  virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr,
    int iSys, int idA, int idB, double eA, double eB,
    double Qt2A, double Qt2B) {
    return TrialIIGCollA::trialPDFratio(beamBPtr, beamAPtr, iSys,
      idB, idA, eB, eA, Qt2B, Qt2A);
  }
  
};


// ===================================================================

// A splitting trial function for II (derived base class), q<-gqbar

class TrialIISplitA : public TrialGeneratorISR {

 public:

  // Name of trial generator
  virtual string name() {
    return "TrialIISplitA";
  }

  // Trial antenna function
  virtual double aTrial(double saj, double sjb, double sAB);

  // Evolution scale
  virtual double getQ2(double saj, double sjb, double sAB) {
    return ( (useMevolSav) ? saj : saj*sjb/(saj+sjb+sAB) );
  }
  virtual double getQ2max(double sAB, double, double) {
    return ( (useMevolSav) ? shhSav-sAB : 0.25*pow2(shhSav-sAB)/shhSav );
  }

  // Generate new Q value, with first-order running alphaS
  virtual double genQ2run(double q2old, double sAB, double zMin, double zMax,
    double colFac, double PDFratio, double b0, double kR, double Lambda,
    double eA, double eB, double headroomFac=1.0, double enhanceFac=1.0);

  // Generate new Q value, with constant trial alphaS
  virtual double genQ2(double q2old, double sAB, double zMin, double zMax,
    double colFac, double alphaSvalue, double PDFratio,
    double eA, double eB, double headroomFac=1.0, double enhanceFac=1.0);

  // Generate new Q value, with running of the PDFs towards the mass threshold
  virtual double genQ2thres(double q2old, double sAB,
    double zMin, double zMax, double colFac, double alphaSvalue,
    double PDFratio, int idA, int idB, double eA, double eB, bool useMpdf,
    double headroomFac=1.0, double enhanceFac=1.0);

  // Generate a new zeta value in [zMin,zMax]
  virtual double genZ(double zMin, double zMax);

  // The zeta integral
  virtual double getIz(double zMin, double zMax);

  // The zeta boundaries, for a given value of the evolution scale
  virtual double getZmin(double Qt2, double sAB, double eA, double eBeamUsed);
  virtual double getZmax(double Qt2, double sAB, double eA, double eBeamUsed);

  // Inverse transforms
  // To obtain saj and sjb from Qt2 and zeta
  virtual double getS1j(double Qt2, double zeta, double sAB);
  virtual double getSj2(double Qt2, double zeta, double sAB);

  // Trial PDF ratio
  virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr,
    int iSys, int idA, int idB, double eA, double eB,
    double Qt2A, double Qt2B);

};

// ===================================================================

// B splitting trial function for II (derived base class), q<-gqbar

class TrialIISplitB : public TrialIISplitA {

 public:

  // Name of trial generator
  virtual string name() {
    return "TrialIISplitB";
  }

  // Trial antenna function
  virtual double aTrial(double saj, double sjb, double sAB) {
    return TrialIISplitA::aTrial(sjb, saj, sAB);
  }

  // Evolution scale
  virtual double getQ2(double saj, double sjb, double sAB) {
    return TrialIISplitA::getQ2(sjb, saj, sAB);
  }

  // Generate new Q value, with first-order running alphaS
  virtual double genQ2run(double q2old, double sAB, double zMin, double zMax,
    double colFac, double PDFratio, double b0, double kR, double Lambda,
    double eA, double eB, double headroomFac=1.0, double enhanceFac=1.0) {
    return TrialIISplitA::genQ2run(q2old, sAB, zMin, zMax, colFac, PDFratio,
      b0, kR, Lambda, eB, eA, headroomFac, enhanceFac);
  }

  // Generate new Q value, with constant trial alphaS
  virtual double genQ2(double q2old, double sAB, double zMin, double zMax,
    double colFac, double alphaSvalue, double PDFratio,
    double eA, double eB, double headroomFac=1.0, double enhanceFac=1.0) {
    return TrialIISplitA::genQ2(q2old, sAB, zMin, zMax, colFac, alphaSvalue,
      PDFratio, eB, eA, headroomFac, enhanceFac);
  }

  // Generate new Q value, with running of the PDFs towards the mass threshold
  virtual double genQ2thres(double q2old, double sAB,
    double zMin, double zMax, double colFac, double alphaSvalue,
    double PDFratio, int idA, int idB, double eA, double eB, bool useMpdf,
    double headroomFac=1.0, double enhanceFac=1.0) {
    return TrialIISplitA::genQ2thres(q2old, sAB, zMin, zMax, colFac,
      alphaSvalue, PDFratio, idB, idA, eB, eA, useMpdf, headroomFac,
      enhanceFac);
  }

  // Inverse transforms
  // To obtain saj and sjb from Qt2 and zeta
  virtual double getS1j(double Qt2, double zeta, double sAB) {
    return TrialIISplitA::getSj2(Qt2, zeta, sAB);
  }
  virtual double getSj2(double Qt2, double zeta, double sAB) {
    return TrialIISplitA::getS1j(Qt2, zeta, sAB);
  }

  // Trial PDF ratio
  virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr,
    int iSys, int idA, int idB, double eA, double eB,
    double Qt2A, double Qt2B) {
    return TrialIISplitA::trialPDFratio(beamBPtr, beamAPtr, iSys,
      idB, idA, eB, eA, Qt2B, Qt2A);
  }

};

// ===================================================================

// A conversion trial function for II (derived base class), g<-qqbar

class TrialIIConvA : public TrialGeneratorISR {

 public:

  // Name of trial generator
  virtual string name() {
    return "TrialIIConvA";
  }

  // Trial antenna function
  virtual double aTrial(double saj, double sjb, double sAB);

  // Evolution scale
  virtual double getQ2(double saj, double sjb, double sAB) {
    return ( (useMevolSav) ? saj : (saj*sjb/(saj+sjb+sAB)) );
  }
  virtual double getQ2max(double sAB, double, double) {
    return ( (useMevolSav) ? (shhSav-sAB) : 0.25*pow2(shhSav-sAB)/shhSav );
  }

  // Generate a new Q value, with first-order running alphaS
  virtual double genQ2run(double q2old, double sAB, double zMin, double zMax,
    double colFac, double PDFratio, double b0, double kR, double Lambda,
    double eA, double eB, double headroomFac=1.0, double enhanceFac=1.0);

  // Generate a new Q value, with constant trial alphaS
  virtual double genQ2(double q2old, double sAB, double zMin, double zMax,
    double colFac, double alphaSvalue, double PDFratio,
    double eA, double eB, double headroomFac=1.0, double enhanceFac=1.0);

  // Generate a new zeta value in [zMin,zMax]
  virtual double genZ(double zMin, double zMax);

  // The zeta integral
  virtual double getIz(double zMin, double zMax);

  // The zeta boundaries, for a given value of the evolution scale
  virtual double getZmin(double Qt2, double sAB, double eA, double eBeamUsed);
  virtual double getZmax(double Qt2, double sAB, double eA, double eBeamUsed);

  // Inverse transforms
  // To obtain saj and sjb from Qt2 and zeta
  virtual double getS1j(double Qt2, double zeta, double sAB);
  virtual double getSj2(double Qt2, double zeta, double sAB);

  // Trial PDF ratio
  virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr,
    int iSys, int idA, int idB, double eA, double eB,
    double Qt2A, double Qt2B);

 private:

};

// ===================================================================

// B conversion trial function for II (derived base class), g<-qqbar

class TrialIIConvB : public TrialIIConvA {

 public:

  // Name of trial generator
  virtual string name() {
    return "TrialIIConvB";
  }

  // Trial antenna function
  virtual double aTrial(double saj, double sjb, double sAB) {
    return TrialIIConvA::aTrial(sjb, saj, sAB);
  }

  // Evolution scale
  virtual double getQ2(double saj, double sjb, double sAB) {
    return TrialIIConvA::getQ2(sjb, saj, sAB);
  }

  // Generate a new Q value, with first-order running alphaS
  virtual double genQ2run(double q2old, double sAB, double zMin, double zMax,
    double colFac, double PDFratio, double b0, double kR, double Lambda,
    double eA, double eB, double headroomFac=1.0, double enhanceFac=1.0) {
    return TrialIIConvA::genQ2run(q2old, sAB, zMin, zMax, colFac, PDFratio,
      b0, kR, Lambda, eB, eA, headroomFac, enhanceFac);
  }

  // Generate a new Q value, with constant trial alphaS
  virtual double genQ2(double q2old, double sAB, double zMin, double zMax,
    double colFac, double alphaSvalue, double PDFratio,
    double eA, double eB, double headroomFac=1.0, double enhanceFac=1.0) {
    return TrialIIConvA::genQ2(q2old, sAB, zMin, zMax, colFac, alphaSvalue,
      PDFratio, eB, eA, headroomFac, enhanceFac);
  }

  // Inverse transforms: to obtain saj and sjb from Qt2 and zeta
  virtual double getS1j(double Qt2, double zeta, double sAB) {
    return TrialIIConvA::getSj2(Qt2, zeta, sAB);
  }
  virtual double getSj2(double Qt2, double zeta, double sAB) {
    return TrialIIConvA::getS1j(Qt2, zeta, sAB);
  }

  // Trial PDF ratio
  virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr,
    int iSys, int idA, int idB, double eA, double eB,
    double Qt2A, double Qt2B) {
    return TrialIIConvA::trialPDFratio(beamBPtr, beamAPtr, iSys,
      idB, idA, eB, eA, Qt2B, Qt2A);
  }

};

// ===================================================================

// Soft-eikonal trial function for IF (derived base class)

class TrialIFSoft : public TrialGeneratorISR {

 public:

  // Name of trial generator
  virtual string name() {
    return "TrialIFSoft";
  }

  // Trial antenna function
  virtual double aTrial(double saj, double sjk, double sAK);

  // Evolution scale
  virtual double getQ2(double saj, double sjk, double sAK) {
    return (saj*sjk/(sAK+sjk));
  }
  virtual double getQ2max(double sAK, double eA, double eBeamUsed) {
    double eAmax = ( (sqrt(shhSav)/2.0) - (eBeamUsed-eA) );
    return (sAK*(eAmax-eA)/eA);
  }

  // Generate a new Q value, with first-order running alphaS
  virtual double genQ2run(double q2old, double sAK, double zMin, double zMax,
    double colFac, double PDFratio, double b0, double kR, double Lambda,
    double eA, double eK, double headroomFac=1.0, double enhanceFac=1.0);

  // Generate a new Q value, with constant trial alphaS
  virtual double genQ2(double q2old, double sAK, double zMin, double zMax,
    double colFac, double alphaSvalue, double PDFratio,
    double eA, double eK, double headroomFac=1.0, double enhanceFac=1.0);

  // Generate a new zeta value in [zMin,zMax]
  virtual double genZ(double zMin, double zMax);

  // The zeta integral
  virtual double getIz(double zMin, double zMax);

  // The zeta boundaries, for a given value of the evolution scale
  virtual double getZmin(double Qt2, double sAK, double eA, double eBeamUsed);
  virtual double getZmax(double Qt2, double sAK, double eA, double eBeamUsed);

  // Inverse transforms
  // To obtain saj and sjk from Qt2 and zeta
  virtual double getS1j(double Qt2, double zeta, double sAK);
  virtual double getSj2(double Qt2, double zeta, double sAK);

  // Trial PDF ratio
  virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr,
    int iSys, int idA, int idK, double eA, double eK,
    double Qt2A, double Qt2B);

};

// ===================================================================

// Specialised Soft-eikonal trial function for IF (derived base class)
// when initial-state parton is a valence quark.

class TrialVFSoft : public TrialIFSoft {

 public:

  // Name of trial generator
  virtual string name() {
    return "TrialVFSoft";
  }

  // This generator uses f <= const overestimate so needs x-factor on trial
  virtual double aTrial(double saj, double sjk, double sAK);

  // This generator uses a different zeta integral 
  virtual double getIz(double zMin, double zMax);
  virtual double genZ(double zMin, double zMax);

};

// ===================================================================

// A gluon collinear trial function for IF (derived base class)

class TrialIFGCollA : public TrialGeneratorISR {

 public:

  // Name of trial generator
  virtual string name() {
    return "TrialIFGCollA";
  }

  // Trial antenna function
  virtual double aTrial(double saj, double sjk, double sAK);

  // Evolution scale
  virtual double getQ2(double saj, double sjk, double sAK) {
    return (saj*sjk/(sAK+sjk));
  }
  virtual double getQ2max(double sAK, double eA, double eBeamUsed) {
    double eAmax = ( (sqrt(shhSav)/2.0) - (eBeamUsed-eA) );
    return (sAK*(eAmax-eA)/eA);
  }

  // Generate a new Q value, with first-order running alphaS
  virtual double genQ2run(double q2old, double sAK, double zMin, double zMax,
    double colFac, double PDFratio, double b0, double kR, double Lambda,
    double eA, double eK, double headroomFac=1.0, double enhanceFac=1.0);

  // Generate a new Q value, with constant trial alphaS
  virtual double genQ2(double q2old, double sAK, double zMin, double zMax,
    double colFac, double alphaSvalue, double PDFratio,
    double eA, double eK, double headroomFac=1.0, double enhanceFac=1.0);

  // Generate a new zeta value in [zMin,zMax]
  virtual double genZ(double zMin, double zMax);

  // The zeta integral
  virtual double getIz(double zMin, double zMax);

  // The zeta boundaries, for a given value of the evolution scale
  virtual double getZmin(double Qt2, double sAK, double eA, double eBeamUsed);
  virtual double getZmax(double Qt2, double sAK, double eA, double eBeamUsed);

  // Inverse transforms
  // To obtain saj and sjk from Qt2 and zeta
  virtual double getS1j(double Qt2, double zeta, double sAK);
  virtual double getSj2(double Qt2, double zeta, double sAK);

  // Trial PDF ratio
  virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr,
    int iSys, int idA, int idK, double eA, double eK,
    double Qt2A, double Qt2B);

};

// ===================================================================

// A splitting trial function for IF (derived base class), q<-gqbar

class TrialIFSplitA : public TrialGeneratorISR {

 public:

  // Name of trial generator
  virtual string name() {
    return "TrialIFSplitA";
  }

  // Trial antenna function
  virtual double aTrial(double saj, double sjk, double sAK);

  // Evolution scale
  virtual double getQ2(double saj, double sjk, double sAK) {
    return ( (useMevolSav) ? saj : saj*sjk/(sAK+sjk) );
  }
  virtual double getQ2max(double sAK, double eA, double eBeamUsed) {
    double xA    = eA/(sqrt(shhSav)/2.0);
    double eAmax = ( (sqrt(shhSav)/2.0) - (eBeamUsed-eA) );
    return ( (useMevolSav) ? sAK/xA : sAK*(eAmax-eA)/eA );
  }

  // Generate new Q value, with first-order running alphaS
  virtual double genQ2run(double q2old, double sAK, double zMin, double zMax,
    double colFac, double PDFratio, double b0, double kR, double Lambda,
    double eA, double eK, double headroomFac=1.0, double enhanceFac=1.0);

  // Generate new Q value, with constant trial alphaS
  virtual double genQ2(double q2old, double sAK, double zMin, double zMax,
    double colFac, double alphaSvalue, double PDFratio,
    double eA, double eK, double headroomFac=1.0, double enhanceFac=1.0);

  // Generate new Q value, with running of the PDFs towards the mass threshold
  virtual double genQ2thres(double q2old, double sAK,
    double zMin, double zMax, double colFac, double alphaSvalue,
    double PDFratio, int idA, int idK, double eA, double eK, bool useMpdf,
    double headroomFac=1.0, double enhanceFac=1.0);

  // The zeta integral
  virtual double getIz(double zMin, double zMax);

  // Generate a new zeta value in [zMin,zMax]
  virtual double genZ(double zMin, double zMax);

  // The zeta boundaries, for a given value of the evolution scale
  virtual double getZmin(double Qt2, double sAK, double eA, double eBeamUsed);
  virtual double getZmax(double Qt2, double sAK, double eA, double eBeamUsed);

  // Inverse transforms
  // To obtain saj and sjk from Qt2 and zeta
  virtual double getS1j(double Qt2, double zeta, double sAK);
  virtual double getSj2(double Qt2, double zeta, double sAK);

  // Trial PDF ratio
  virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr,
    int iSys, int idA, int idK, double eA, double eK,
    double Qt2A, double Qt2B);

};

// ===================================================================

// K splitting trial function for IF (derived base class), g<-qqbar

class TrialIFSplitK : public TrialGeneratorISR {

 public:

  // Name of trial generator
  virtual string name() {
    return "TrialIFSplitK";
  }

  // Trial antenna function
  virtual double aTrial(double saj, double sjk, double sAK);

  // Evolution scale
  virtual double getQ2(double saj, double sjk, double sAK) {
    return ( (useMevolSav) ? sjk : saj*sjk/(sAK+sjk) );
  }
  virtual double getQ2max(double sAK, double eA, double eBeamUsed) {
    double xA    = eA/(sqrt(shhSav)/2.0);
    double eAmax = ( (sqrt(shhSav)/2.0) - (eBeamUsed-eA) );
    return ( (useMevolSav) ? sAK*(1.0-xA)/xA : sAK*(eAmax-eA)/eA );
  }

  // Generate a new Q value, with first-order running alphaS
  virtual double genQ2run(double q2old, double sAK, double zMin, double zMax,
    double colFac, double PDFratio, double b0, double kR, double Lambda,
    double eA, double eK, double headroomFac=1.0, double enhanceFac=1.0);

  // Generate a new Q value, with constant trial alphaS
  virtual double genQ2(double q2old, double sAK, double zMin, double zMax,
    double colFac, double alphaSvalue, double PDFratio,
    double eA, double eK, double headroomFac=1.0, double enhanceFac=1.0);

  // Generate a new zeta value in [zMin,zMax]
  virtual double genZ(double zMin, double zMax);

  // The zeta integral
  virtual double getIz(double zMin, double zMax);

  // The zeta boundaries, for a given value of the evolution scale
  virtual double getZmin(double Qt2, double sAK, double eA, double eBeamUsed);
  virtual double getZmax(double Qt2, double sAK, double eA, double eBeamUsed);

  // Inverse transforms
  // To obtain saj and sjk from Qt2 and zeta
  virtual double getS1j(double Qt2, double zeta, double sAK);
  virtual double getSj2(double Qt2, double zeta, double sAK);

  // Trial PDF ratio
  virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr,
    int iSys, int idA, int idK, double eA, double eK,
    double Qt2A, double Qt2B);

};

// ===================================================================

// A conversion trial function for IF (derived base class), g<-qqbar

class TrialIFConvA : public TrialGeneratorISR {

 public:

  // Name of trial generator
  virtual string name() {
    return "TrialIFConvA";
  }

  // Trial antenna function
  virtual double aTrial(double saj, double sjk, double sAK);

  // Evolution scale
  virtual double getQ2(double saj, double sjk, double sAK) {
    return ( (useMevolSav) ? saj : saj*sjk/(sAK+sjk) );
  }
  virtual double getQ2max(double sAK, double eA, double eBeamUsed) {
    double xA    = eA/(sqrt(shhSav)/2.0);
    double eAmax = ( (sqrt(shhSav)/2.0) - (eBeamUsed-eA) );
    return ( (useMevolSav) ? sAK/xA : sAK*(eAmax-eA)/eA );
  }

  // Generate a new Q value, with first-order running alphaS
  virtual double genQ2run(double q2old, double sAK, double zMin, double zMax,
    double colFac, double PDFratio, double b0, double kR, double Lambda,
    double eA, double eK, double headroomFac=1.0, double enhanceFac=1.0);

  // Generate a new Q value, with constant trial alphaS
  virtual double genQ2(double q2old, double sAK, double zMin, double zMax,
    double colFac, double alphaSvalue, double PDFratio,
    double eA, double eK, double headroomFac=1.0, double enhanceFac=1.0);

  // Generate a new zeta value in [zMin,zMax]
  virtual double genZ(double zMin, double zMax);

  // The zeta integral
  virtual double getIz(double zMin, double zMax);

  // The zeta boundaries, for a given value of the evolution scale
  virtual double getZmin(double Qt2, double sAK, double eA, double eBeamUsed);
  virtual double getZmax(double Qt2, double sAK, double eA, double eBeamUsed);

  // Inverse transforms
  // To obtain saj and sjk from Qt2 and zeta
  virtual double getS1j(double Qt2, double zeta, double sAK);
  virtual double getSj2(double Qt2, double zeta, double sAK);

  // Trial PDF ratio
  virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr,
    int iSys, int idA, int idK, double eA, double eK,
    double Qt2A, double Qt2B);

};

//==========================================================================

// The BranchelementalISR class:
// Container for 2->3 trial branchings

// Input: i1In     carries colour (or incoming anticolour)
//        i2In     carries anticolour (or incoming colour)
//        colIn    colour tag
//        isVal1In (true if i1In is incoming and is a valence parton on its
//                  side/system)
//        isVal2In (true if i2In is incoming and is a valence parton on its
//                  side/system)
// Internal storage for II:
//        i1sav: whichever of i1In, i2In has positive pz
//        i2sav: whichever of i1In, i2In has negative pz
//        is1A : always true (i1sav always has positive pz)
//        isII : true
// Internal storage for IF:
//        i1sav: whichever of i1In, i2In is the initial-state leg
//        i2sav: whichever of i1In, i2In is the final-state leg
//        is1A : true if i1sav has positive pz, false if it has negative pz
//        isII : false

class BranchelementalISR {

 public:

  // Constructors
  BranchelementalISR() : system(0) {}
  BranchelementalISR(int iSysIn, Event& event, int iOld1In,
    int iOld2In, int colIn, bool isVal1In, bool isVal2In) {
    reset(iSysIn, event, iOld1In, iOld2In, colIn, isVal1In, isVal2In); };

  // Main method to initialise/reset a branchelemental 
  void reset(int iSysIn, Event& event, int i1In, int i2In, int colIn,
    bool isVal1In, bool isVal2In);

  // Antenna Mass (negative if spacelike virtual)
  double mAnt() const { return mAntSav; }
  double m2Ant() const { return m2AntSav; }
  double sAnt() const { return sAntSav; }
  // Dot products of daughters
  double s12() const { return 2*new1.p()*new2.p(); }
  double s23() const { return 2*new2.p()*new3.p(); }
  double s13() const { return 2*new1.p()*new3.p(); }

  // Is this an II or an IF type?
  bool isII() const { return isIIsav; }
  // Is 1 a side A (p+) guy? (need to know for pdfs in IF)
  bool is1A() const { return is1Asav; }
  // Valence?
  bool isVal1() const {return isVal1sav;}
  bool isVal2() const {return isVal2sav;}

  //
  int colType1() const { return colType1sav; }
  int colType2() const { return colType2sav; }
  int col() const { return colSav; }

  int geti1(){  return i1sav; }
  int geti2(){  return i2sav; }
  int getSystem(){ return system;}

  // Function to reset all trial generators for this branchelemental
  void clearTrialGenerators();

  // Add a trial generator to this branchelemental. Return name of generator.
  void addTrialGenerator(int iAntPhysIn, bool swapIn,
    TrialGeneratorISR* trialGenPtrIn);

  // Add to and get rescue levels
  void addRescue(int iTrial) { nShouldRescue[iTrial]++; }
  int getNshouldRescue(int iTrial) { return nShouldRescue[iTrial]; }
  void resetRescue() {
    for (int i=0; i<(int)nShouldRescue.size(); i++) nShouldRescue[i] = 0;
  }

  // Function to return number of trial generators for this antenna
  int nTrialGenerators() const { return trialGenPtrsSav.size(); }

  // Save a generated trial branching
  void saveTrial(int iTrial, double qOld, double qTrial, double zMin,
    double zMax, double colFac,double alphaEff, double pdfRatio, int trialFlav,
    double extraMpdf, double headroom = 1.0, double enhanceFac = 1.0);

  // Add the physical pdf ratio
  void addPDF(int iTrial,double pdfRatio) {
    physPDFratioSav[iTrial] = pdfRatio;
  }

  // Generate invariants for saved branching
  bool genTrialInvariants(double& s1j, double& sj2,
    double eBeamUsed, int iTrial = -1) {
    // Automatically determine which trial function to use if -1 input
    int iGen = iTrial;
    if (iGen == -1) iGen = getTrialIndex();
    if (iGen <= -1) return false;
    double z = trialGenPtrsSav[iGen]->genZ(zMinSav[iGen],zMaxSav[iGen]);
    // Check physical phase space (note: this only checks massless hull)
    // (Use absolute z value since negative z values are used to indicate
    //  swapped invariants for mD ordering)
    double Q2E = pow2(scaleSav[iGen]);
    if ( abs(z) < trialGenPtrsSav[iGen]->getZmin(Q2E,sAntSav,e1sav,eBeamUsed)
      || abs(z) > trialGenPtrsSav[iGen]->getZmax(Q2E,sAntSav,e1sav,eBeamUsed) )
      return false;
    // Convert to s1j, sj2
    s1j = trialGenPtrsSav[iGen]->getS1j(Q2E,z,sAntSav);
    sj2 = trialGenPtrsSav[iGen]->getSj2(Q2E,z,sAntSav);
    return true;
  }

  // Get trial function index of winner
  int getTrialIndex() const {
    double qMax = 0.0;
    int iMax    = -1;
    for (int i = 0; i < int(scaleSav.size()); ++i) {
      if (hasSavedTrial[i]) {
        double qSav = scaleSav[i];
        if (qSav > qMax) {
          qMax = qSav;
          iMax = i;
        }
      }
    }
    return iMax;
  }

  // Check if a saved trial exists for a particular trialGenerator
  bool hasTrial(int iTrial) const {
    if (iTrial < int(hasSavedTrial.size())) return hasSavedTrial[iTrial];
    else return false;
  }

  // Get whether physical antenna is swapped
  bool getIsSwapped(int iTrial=-1) const {
    if (iTrial <= -1) iTrial = getTrialIndex();
    return isSwappedSav[iTrial];
  }

  // Get physical antenna function index of winner
  int getPhysIndex(int iTrial=-1) const {
    if (iTrial <= -1) iTrial = getTrialIndex();
    return iAntPhysSav[iTrial];
  }

  // Get scale for a specific saved trial
  double getTrialScale(int iTrial) const {
    if (iTrial < int(scaleSav.size())) return scaleSav[iTrial];
    else return -1.0;
  }

  // Get scale of winner
  double getTrialScale() const {
    double qMax = 0.0;
    for (int i = 0; i < int(scaleSav.size()); ++i) {
      if (hasSavedTrial[i]) qMax = max(qMax,scaleSav[i]);
      else printOut("getTrialScale","Error! "
        "not all trials have saved scales");
    }
    return qMax;
  }

  // Get colour factor
  double getColFac(int iTrial=-1) {
    if (iTrial <= -1) iTrial = getTrialIndex();
    return colFacSav[iTrial];
  }

  // Get headroom factor
  double getHeadroomFac(int iTrial=-1) {
    if (iTrial <= -1) iTrial = getTrialIndex();
    return headroomSav[iTrial];
  }

  // Get headroom factor
  double getEnhanceFac(int iTrial=-1) {
    if (iTrial <= -1) iTrial = getTrialIndex();
    return enhanceFacSav[iTrial];
  }

  // Get alpha S
  double getAlphaTrial(int iTrial=-1) {
    if (iTrial <= -1) iTrial = getTrialIndex();
    return alphaSav[iTrial];
  }

  // Get pdf ratio
  double getPDFratioTrial(int iTrial=-1) {
    if (iTrial <= -1) iTrial = getTrialIndex();
    return trialPDFratioSav[iTrial];
  }

  // For gluon conversions, get ID of quark flavour to convert to.
  int getTrialFlav(int iTrial=-1) {
    if (iTrial <= -1) iTrial = getTrialIndex();
    return trialFlavSav[iTrial];
  }

  // Get pdf ratio
  double getPDFratioPhys(int iTrial=-1) {
    if (iTrial <= -1) iTrial = getTrialIndex();
    return physPDFratioSav[iTrial];
  }

  // Get the extra factor when getting rid of the heavy quarks
  double getExtraMassPDFfactor(int iTrial=-1) {
    if (iTrial <= -1) iTrial = getTrialIndex();
    return extraMassPDFfactorSav[iTrial];
  }

  // Flag to set if a saved trial should be ignored and a new one generated
  // Default value -1 : force all trials to renew
  void renewTrial(int iTrial = -1) {
    if (iTrial >= 0) hasSavedTrial[iTrial] = false;
    else {
      for (iTrial = 0; iTrial < int(hasSavedTrial.size()); ++iTrial)
        hasSavedTrial[iTrial] = false;
    }
  }

  // List function
  void list(bool header=false, bool footer=false) const;

  // Data storage members
  int i1sav, i2sav, id1sav, id2sav, colType1sav, colType2sav, h1sav, h2sav;
  double e1sav, e2sav;
  bool isVal1sav, isVal2sav, isIIsav, is1Asav;
  Particle new1, new2, new3;
  // Colour, not obvious, since for eg gg->H we have two II antennae
  int colSav;
  // System and counter for vetos
  int system, nVeto, nHull, nHadr;
  // Do we have to force a splitting (heavy quarks)
  bool forceSplittingSav;

  // Trial Generators and properties of saved trials
  vector<TrialGeneratorISR*> trialGenPtrsSav;
  vector<double> zMinSav, zMaxSav, colFacSav, alphaSav;
  vector<double> physPDFratioSav, trialPDFratioSav;
  vector<double> extraMassPDFfactorSav;
  vector<double> scaleSav, scaleOldSav, headroomSav, enhanceFacSav;
  vector<bool> hasSavedTrial, isSwappedSav;
  vector<int> iAntPhysSav, nShouldRescue, trialFlavSav;
  // Note: isSwapped = true for II means physical antenna function is
  // coded for side A but trial generator is for side B. 
  // For IF, is1A = true for 1 being on side A, false for 1 being on side B

 private:

  // Saved antenna invariant mass value
  double m2AntSav, mAntSav, sAntSav;

};

//==========================================================================

// The VinciaISR class
// Main shower class for initial-state (II and IF) antenna showers
// Inherits from SpaceShower in Pythia 8 so can be used as alternative to
// SpaceShower. Added by NF.

class VinciaISR : public Pythia8::SpaceShower {

  // Allow VinciaPlugin and VinciaFSR to access private information
  friend class VinciaPlugin;
  friend class VinciaFSR;

 public:

  // ----------------------------------------------------------------------

  // CONSTRUCTOR AND DESTRUCTOR

  // Constructor.
  VinciaISR() : isInit(false) { };

  // Destructor.
  virtual ~VinciaISR() { };

  // -----------------------------------------------------------------------

  // FUNCTIONS INHERITED FROM PYTHIA 8 FOR INTERFACING

  // Initialize generation. Possibility to force re-initialization by hand.
  void init( BeamParticle* beamAPtrIn, BeamParticle* beamBPtrIn);

  // Possible limitation of first emission.
  bool limitPTmax( Event& event, double Q2Fac = 0., double Q2Ren = 0.);

  // Prepare system for evolution; identify ME.
  void prepare( int iSys, Event& event, bool limitPTmaxIn = false);

  // Update dipole list after each ISR emission.
  void update( int iSys, Event& event, bool hasWeakRad = false);

  // Select next pT in downwards evolution.
  double pTnext( Event& event, double pTbegAll, double pTendAll,
    int nRadIn = -1, bool doTrialIn = false);

  // ME corrections and kinematics that may give failure.
  bool branch( Event& event);

  // Utility to print dipole list; for debug mainly.
  void list() const;

  // Initialize data members for calculation of uncertainty bands.
  // (So far purely dummy in Vincia)
  bool initUncertainties() { return false; }

  // Flag for failure  in branch(...) that will force a retry of parton level.
  // (Returns false in base class, and rescatterFail in simpleTimeShower)
  // (Not implemented in Vincia yet since we do not do rescattering)
  bool doRestart() const { return false; }

  // Flag for last ISR branching being gamma -> qqbar
  // (not implemented in Vincia's QED evolution yet)
  bool wasGamma2qqbar() { return false; }
    
  // Tell whether ISR has done a weak emission.
  // (not implemented in Vincia yet)
  bool getHasWeaklyRadiated() { return false; }

  // Tell which system was the last processed one.
  int system() const {return iSysWin;}

  // Potential enhancement factor of pTmax scale for hardest emission.
  // Used if limitPTmax = true.
  double enhancePTmax() const { return pTmaxFudge; }
  
  // END OF FUNCTIONS INHERITED FROM PYTHIA 8 FOR INTERFACING

  // -----------------------------------------------------------------------

  // PUBLIC NON-INHERITED MEMBERS

  // Initialise pointers to Vincia objects
  void initVinciaPtrs(Colour* colourPtrIn, VinciaFSR* fsrPtrIn,
    QEDShower* qedPtrIn,MECs* mecsPtrIn,Resolution* resolutionPtrIn,
    VinciaCommon* vinComPtrIn,VinciaWeights* vinWeightsPtrIn) {
    colourPtr     = colourPtrIn;
    fsrPtr        = fsrPtrIn;
    qedShowerPtr  = qedPtrIn;
    mecsPtr       = mecsPtrIn;
    resolutionPtr = resolutionPtrIn;
    vinComPtr     = vinComPtrIn;
    weightsPtr    = vinWeightsPtrIn;
  }

  // Initialize Pointers to Antenna sets
  void initAntPtr(AntennaSetISR* antSetIn) { antSetPtr = antSetIn; }

  // Clear all containers
  void clearContainers() {
    hasPrepared.clear();
    branchelementals.clear();
    Q2hat.clear();
    isHardSys.clear();
    isResonanceSys.clear();
    polarisedSys.clear();
    doMECsSys.clear();
    indexSav.clear();
    partsSav.clear();
    nBranch.clear();
    nBranchISR.clear();
    nG.clear();
    nQQ.clear();
    initialA.clear();
    initialB.clear();
  }

  // Function to tell if VinciaISR::prepare() has treated this system
  bool prepared(int iSys) {
    if ( hasPrepared.find(iSys) != hasPrepared.end() )
      return hasPrepared[iSys];
    else return false;
  }

  // Wrapper function to return a specific antenna inside antenna set
  AntennaFunctionIX* getAnt(int iAnt) { return antSetPtr->getAnt(iAnt); }

  // Print out the antennae
  void printAnts(int iSys);

  // Print final statistics information
  void printInfo(bool pluginCall = false );

  // Evolution Windows: Phase space region boundaries
  int getRegion(double q) {
    for (int i=1; i<(int)regMinScalesSav.size(); i++)
      if (q < regMinScalesSav[i]) return i-1;
    return (int)regMinScalesSav.size()-1;
  }

  // Evolution Windows: Phase space region boundaries
  double getQmin(int iRegion) {
    iRegion = max(0,iRegion);
    iRegion = min(iRegion,(int)regMinScalesSav.size()-1);
    return regMinScalesSav[iRegion];
  }

  // Number of active flavors
  int getNf(int iRegion) {
    if (iRegion <= 1) return 3;
    else if (iRegion <= 2) return 4;
    else if (iRegion <= 4) return 5;
    else return 6;
  }

  // Pointer to global AlphaS instance
  AlphaStrong* alphaSptr;

  // Lambda value
  double getLambda(int nFin) {
    if (nFin <= 3) return alphaSptr->Lambda3();
    else if (nFin <= 4) return alphaSptr->Lambda4();
    else if (nFin <= 5) return alphaSptr->Lambda5();
    else return alphaSptr->Lambda6();
  }

  // Add trial functions to a Branchelemental
  void resetTrialGenerators(BranchelementalISR* trial);

  // Method to check if a gluon splitting in the initial state (to get rid of
  // heavy quarks) is still possible after the current branching
  bool checkHeavyQuarkPhaseSpace(vector<Particle> parts, int iSyst);

  // Method to check if heavy quark left after passing the evolution window
  bool heavyQuarkLeft(double qTrial);

  // Function to ask if a system is hard
  bool isSysHard(int iSys) {
    if (!isInit) return false;
    if ((int)isHardSys.size() <= iSys) return false;
    return isHardSys[iSys];
  }

  vector<double> getMasses() {
    vector<double> masses;
    masses.resize(0);
    masses.push_back(mt);
    masses.push_back(mtb);
    masses.push_back(mb);
    masses.push_back(mc);
    masses.push_back(ms);
    return masses;
  }

  // Information.
  // -----------------------------------------------------------------------
  // Get number of systems
  int getNsys() {return nBranchISR.size();}
  // Get number of branchings in a system (return -1 if no such system)
  int getNbranch(int iSys = -1) {
    // iSys < 0 : return sum over all
    int n = 0;
    if (iSys < 0) for (int i=0; i<(int)nBranchISR.size(); ++i)
                    n += nBranchISR[iSys];
    else if (iSys < (int)nBranchISR.size()) n = nBranchISR[iSys];
    else n = -1;
    return n;
  }
  // Get scale of branchings; use (0,1) for first branching in 1st system.
  // (Could be extended so (i,0) would return starting scale for system i.)
  double getQbranch(int iSys, int iBranch) {
    if (iSys < 4 && iSys >= 0 && iBranch <= 10 && iBranch >= 1)
      return qBranch[iSys][iBranch];
    else return 0.;
  }
  double getPTphys(int iSys, int iBranch) {
    if (iSys < 4 && iSys >= 0 && iBranch <= 10 && iBranch >= 1)
      return pTphys[iSys][iBranch];
    else return 0.;
  }

  // Set verbose level
  void setVerbose(int verboseIn) { verbose = verboseIn; }

  bool checkAntennae(const Event& event);


 private :

  // -----------------------------------------------------------------------

  // PRIVATE MEMBERS

  // Set starting scale of shower (power vs wimpy) for system iSys
  void setStartScale(int iSys, Event& event);

  // Function to return headroom factor
  double getHeadroomFac(int iSys, int iAntPhys, double qMinNow);

  // Generate trial branching kinematics and check physical phase space
  bool generateKinematics(Event& event, BranchelementalISR* trialPtr,
    vector<Vec4>& pRec) {
    return ( trialPtr->isII()
      ? generateKinematicsII(event, trialPtr, pRec)
      : generateKinematicsIF(event, trialPtr, pRec) ); }

  bool generateKinematicsII(Event& event, BranchelementalISR* trialPtr,
    vector<Vec4>& pRec);
  bool generateKinematicsIF(Event& event, BranchelementalISR* trialPtr,
    vector<Vec4>& pRec);

  // Main trial accept function
  bool acceptTrial(const Event& event, BranchelementalISR* winnerPtr);

  // Method to assign colour flow
  bool assignColourFlow(Event& event, BranchelementalISR* trialPtr);

  // Initialised?
  bool isInit;
  bool isPrepared;

  // Possibility to allow user veto of emission step.
  bool hasUserHooks, canVetoEmission;

  // Beams, saved as positive and negative pz respectively. 
  int beamFrameType;
  BeamParticle* beamPosPtr;
  BeamParticle* beamNegPtr;
  double eBeamPos, eBeamNeg, eCMBeamsSav, m2BeamsSav;

  double TINYPDF;

  // Main ISR on/off switch
  bool doISR;

  // Map of which systems ISR::prepare() has treated
  map<int, bool> hasPrepared;

  // Shower parameters
  bool helicityShower, sectorShower, convGluonToQuarkI, convQuarkToGluonI;
  bool kineMapIFretry;
  int nGluonToQuarkI, nGluonToQuarkF;
  double cutoffScaleII, cutoffScaleIF;
  bool evAllPT;
  int nFlavZeroMass;

  // Factorization scale and shower starting settings
  int    pTmaxMatch, pTdampMatch;
  double pTmaxFudge, pT2maxFudge, pT2maxFudgeMPI, pT2dampFudge;

  // AlphaS parameters
  bool useCMW;
  int alphaSorder;
  double alphaSvalue, alphaSmax, alphaSmuFreeze, alphaSmuMin;
  double aSkMu2EmitI, aSkMu2SplitI, aSkMu2SplitF, aSkMu2Conv;
  double mt, mtb, ms, mb, mc;
  // Calculated values
  double mu2freeze, mu2min;

  // Trial Generators
  TrialIISoft    trialIISoft;
  TrialIIGCollA  trialIIGCollA;
  TrialIIGCollB  trialIIGCollB;
  TrialIISplitA  trialIISplitA;
  TrialIISplitB  trialIISplitB;
  TrialIIConvA   trialIIConvA;
  TrialIIConvB   trialIIConvB;
  TrialIFSoft    trialIFSoft;
  TrialVFSoft  trialVFSoft;
  TrialIFGCollA  trialIFGCollA;
  TrialIFSplitA  trialIFSplitA;
  TrialIFSplitK  trialIFSplitK;
  TrialIFConvA   trialIFConvA;

  // Uncertainty Bands and Weights
  bool uncertaintyBands, allowReweighting;
  double uncertaintyBandsCutoff;

  // Enhanceing switches and parameters
  bool enhanceInHard, enhanceInResDec, enhanceInMPI;
  double enhanceAll, enhanceBottom, enhanceCharm, enhanceCutoff;

  // Colour Reconnection parameters
  bool doCR, CRjunctions;

  // Pointer to VINCIA objects
  AntennaSetISR* antSetPtr;
  MECs*          mecsPtr;
  Colour*        colourPtr;
  Resolution*    resolutionPtr;
  QEDShower*     qedShowerPtr;
  VinciaFSR*     fsrPtr;
  VinciaCommon*  vinComPtr;
  VinciaWeights* weightsPtr;

  // Total and MEC accept probability
  vector<double> Paccept;

  // Evolution windows
  vector<double> regMinScalesMtSav;
  vector<double> regMinScalesSav;
  vector<double> regMinScalesNow;

  // vector of dipoles (with trial branchings, 4 at most)
  vector<BranchelementalISR > branchelementals;

  // Current winner
  BranchelementalISR* winnerPtr;
  int indxWin;
  int iSysWin;

  // ------------------ Every system gets own copy ------------------------- //

  // Flags to tell a few basic properties of each parton system.
  map<int, bool> isHardSys, isResonanceSys, polarisedSys, doMECsSys;

  // Saved particle state and number in event record
  map<int, vector< Particle > > partsSav;
  map<int, vector< int      > > indexSav;

  // Save initial ISR starting scale system by system
  map<int, double> Q2hat;

  // Count the number of branchings in the system
  map<int, int> nBranch, nBranchISR;

  // Saved incoming guys
  map<int, Particle> initialA;
  map<int, Particle> initialB;
  double eBeamPosUsed;
  double eBeamNegUsed;

  // Count numbers of quarks and gluons
  map<int, int> nG, nQQ;

  // ---------------------------- //

  // Statistics
  long nTrialsSum;
  vector<long> nTrials, nTrialsAccepted, nFailedVeto, nFailedHull, nFailedKine;
  vector<long> nFailedMass, nFailedCutoff, nClosePSforHQ, nSectorReject;
  vector<double> rFailedVetoPDF;
  int nForceSplit;

  // Rescue mechanism
  bool doRescue;
  int nRescue;
  double rescueMin;

  // Verbose setting, P>1 warning settings, and version number
  int verbose;
  double pAcceptFuss, pAcceptPT0;
  double qBranch[4][11], pTphys[4][11];

  // Counter for numbers of events
  long nAccepted, nSelected;
  int nVetoUserHooks, nFailHadLevel, nCallPythiaNext;

};

}

#endif
