// VinciaMG5cxxInterface.h is a part of the PYTHIA event generator.
// Copyright (C) 2019 Torbjorn Sjostrand.
// PYTHIA is licenced under the GNU GPL v2 or later, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// This file contains the functionality to interface C++ matrix elements
// generated with the PY8Kernels/MG5MES plugin to MadGraph 5

#ifndef VINCIA_MG5CXXinterface_H
#define VINCIA_MG5CXXinterface_H

// Include Pythia headers
#include "Pythia8/Basics.h"
#include "Pythia8/PythiaComplex.h"
#include "Pythia8/Event.h"
#include "Pythia8/Info.h"
#include "Pythia8/ParticleData.h"
#include "Pythia8/PythiaStdlib.h"
#include "Pythia8/Settings.h"
#include "Pythia8/StandardModel.h"
#include "Pythia8/SusyLesHouches.h"

// Include MG5 PY8MEs plugin headers (if preprocessor flag MG5MES has been set)
#ifdef MG5MES
#include "Processes_sm/PY8ME.h"
#include "Processes_sm/PY8MEs.h"
#endif

// Include Vincia Standard Library
#include "Pythia8/VinciaCommon.h"

namespace Pythia8 {

#ifdef MG5MES
using namespace PY8MEs_namespace;
#endif

//==========================================================================

class VinciaMG5cxxInterface {

 public:

  // Constructor (depends on whether MG5MES enabled)
  VinciaMG5cxxInterface() {
    isInitPtr = false;
    isInit    = false;
    mg5libPtr = NULL;
  };

  // Destructor
  virtual ~VinciaMG5cxxInterface() {
    if (mg5libPtr != NULL) delete mg5libPtr;
  };

  // Set pointers to required PYTHIA 8 objects
  void initPtr(ParticleData* particleDataPtrIn,Settings* settingsPtrIn,
    Couplings* couplingsPtrIn, Rndm* rndmPtrIn, SusyLesHouches* slhaPtrIn,
    VinciaCommon* vinComPtrIn) {
    couplingsPtr    = couplingsPtrIn;
    particleDataPtr = particleDataPtrIn;
    settingsPtr     = settingsPtrIn;
    rndmPtr         = rndmPtrIn;
    slhaPtr         = slhaPtrIn;
    vinComPtr       = vinComPtrIn;
    isInitPtr       = true;
  }

  // Main initialization routine.
  bool init();

  // Get pointer to matrix element (eg to check if process exists in library)
  // Use #ifdef to decide whether the MG5 ME plugin is actually linked,
  // otherwise dummy implementation which returns NULL.
#ifdef MG5MES
  PY8ME* getProcess(vector<int> idIn, vector<int> idOut, set<int> sChan) {
    // Verbose output: tell which process we are looking for.
    if (verbose >= 9) {
      cout<<" VinciaMG5interface::getProcess(): checking for process";
      for (int i=0; i<(int)idIn.size(); ++i) cout<<" "<<idIn[i];
      cout<<" > ";
      for (int i=0; i<(int)idOut.size(); ++i) cout<<" "<<idOut[i];
      cout<<endl;
    }
    if (mg5libPtr != NULL && mg5libPtr != 0) {
      return mg5libPtr->getProcess(idIn, idOut, sChan);
    }
    cout<<"      returning NULL"<<endl;
    return NULL;
  }
#endif

  // Get the matrix element squared for a particle state
  double ME2( vector<Particle> state, int nIn);

  // Use ME2 to set helicities for a state.
  // (Note: takes a reference as input and operates on it)
  bool selectHelicities( vector<Particle>& state, int nIn);

  // Set colour depth
  void setColourDepth(int colourDepthIn) {colourDepth = colourDepthIn;}
  // Get colour depth
  int getColourDepth() {return colourDepth;}

  // Convert a process label to a string, e.g. for printing to stdout
  string makeLabel(vector<int>& id, int nIn, bool convertToNames=false);

  // Set verbose level
  void setVerbose(int verboseIn) {verbose = verboseIn;}

 private:

  // Is initialized?
  bool isInitPtr, isInit;

  // Saved list of helicity components for last ME evaluated
  map< vector<int> , double > me2hel;

  // Pointers to VINCIA and PYTHIA 8 objects
  Couplings*      couplingsPtr;
  ParticleData*   particleDataPtr;
  Rndm*           rndmPtr;
  Settings*       settingsPtr;
  VinciaCommon*   vinComPtr;
  SusyLesHouches* slhaPtr;

  // MG5 matrix-element library
#ifdef MG5MES
  PY8MEs*         mg5libPtr;
#else
  int*            mg5libPtr;
#endif

  // Colour Mode (0: leading colour, 1: Vincia colour)
  int colourDepth;

  // Verbose level
  int verbose;

};

//==========================================================================

} // end namespace Pythia8

#endif
