// VinciaWeights.h is a part of the PYTHIA event generator.
// Copyright (C) 2019 Torbjorn Sjostrand.
// PYTHIA is licenced under the GNU GPL v2 or later, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// This file contains header information for the VinciaWeights class.

#ifndef Vincia_VinciaWeights_H
#define Vincia_VinciaWeights_H

#include "Pythia8/Info.h"
#include "Pythia8/Settings.h"
#include "Pythia8/VinciaCommon.h"

namespace Pythia8 {

class VinciaWeights {

 public:

  // Constructor.
  VinciaWeights() {}

  // Destructor.
  ~VinciaWeights() {}

  // Initilize Pointers.
  bool initPtr(Settings* settingsPtrIn, Info* infoPtrIn,
    VinciaCommon* vinComPtrIn) {
    settingsPtr = settingsPtrIn;
    infoPtr     = infoPtrIn;
    vinComPtr   = vinComPtrIn;
    isInitPtr   = true;
    return true;
  }

  // Initilization.
  void init();

  // Check if particular weight exists.
  bool existsWeight(int iWeightIn=0) {
    if (iWeightIn >= 0 && iWeightIn < nWeightsSav) return true;
    else return false;
  }

  // Get weight of current event weight
  // Normally 1 for set 0 (user settings)
  // Use iWeightIn > 0 to access uncertainty weights
  double weight(int iWeightIn=0) {
    if (existsWeight(iWeightIn)) return weightsSav[iWeightIn];
    return 0.0;
  }

  // Access the weight labels
  string weightLabel(int iWeightIn=0) {
    if (iWeightIn == 0) return "no variation";
    if(existsWeight(iWeightIn) && iWeightIn-1 < (int)varLabels.size())
      return varLabels[iWeightIn-1];
    return "";
  }

  // Functions to access by VINCIA plugin
  int  nWeights()         { return nWeightsSav; }
  bool reweightingOccurred() { return didReweight; }

  // Reset the weights
  void resetWeights(int nAccepted);

  // Scale all event weights
  void scaleWeightAll(double scaleFacIn) {
    for (int iWeight=0; iWeight<nWeightsSav; iWeight++)
      weightsSav[iWeight] *= scaleFacIn;
  }

  // Scale a particular event weight
  void scaleWeight(double scaleFacIn, int iWeightIn=0) {
    if (existsWeight(iWeightIn)) weightsSav[iWeightIn] *= scaleFacIn;
  }

  // Scale the uncertainty band weights
  void scaleWeightVar(vector<double> Paccept, bool accept, bool isHard);

  // Scale the uncertainty band weights if branching is accepted
  void scaleWeightVarAccept(vector<double> Paccept);

  // Scale the uncertainty band weights if branching is rejected
  void scaleWeightVarReject(vector<double> Paccept);

  // Enhanced kernels: reweight if branching is accepted
  void scaleWeightEnhanceAccept(double enhanceFac=1.) {
    if (enhanceFac == 1.0) return;
    else scaleWeightAll(1./enhanceFac);
  }

  // Enhanced kernels: reweight if branching is rejected
  void scaleWeightEnhanceReject(double PacceptUnenhanced,
    double enhanceFac=1.) {
    if (enhanceFac == 1.0) return;
    if (enhanceFac > 1.0) {
      double rRej =
        (1. - PacceptUnenhanced/enhanceFac)/(1. - PacceptUnenhanced);
      scaleWeightAll(rRej);
    } else {
      double rRej =
        (1. - PacceptUnenhanced)/(1. - enhanceFac*PacceptUnenhanced);
      scaleWeightAll(rRej);
    }
  }

  // Helper function for keyword evaluation
  int doVarNow(string keyIn, int iAntPhys, string type) ;

  // Helper function for antenna function
  double ant(double antIn, double cNSIn) {
    if (allowNegAnt) return (antIn+cNSIn);
    return max(TINYANT,antIn+cNSIn);
  }

  // Main function to perform the ISR uncertainty variations
  // TMP: disabled. Needs new structure.
  // (fsrTrial no longer has information on new state inside)
  /*
    void doVariationsISR(vector<double>& Paccept, bool isHard, bool doMEC,
    BranchelementalISR* isrTrial, int iTrial, AlphaStrong* asPtr,
    double asUsr, double mu2usr, double mu2min, double mu2freeze,
    double asMax, bool isPolarized, double antUsr, double helSumUsr,
    double aHelUsr);
  */

  // Main function to perform the FSR uncertainty variations
  // TMP: disabled. Needs new structure.
  // (fsrTrial no longer has information on new state inside)
  /*
    void doVariationsFSR(vector<double>& Paccept, bool isHard, bool doMEC,
    BranchelementalFSR* fsrTrial, int iTrial, AlphaStrong* asPtr,
    double asUsr, double mu2usr, double mu2min, double mu2freeze,
    double asMax, bool isPolarized, double antUsr, double helSumUsr,
    double aHelUsr);
  */

  // Main function to perform the weighting
  void doWeighting();

  // Verbose
  int verbose;

  // Pointers
  Settings*     settingsPtr;
  Info*         infoPtr;
  VinciaCommon* vinComPtr;

  // Internal flag
  bool isInitPtr;

  // Constants
  static const double TINYANT, PACCEPTVARMAX, MINVARWEIGHT;

  // Parameters taken from settings
  bool   uncertaintyBands;
  vector<string> varLabels;
  vector< vector<string> > varKeys;
  vector< vector<double> > varVals;
  double deltaASmax, cNSpTmin;
  bool   varInMPIshowers, muSoftCorr, allowNegAnt;
  bool   printPaccVarViol, printReWeightViol;

  // Helper parameters
  vector<string> allKeywords;
  map<int,string> iAntToKeyFSR, iAntToKeyISR;
  double nWeightsSav;
  int    nReportWeight, nReportedWeight;
  bool   doAlphaSvar, doFiniteVar;

  // Weights (from input events, enhancement, MC violations, or variations)
  vector<double> weightsSav; // Weights
  vector<double> weightsOld; // Weights from last step
  vector<double> weightsMax, weightsMin;
  // WeightSums
  vector<double> weightSum, weightSum2;
  // Contribution of current event
  vector<double> contribSum, contribSum2;

  // Counters:
  // - total number of weights = nr of events
  int nTotWeights;
  // - for initial and main weight (MC violation only)
  int nNonunityWeight, nNegativeWeight,
    nNonunityInitialWeight, nNegativeInitialWeight;
  // - for current event
  int nNonunityWeightNow, nNegativeWeightNow,
    nNonunityInitialWeightNow, nNegativeInitialWeightNow;

  // Flag to tell if reweighting due to MC violations occured
  bool didReweight;

  // doWeighting called for the first time?
  bool firstCall;

};

}

#endif
